/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


// DID type string length
#define MAX_TYPE_LENGTH  5
// DID type strings
char          DIDTypeString[4][MAX_TYPE_LENGTH]= {"PID", "MID", "TID", "INF"};


/*
*******************************************************************************
**  Function:   VerifyDidSupportData
**
**  Purpose:    Verify each controller supports at a minimum one PID.
**             Any ECU that responds that does not support at least
**              one PID is flagged as an error.
**
*******************************************************************************
*/
STATUS VerifyDidSupportData ( DIDTYPE DidType )
{
	STATUS          bReturn = PASS;
	int             bEcuResult;
	unsigned long   EcuIndex;
	unsigned long   Index;

	unsigned char   SupSize;
	ID_SUPPORT     *pSup;


	// For each ECU
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		if ( DidType == PIDREQUEST )
		{
			SupSize = gOBDResponse[EcuIndex].PidSupportSize;
			pSup = &gOBDResponse[EcuIndex].PidSupport[0];
		}
		else if ( DidType == MIDREQUEST )
		{
			SupSize = gOBDResponse[EcuIndex].MidSupportSize;
			pSup = &gOBDResponse[EcuIndex].MidSupport[0];
		}
		else
		{
			SupSize = gOBDResponse[EcuIndex].InfSupportSize;
			pSup = &gOBDResponse[EcuIndex].InfSupport[0];
		}

		bEcuResult = FAIL;
		// PID $00 not required to indicate PIDs Supported
		for ( Index = 1; Index < SupSize; Index++ )
		{
			// If PID is supported, keep looking
			if ( pSup[Index].IDBits[0] ||
			     pSup[Index].IDBits[1] ||
			     pSup[Index].IDBits[2] ||
			     (pSup[Index].IDBits[3] & 0xFE) != 0x00 )
			{
				bEcuResult = PASS;
				break;
			}
		}

		if ( bEcuResult == FAIL && SupSize > 1 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  DID-Supported DIDs indicate no %ss supported",
			      GetEcuId ( EcuIndex ),
			      DIDTypeString[DidType] );
			bReturn = FAIL;
		}
	}

	return bReturn;
}


/*
********************************************************************************
**  FUNCTION    RequestDIDSupportData
**
**  Purpose     Isolated function to identify support for SID 22 DID x
********************************************************************************
*/
STATUS RequestDIDSupportData ( DIDTYPE DidType, BOOL bVerifyLinkActive )
{
	unsigned long EcuIndex = 0;
	unsigned int  IdIndex;
	BOOL          bDIDSupport;  // Evaluate $E0 DID support indication.
	SID_REQ       SidReq;

	unsigned int  DidMsb;
	ID_SUPPORT   *pSup;

	BOOL          bTestFailed = FALSE;


	if ( DidType == PIDREQUEST )
	{
		DidMsb = 0xF400;
	}
	else if ( DidType == MIDREQUEST )
	{
		DidMsb = 0xF600;
	}
	else
	{
		DidMsb = 0xF800;
	}


	// Request SID 22 support data
	for ( IdIndex = 0x00; IdIndex < MAX_PIDS; IdIndex += 0x20 )
	{
		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = DidMsb + IdIndex;
		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Error in response to DID-Supported DID $%04X request\n",
			      DidMsb + IdIndex );
			bTestFailed = TRUE;
		}

		// Check if we need to request the next group
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( ( DidType == PIDREQUEST && (gOBDResponse[EcuIndex].PidSupport[(IdIndex >> 5)].IDBits[3] & 0x01) ) ||
			     ( DidType == MIDREQUEST && (gOBDResponse[EcuIndex].MidSupport[(IdIndex >> 5)].IDBits[3] & 0x01) ) ||
			     ( DidType == INFREQUEST && (gOBDResponse[EcuIndex].InfSupport[(IdIndex >> 5)].IDBits[3] & 0x01) ) )
			{
				break;
			}
		}
		// If no ECU supports the next DID support DID, end requests
		if ( EcuIndex >= gOBDNumEcus )
		{
			break;
		}
	}


	// Verify support information
	if ( EcuIndex != 0 && VerifyDidSupportData ( DidType ) == FAIL )
	{
		bTestFailed = TRUE;
	}


	// Verify support information if request is at upper limit of $E0
	if ( ( DidType == PIDREQUEST && IdIndex == 0x1E0 ) ||
	     ( (DidType == MIDREQUEST || DidType == INFREQUEST) && IdIndex == 0xE0 ) )
	{
		// Init variable to no-support
		bDIDSupport = FALSE;

		// For each ECU
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( DidType == PIDREQUEST )
			{
				pSup = &gOBDResponse[EcuIndex].PidSupport[0];
			}
			else if ( DidType == MIDREQUEST )
			{
				pSup = &gOBDResponse[EcuIndex].MidSupport[0];
			}
			else
			{
				pSup = &gOBDResponse[EcuIndex].InfSupport[0];
			}

			// If DID is supported, keep looking
			if ( pSup[IdIndex >> 5].IDBits[0] ||
			     pSup[IdIndex >> 5].IDBits[1] ||
			     pSup[IdIndex >> 5].IDBits[2] ||
			     (pSup[IdIndex >> 5].IDBits[3] & 0xFE) != 0x00  )
			{
				// Flag as support indicated!
				bDIDSupport = TRUE;
			}
		}

		// Flag as error if no support indicated in $E0
		if ( bDIDSupport == FALSE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "DID $%04X support failure.  No DID support indicated!\n",
			      DidMsb + IdIndex);
			bTestFailed = TRUE;
		}
	}
	// Request next unsupported DID-Supported DID
	else if ( DidType != INFREQUEST || gTestPhase != eTestNoDTC || gTestSubsection != 17 )  // IF not Test 5.17
	{
		IdIndex += 0x20;

		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = DidMsb + IdIndex;

		if ( SidRequest ( &SidReq, SID_REQ_NORMAL|SID_REQ_IGNORE_NO_RESPONSE ) != FAIL  )
		{
			for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
			{
				if ( (DidType == PIDREQUEST && gOBDResponse[EcuIndex].PidSize != 0) ||
				     (DidType == MIDREQUEST && gOBDResponse[EcuIndex].MidSize != 0) ||
				     (DidType == INFREQUEST && gOBDResponse[EcuIndex].InfSize != 0) )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  Unexpected response to unsupported SID $22 DID $%04X!\n",
					      GetEcuId ( EcuIndex ),
					      DidMsb + IdIndex );
				}
			}
			bTestFailed = TRUE;
		}
	}

	// Verify ECU did not drop out
	if ( bVerifyLinkActive && VerifyLinkActive () != PASS )
	{
		bTestFailed = TRUE;
	}

	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}

	return PASS;
}


/*
*******************************************************************************
**  RequestGroupDIDSupportData - Function to verify DID group support
*******************************************************************************
*/
STATUS RequestGroupDIDSupportData ( DIDTYPE DidType )
{
	unsigned long EcuIndex;
	SID_REQ       SidReq;
	unsigned int  DidMsb;
	unsigned int  NumIds;
	unsigned int  NextSupportOffset = 2;
	ID_SUPPORT   *pSup;
	unsigned char SupSize;
	ID_SUPPORT   *pCompSup;
	unsigned char CompSupSize;
	BOOL          bReqNextGroup;  // Binary logic for request for next group!
	BOOL          bTestFailed = FALSE;

	// Save the original response data for comparison
	memcpy ( &gOBDCompareResponse[0], &gOBDResponse[0], sizeof ( OBD_DATA ) * gNumEcusForAlloc );

	// Set starting DID Support DID
	if ( DidType == PIDREQUEST )
	{
		DidMsb = 0xF400;
	}
	else if ( DidType == MIDREQUEST )
	{
		DidMsb = 0xF600;
	}
	else
	{
		DidMsb = 0xF800;
	}


	do
	{
		bReqNextGroup = FALSE;
		// MID and INF Support's highest group request is $F4C0 $F4E0
		// But     PID Support's highest group request is $F4C0 $F4E0 $F500
		if ( (DidMsb & 0x00FF) < 0xC0 ||
		     DidType == PIDREQUEST )
		{
			NumIds = 3;
		}
		else
		{
			NumIds = 2;
		}
		
		// Request DID support data as a group
		SidReq.SID    = 0x22;
		SidReq.NumIds = NumIds;
		SidReq.DID[0] = DidMsb;
		SidReq.DID[1] = DidMsb + 0x20;
		if ( NumIds == 3 )
		{
			SidReq.DID[2] = DidMsb + 0x40;
		}
		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "SID $22 group $%04X-$%04X support request failed\n",
			      DidMsb,
			      DidMsb + (0x20 * (NumIds-1)) );
			bTestFailed = TRUE;
		}
		else
		{
			// Set next group select to false
			bReqNextGroup = FALSE;
	
			// Determine if next group support request is required
			for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
			{
				if ( ( DidType == PIDREQUEST && (gOBDResponse[EcuIndex].PidSupport[NextSupportOffset].IDBits[3] & 0x01) ) ||
				     ( DidType == MIDREQUEST && (gOBDResponse[EcuIndex].MidSupport[NextSupportOffset].IDBits[3] & 0x01) ) ||
				     ( DidType == INFREQUEST && (gOBDResponse[EcuIndex].InfSupport[NextSupportOffset].IDBits[3] & 0x01) ) )
				{
					bReqNextGroup = TRUE;
					DidMsb += 0x60;
					NextSupportOffset += 3;
					break;
				}
			}
		}
	}
	while ( bReqNextGroup == TRUE );

	// Verify that all INF group/individual support data compares
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		if ( DidType == PIDREQUEST )
		{
			SupSize = gOBDResponse[EcuIndex].PidSupportSize;
			pSup = &gOBDResponse[EcuIndex].PidSupport[0];
			CompSupSize = gOBDCompareResponse[EcuIndex].PidSupportSize;
			pCompSup = &gOBDCompareResponse[EcuIndex].PidSupport[0];
		}
		else if ( DidType == MIDREQUEST )
		{
			SupSize = gOBDResponse[EcuIndex].MidSupportSize;
			pSup = &gOBDResponse[EcuIndex].MidSupport[0];
			CompSupSize = gOBDCompareResponse[EcuIndex].MidSupportSize;
			pCompSup = &gOBDCompareResponse[EcuIndex].MidSupport[0];
		}
		else
		{
			SupSize = gOBDResponse[EcuIndex].InfSupportSize;
			pSup = &gOBDResponse[EcuIndex].InfSupport[0];
			CompSupSize = gOBDCompareResponse[EcuIndex].InfSupportSize;
			pCompSup = &gOBDCompareResponse[EcuIndex].InfSupport[0];
		}

		if ( CompareDIDSupport ( EcuIndex,
		                         DidType,
		                         SupSize,
		                         CompSupSize,
		                         pSup,
		                         pCompSup,
		                         TRUE ) == FAIL )
		{
			bTestFailed = TRUE;
		}
	}

	if ( bTestFailed == TRUE )
	{
		// Something was wrong, restore the original response data
		memcpy ( &gOBDResponse[0], &gOBDCompareResponse[0], sizeof ( OBD_DATA ) * gNumEcusForAlloc );
		return FAIL;
	}

	return PASS;
}


/*
********************************************************************************
**
**  FUNCTION  CompareDIDSupport
**
**  Purpose   Isolated function to compare normal/single and reverse/group
**            PID Support responses for SID $22 DID $F4xx, 5, $F6xx, 8, $F8xx
**
********************************************************************************
*/
STATUS CompareDIDSupport ( unsigned long EcuIndex,       // The index of the current ECU
                           DIDTYPE       DidType,        // Type of DID support array
                           unsigned char ResponseSize,   // Size of group/reverse DID Support response
                           unsigned char CompareSize,    // Size of single/normal DID Support response
                           ID_SUPPORT    ResponseData[], // Data of group/reverse DID Support response
                           ID_SUPPORT    CompareData[],  // Data of single/normal DID Support response
                           BOOL          bGroup )        // Flag indicating if this is a group DID Support response
{
	unsigned int  IdIndex;
	unsigned int  IdBitIndex;
	unsigned int  BitIndex;
	unsigned char BitMask;
	BOOL          bTestFailed = FALSE;

	if ( ResponseSize != CompareSize )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  SID $22 %s Supported %s %s size mismatch\n"
		      "                          %s response size = %d\n"
		      "                          %s response size = %d\n",
		      GetEcuId ( EcuIndex ),
		      DIDTypeString[DidType],
		      DIDTypeString[DidType],
		      bGroup ? "group/single" : "reverse/normal",
		      bGroup ? "Group " : "Reverse",
		      ResponseSize,
		      bGroup ? "Single" : "Normal ",
		      CompareSize );
		bTestFailed = TRUE;
	}

	// Verify that all SID X PID support data compares
	for ( IdIndex = 0; IdIndex < 8; IdIndex++ )
	{
		if ( CompareData[IdIndex].FirstID != ResponseData[IdIndex].FirstID )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  SID $22 %s $%02X %s support response Support ID mismatch\n",
			      GetEcuId ( EcuIndex ),
			      DIDTypeString[DidType],
			      IdIndex * 0x20,
			      bGroup ? "group/single" : "reverse/normal",
			      CompareData[IdIndex].FirstID );
			bTestFailed = TRUE;

			if ( CompareData[IdIndex].FirstID != IdIndex * 0x20 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  SID $22 %s $%02X invalid %s support response Support ID ($%02X)\n",
				      GetEcuId ( EcuIndex ),
				      DIDTypeString[DidType],
				      IdIndex * 0x20,
				      bGroup ? "single" : "normal",
				      CompareData[IdIndex].FirstID );
			}
			if ( ResponseData[IdIndex].FirstID != IdIndex * 0x20 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  SID $22 %s $%02X invalid %s support response Support ID ($%02X)\n",
				      GetEcuId ( EcuIndex ),
				      DIDTypeString[DidType],
				      IdIndex * 0x20,
				      bGroup ? "group" : "reverse",
				      ResponseData[IdIndex].FirstID );
			}
		}

		if ( (ResponseData[IdIndex].IDBits[0] != CompareData[IdIndex].IDBits[0] ) ||
		     (ResponseData[IdIndex].IDBits[1] != CompareData[IdIndex].IDBits[1] ) ||
		     (ResponseData[IdIndex].IDBits[2] != CompareData[IdIndex].IDBits[2] ) ||
		     (ResponseData[IdIndex].IDBits[3] != CompareData[IdIndex].IDBits[3] ) )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  SID $22 %s $%02X %s support response data mismatch\n"
			      "                         %7s response = %02X  %02X  %02X  %02X\n"
			      "                         %7s response = %02X  %02X  %02X  %02X\n",
			      GetEcuId ( EcuIndex ),
			      DIDTypeString[DidType],
			      IdIndex * 0x20,
			      bGroup ? "group/single" : "reverse/normal",
			      bGroup ? "Group " : "Reverse",
			      ResponseData[IdIndex].IDBits[0],
			      ResponseData[IdIndex].IDBits[1],
			      ResponseData[IdIndex].IDBits[2],
			      ResponseData[IdIndex].IDBits[3],
			      bGroup ? "Single" : "Normal ",
			      CompareData[IdIndex].IDBits[0],
			      CompareData[IdIndex].IDBits[1],
			      CompareData[IdIndex].IDBits[2],
			      CompareData[IdIndex].IDBits[3] );
			bTestFailed = TRUE;

			for ( IdBitIndex = 0; IdBitIndex < 4; IdBitIndex++ )
			{
				if ( (BitMask = (ResponseData[IdIndex].IDBits[IdBitIndex] ^ CompareData[IdIndex].IDBits[IdBitIndex])) != 0 )
				{
					for ( BitIndex = 8; BitIndex != 0; BitIndex--)
					{
						if ( (BitMask & 0x80) != 0 )
						{
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "ECU %X  SID $22 %s $%02X Byte %d:  %s $%02X support response mismatch\n",
							      GetEcuId ( EcuIndex ),
							      DIDTypeString[DidType],
							      IdIndex * 0x20,
							      IdBitIndex,
							      DIDTypeString[DidType],
							      (IdIndex * 0x20) + (IdBitIndex * 8) + (8 - (BitIndex - 1)) );
						}
						BitMask <<= 1;
					}
				}
			}
		}
	}

	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}
	else
	{
		return PASS;
	}
}


/*
*****************************************************************************
**
**  Function:  IsDIDSupported
**
**  Purpose:   Determine if SID $22 DID $xx is supported on specific ECU.
**             Need to have called RequestDIDSupportData() previously.
**             If EcuIndex < 0 then check all ECUs.
**
*****************************************************************************
*/
BOOL IsDIDSupported ( unsigned int EcuIndex, unsigned int DidIndex )
{
	unsigned int index1;
	unsigned int index2;
	unsigned int mask;

	DIDTYPE      DidType;


	if ( (DidIndex & 0xFF00) == 0xF400 || (DidIndex & 0xFF00) == 0xF500 )
	{
		DidIndex -= 0xF400;
		DidType = PIDREQUEST;
	}
	else if ( (DidIndex & 0xFF00) == 0xF600 )
	{
		DidIndex &= 0x00FF;
		DidType = MIDREQUEST;
	}
	else if ( (DidIndex & 0xFF00) == 0xF800 )
	{
		DidIndex &= 0x00FF;
		DidType = INFREQUEST;
	}

	if ( DidIndex == 0 )
	{
		return TRUE;            // all modules must support PID/MID/INF $00
	}

	DidIndex--;
	index1 = DidIndex >> 5;
	index2 = (DidIndex >> 3) & 0x03;
	mask   = 0x80 >> (DidIndex & 0x07);

	if ( (signed int)EcuIndex < 0 )
	{
		for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
		{
			if ( ( DidType == PIDREQUEST && (gOBDResponse[EcuIndex].PidSupport[index1].IDBits[index2] & mask) ) ||
			     ( DidType == MIDREQUEST && (gOBDResponse[EcuIndex].MidSupport[index1].IDBits[index2] & mask) ) ||
			     ( DidType == INFREQUEST && (gOBDResponse[EcuIndex].InfSupport[index1].IDBits[index2] & mask) ) )
			{
				return TRUE;
			}
		}
	}
	else
	{
		if ( ( DidType == PIDREQUEST && (gOBDResponse[EcuIndex].PidSupport[index1].IDBits[index2] & mask) ) ||
		     ( DidType == MIDREQUEST && (gOBDResponse[EcuIndex].MidSupport[index1].IDBits[index2] & mask) ) ||
		     ( DidType == INFREQUEST && (gOBDResponse[EcuIndex].InfSupport[index1].IDBits[index2] & mask) ) )
		{
			return TRUE;
		}
	}

	return FALSE;
}
