/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


/*  Local Funtion prototypes  */
STATUS SetupRequestMSG    ( SID_REQ *, PASSTHRU_MSG * );
STATUS ProcessISO15765Msg ( SID_REQ *, PASSTHRU_MSG *, unsigned long *, unsigned long *, unsigned long *, unsigned long *, unsigned long * );

/*  External Funtion prototypes  */
extern STATUS SidResetResponseData ( PASSTHRU_MSG * );
extern STATUS SidSaveResponseData ( PASSTHRU_MSG *, SID_REQ *, unsigned long * );
extern STATUS LookupEcuIndex ( PASSTHRU_MSG *RxMsg, unsigned long *pulEcuIndex );


/*  Wait / Pending data  */
static unsigned long ulEcuWaitFlags = 0;            // up to 32 ECUs
static unsigned long ulResponsePendingDelay = 0;
static unsigned char bPadErrorPermanent = FALSE;


/*
*******************************************************************************
**  SidRequest - Function to request a service ID
**
**  Returns:    RETRY  - NRC=$21 durning initialization of ISO 15765
**              ERRORS - One or more correct early/late responses
**              FAIL   - No response, wrong response, or catestrophic error
*	             PASS   - One or more correct responses, all on time
*******************************************************************************
*/
STATUS SidRequest ( SID_REQ *SidReq, unsigned long Flags )
{
	PASSTHRU_MSG  RxMsg;
	PASSTHRU_MSG  TxMsg;
	unsigned long NumMsgs;
	unsigned long RetVal;
	unsigned long StartTimeMsecs;
	unsigned long NumResponses;
	unsigned long NumFirstFrames;
	unsigned long TxTimestamp;
	unsigned long ExtendResponseTimeMsecs;
	unsigned long SOMTimestamp;
	unsigned char fFirstResponse;
	unsigned long ulResponseTimeoutMsecs;
	char          bString[MAX_LOG_STRING_SIZE];
	unsigned long EcuTimingIndex;

	STATUS eReturnCode = PASS;    // saves the return code from function calls


	// Initialize local variables
	ExtendResponseTimeMsecs = 0;
	TxTimestamp             = 0;
	SOMTimestamp            = 0;

	// Reset wait variables
	ulEcuWaitFlags          = 0;
	ulResponsePendingDelay  = 0;

	// Initialize ECU variables
	for ( EcuTimingIndex = 0; EcuTimingIndex < gNumEcusForAlloc; EcuTimingIndex++ )
	{
		gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs = 0;
		gEcuTimingData[EcuTimingIndex].ResponsePendingDelay    = 0;
		gEcuTimingData[EcuTimingIndex].FFReceived              = FALSE;
		gEcuTimingData[EcuTimingIndex].NAKReceived             = FALSE;
	}

	// if gSuspendLogOutput is true, then clear buffer
	if ( gSuspendLogOutput == TRUE )
	{
		SaveTransactionStart ();
	}

	// If not burst test, stop tester present message and delay
	// before each request to avoid exceeding minimum OBD request timing
	if ( (Flags & SID_REQ_NO_PERIODIC_DISABLE) == 0)
	{
		// Stop the tester present message before each request
		if ( StopPeriodicMsg ( TRUE ) == FAIL )
		{
			return FAIL;
		}

		gOBDList[gOBDListIndex].TesterPresentID = -1;

		Sleep ( gOBDRequestDelay );
	}

	// Setup request message based on the protocol
	if ( SetupRequestMSG ( SidReq, &TxMsg ) != PASS )
	{
		return FAIL;
	}

	// Clear the transmit queue before sending request
	RetVal = PassThruIoctl ( gOBDList[gOBDListIndex].ChannelID, CLEAR_TX_BUFFER, NULL, NULL );

	if ( RetVal != STATUS_NOERROR )
	{
		Log ( J2534_FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "%s returned %ld",
		      "PassThruIoctl(CLEAR_TX_BUFFER)",
		      RetVal );
		return FAIL;
	}

	// Clear the receive queue before sending request
	RetVal = PassThruIoctl ( gOBDList[gOBDListIndex].ChannelID, CLEAR_RX_BUFFER, NULL, NULL );

	if ( RetVal != STATUS_NOERROR )
	{
		Log ( J2534_FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "%s returned %ld",
		      "PassThruIoctl(CLEAR_RX_BUFFER)",
		      RetVal );
		return FAIL;
	}

	// Send the request
	NumMsgs = 1;
	RetVal  = PassThruWriteMsgs ( gOBDList[gOBDListIndex].ChannelID, &TxMsg, &NumMsgs, 500 );

	if ( RetVal != STATUS_NOERROR )
	{
		//  don't log timeouts during DetermineProtocol
		if ( !(gDetermineProtocol == TRUE && RetVal == ERR_TIMEOUT) )
		{
			Log ( J2534_FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "%s returned %ld",
			      "PassThruWriteMsgs",
			      RetVal );
			return FAIL;
		}
	}

	// Log the request message to compare to what is sent
	LogMsg ( &TxMsg, LOG_REQ_MSG );

	// Reset the response data buffers
	if ( SidResetResponseData ( &TxMsg ) != PASS )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Cannot reset SID $%02 response data\n",
		      TxMsg.Data[gOBDList[gOBDListIndex].HeaderSize] );
		return FAIL;
	}

	// Read the response(s) with a timeout of twice what is allowed
	// so we can see late responses.
	NumResponses    = 0;
	NumFirstFrames  = 0;
	fFirstResponse  = TRUE;
	StartTimeMsecs  = GetTickCount ();

	do
	{
		if ( fFirstResponse == TRUE )
		{
			fFirstResponse = FALSE;
			ulResponseTimeoutMsecs =  5 * gOBDMaxResponseTimeMsecs;
			sprintf_s ( bString, MAX_LOG_STRING_SIZE, "In SidRequest - Initial PassThruReadMsgs" );
		}
		else
		{
			ulResponseTimeoutMsecs =  (5 * gOBDMaxResponseTimeMsecs) + ExtendResponseTimeMsecs;
			sprintf_s ( bString, MAX_LOG_STRING_SIZE, "In SidRequest - Loop PassThruReadMsgs" );

		}

		// Read the next response
		NumMsgs = 1;
		RetVal = PassThruReadMsgs ( gOBDList[gOBDListIndex].ChannelID,
		                            &RxMsg,
		                            &NumMsgs,
		                            ulResponseTimeoutMsecs );

		if ( RetVal != STATUS_NOERROR &&
		     RetVal != ERR_BUFFER_EMPTY &&
		     RetVal != ERR_NO_FLOW_CONTROL )
		{
			// Log undesirable returns
			Log ( J2534_FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "%s returned %ld",
			      bString,
			      RetVal );
			eReturnCode |= FAIL;
		}

		// If a message was received, process it
		if ( NumMsgs == 1 )
		{
			// Save all read messages in the log file
			LogMsg ( &RxMsg, LOG_NORMAL_MSG );

			// Process response based on protocol
			switch ( gOBDList[gOBDListIndex].Protocol )
			{
				case ISO15765:
				{
					eReturnCode |= ( ProcessISO15765Msg ( SidReq,
					                                      &RxMsg,
					                                      &StartTimeMsecs,
					                                      &NumResponses,
					                                      &NumFirstFrames,
					                                      &TxTimestamp,
					                                      &ExtendResponseTimeMsecs) );
				}
				break;
				default:
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "Invalid protocol specified for response.\n" );
					return FAIL;
				}
			}
		}

		// If all expected ECUs responded and flag is set, don't wait for timeout
		// NOTE: This mechanism is only good for single message response per ECU
		if ( NumResponses >= gOBDNumEcus &&
		     (Flags & SID_REQ_RETURN_AFTER_ALL_RESPONSES) )
		{
			break;
		}
	}
	while ( NumMsgs == 1 &&
	        ( GetTickCount () - StartTimeMsecs ) < ( ( 5 * gOBDMaxResponseTimeMsecs ) + ExtendResponseTimeMsecs ) );

	// Restart the periodic message if protocol determined and not in burst test
	if ( gOBDDetermined == TRUE &&
	     (Flags & SID_REQ_NO_PERIODIC_DISABLE) == 0 )
	{
		if ( gOBDList[gOBDListIndex].TesterPresentID == -1 )
		{
			if ( StartPeriodicMsg () != PASS )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "Problems starting periodic messages.\n" );
				eReturnCode |= FAIL;
			}
		}
	}

	// Preserve the total number of ECUs responding and
	// allow calling routine to access information.
	gOBDNumEcusResp = NumResponses;

	// Return code based on whether this protocol supports OBD
	if ( NumResponses > 0 )
	{
		if ( gOBDDetermined == FALSE &&
		     (eReturnCode & RETRY) != RETRY )
		{
			gOBDNumEcus = NumResponses;
			Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "%d OBD ECU(s) found\n",
			      gOBDNumEcus);

//			// IF this is the first ECU check save the number of responses
//			if ( gTestPhase == eTestNoDTC && gTestSubsection == 2 )
//			{
//				gUserNumEcus = gOBDNumEcus;
//			}
			// Check if number of OBD-II ECUs entered by user matches the number detected
			if ( gOBDNumEcus != gUserNumEcus )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "Number of OBD-II ECUs currently detected (%d) does not match the number previously entered (%d).\n",
				      gOBDNumEcus,
				      gUserNumEcus );
				eReturnCode |= FAIL;
			}

		}

		// If there weren't any errors since sending the request, pass
		if ( (eReturnCode & FAIL) != FAIL )
		{
			if ( (eReturnCode & RETRY) == RETRY )
			{
				return RETRY;
			}
			else if ( (eReturnCode & ERRORS) == ERRORS )
			{
				return ERRORS;
			}
			else
			{
				return PASS;
			}
		}
	}

	// If there was at least one failure since sending the request, return FAIL
	if ( (eReturnCode & FAIL) == FAIL )
	{
		return FAIL;
	}

	if ( Flags & SID_REQ_ALLOW_NO_RESPONSE )
	{
		// calling function must determine any actual responses
		return PASS;
	}

	if ( gOBDDetermined == TRUE &&
	     (Flags & SID_REQ_IGNORE_NO_RESPONSE) != SID_REQ_IGNORE_NO_RESPONSE )
	{
		Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "No response to OBD request\n" );
	}
	return FAIL;
}




/*
********************************************************************************
**
**  Function:  ProcessISO15765Msg
**
**  Purpose:   Purpose is to process response based on connection to
**             ISO15765.
**
**  Returns:   RETRY, ERRORS, FAIL, or PASS
**
********************************************************************************
*/
STATUS ProcessISO15765Msg ( SID_REQ       *pSidReq,
                            PASSTHRU_MSG  *pRxMsg,
                            unsigned long *ulStartTimeMsecs,
                            unsigned long *pulNumResponses,
                            unsigned long *ulNumFirstFrames,
                            unsigned long *ulTxTimestamp,
                            unsigned long *ulExtendResponseTimeMsecs )
{
	unsigned long ulResponseTimeMsecs;
	unsigned long EcuTimingIndex;
	unsigned long EcuIndex;
	unsigned long EcuId = 0;
	unsigned short i;
	STATUS eReturnCode = PASS;    // saves the return code from function calls
	BOOL RequestMsgLpbk = FALSE;

	// create the ECU ID for later use
	for ( i = 0; i < 4; i++ )
	{
		EcuId = (EcuId << 8) + (pRxMsg->Data[i]);
	}

	// if this is an echo of the request, don't track it's timing data
	if ( pRxMsg->RxStatus & TX_MSG_TYPE )
	{
		RequestMsgLpbk = TRUE;
	}
	else
	{
		// Find this ECU's Timing Structure
		for ( EcuTimingIndex = 0; EcuTimingIndex < gNumEcusForAlloc; EcuTimingIndex++ )
		{
			if ( gEcuTimingData[EcuTimingIndex].EcuId == EcuId || gEcuTimingData[EcuTimingIndex].EcuId == 0x00)
			{
				if ( gEcuTimingData[EcuTimingIndex].EcuId == 0x00 )
				{
					gEcuTimingData[EcuTimingIndex].EcuId = EcuId;
				}
				break;
			}
		}
	}

	// FAIL for padding error only once
	if ( bPadErrorPermanent == FALSE )
	{
		if ( pRxMsg->RxStatus & ISO15765_PADDING_ERROR )
		{
			bPadErrorPermanent = TRUE;

			Log ( ERROR_FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  ISO15765 message padding error.\n",
			      EcuId );
			eReturnCode = ERRORS;
		}
	}

	ulResponseTimeMsecs = (pRxMsg->Timestamp - *ulTxTimestamp) / 1000;

	// Check for FirstFrame indication
	if ( pRxMsg->RxStatus & ISO15765_FIRST_FRAME )
	{
		// Check if response was late (all first frame indications due in P2_MAX)
		if ( (pRxMsg->Timestamp - *ulTxTimestamp) > (gOBDMaxResponseTimeMsecs * 1000) )
		{
			// Exceeded maximum response time
			Log ( ERROR_FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  OBD First Frame Indication was later than allowed (> %dmsec)\n",
			      EcuId,
			      gOBDMaxResponseTimeMsecs );
			eReturnCode = ERRORS;

			if ( !RequestMsgLpbk )
			{
				gEcuTimingData[EcuTimingIndex].RespTimeOutofRange++;
				gEcuTimingData[EcuTimingIndex].RespTimeTooLate++;

				// Tally up the response statistics
				gEcuTimingData[EcuTimingIndex].AggregateResponseTimeMsecs += ulResponseTimeMsecs;
				gEcuTimingData[EcuTimingIndex].AggregateResponses += ( ( ( pRxMsg->DataSize - 4 ) / ISO15765_MAX_BYTES_PER_FRAME ) + 1 );
				if ( ulResponseTimeMsecs > gEcuTimingData[EcuTimingIndex].LongestResponsesTime )
				{
					gEcuTimingData[EcuTimingIndex].LongestResponsesTime = ulResponseTimeMsecs;
				}
			}
			Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Calculated OBD Response time: %2.4f msec\n",
			      (float)(pRxMsg->Timestamp - *ulTxTimestamp) / 1000 );
		}

		// Extend the response time to the worst case for segmented responses
		gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs = 30000;
		if ( gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs > *ulExtendResponseTimeMsecs )
		{
			*ulExtendResponseTimeMsecs = gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs;
		}
		(*ulNumFirstFrames)++;
		gEcuTimingData[EcuTimingIndex].FFReceived = TRUE;
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Receiving segmented responses, please wait...\n" );
	}

	// Check for echoed request message
	else if ( pRxMsg->RxStatus & TX_MSG_TYPE )
	{
		if ( ( pRxMsg->DataSize >= 5 ) &&
		     ( pRxMsg->Data[4]  == pSidReq->SID ) )
		{
			// Save the timestamp
			*ulTxTimestamp = pRxMsg->Timestamp;
		}
	}

	// Check for NAK response message
	else if ( pRxMsg->DataSize >= 7    &&
	          pRxMsg->Data[4]  == NAK  &&
	          pRxMsg->Data[5]  == pSidReq->SID )
	{
		if ( gEcuTimingData[EcuTimingIndex].NAKReceived == FALSE )
		{
			// Tally up the response statistics
			gEcuTimingData[EcuTimingIndex].AggregateResponseTimeMsecs += ulResponseTimeMsecs;
			gEcuTimingData[EcuTimingIndex].AggregateResponses += ( ( ( pRxMsg->DataSize - 4 ) / ISO15765_MAX_BYTES_PER_FRAME ) + 1 );
			if ( ulResponseTimeMsecs > gEcuTimingData[EcuTimingIndex].LongestResponsesTime )
			{
				gEcuTimingData[EcuTimingIndex].LongestResponsesTime = ulResponseTimeMsecs;
			}

			gEcuTimingData[EcuTimingIndex].NAKReceived = TRUE;
		}

		// Check if response was late (all negative respones due in P2_MAX)
		if ( (pRxMsg->Timestamp - *ulTxTimestamp) > ((gOBDMaxResponseTimeMsecs +  gEcuTimingData[EcuTimingIndex].ResponsePendingDelay) * 1000) )
		{
			// Exceeded maximum response time
			Log ( ERROR_FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  OBD Negative Response was later than allowed (> %dmsec)\n",
			      EcuId,
			      gOBDMaxResponseTimeMsecs + gEcuTimingData[EcuTimingIndex].ResponsePendingDelay );
			eReturnCode = ERRORS;

			gEcuTimingData[EcuTimingIndex].RespTimeOutofRange++;
			gEcuTimingData[EcuTimingIndex].RespTimeTooLate++;
			Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Calculated OBD Response time: %2.4f msec\n",
			      (float)(pRxMsg->Timestamp - *ulTxTimestamp) / 1000 );
		}

		// Save the response information
		if ( SidSaveResponseData ( pRxMsg, pSidReq, pulNumResponses ) != PASS )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Cannot save SID response data\n" );
			return FAIL;
		}

		// check the kind of response received for the vehicle
		switch ( pRxMsg->Data[6] )
		{
			case NAK_GENERAL_REJECT: // 0x10
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  General reject to Service $%02X.\n",
				      EcuId,
				      pRxMsg->Data[5] );
			}
			break;

			case NAK_NOT_SUPPORTED: // 0x11
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Service $%02X not supported.\n",
				      EcuId,
				      pRxMsg->Data[5] );
			}
			break;

			case NAK_INVALID_FORMAT: // 0x12
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Service $%02X supported. Unsupported PID request.\n",
				      EcuId,
				      pRxMsg->Data[5] );
			}
			break;

			case NAK_REPEAT_REQUEST: // 0x21
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Service $%02X supported. Busy.\n",
				      EcuId,
				      pRxMsg->Data[5] );
			}
			break;

			case NAK_SEQUENCE_ERROR: // 0x22
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Service $%02X supported. Conditions not correct.\n",
				      EcuId,
				      pRxMsg->Data[5] );
			}
			break;

			case NAK_RESPONSE_PENDING: // 0x78
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Service $%02X supported. Response pending.\n",
				      EcuId,
				      pRxMsg->Data[5] );
			}
			break;

			default:
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Unknown negative response to Service $%02X.\n",
				      EcuId,
				      pRxMsg->Data[5] );
			}
			break;
		}

		// NAK Truth table
		switch ( pSidReq->SID )
		{
			case 0x22:
			{
				if ( gDetermineProtocol == TRUE &&
					pSidReq->DID[0] == 0xF400 &&
					pRxMsg->Data[6] == NAK_REPEAT_REQUEST )
				{
					return RETRY;
				}
				else
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  SID $22 NAK response error!\n",
					      EcuId );
					return FAIL;
				}
			}
			break;

			case 0x04:
			{
				if ( pRxMsg->Data[6] == NAK_RESPONSE_PENDING )
				{
					// If response pending, extend the wait time
					gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs = 30000;
					if ( gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs > *ulExtendResponseTimeMsecs )
					{
						*ulExtendResponseTimeMsecs = gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs;
					}
					gEcuTimingData[EcuTimingIndex].ResponsePendingDelay = 5000;
					if ( gEcuTimingData[EcuTimingIndex].ResponsePendingDelay > ulResponsePendingDelay )
					{
						ulResponsePendingDelay = gEcuTimingData[EcuTimingIndex].ResponsePendingDelay;
					}

					// set wait flag
					if ( LookupEcuIndex ( pRxMsg, &EcuIndex ) == PASS )
					{
						ulEcuWaitFlags |= (1<<EcuIndex);
						gOBDResponse[EcuIndex].bResponseReceived = FALSE;  // allow ECU to respond again
					}
					else
					{
						return FAIL;
					}
				}
				else if ( pRxMsg->Data[6] != NAK_SEQUENCE_ERROR )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  SID $14 NAK response error!\n",
					      EcuId );
					return FAIL;
				}
			}
			break;

			case 0x08:
			{
				if ( pRxMsg->Data[6] == NAK_NOT_SUPPORTED )
				{
					break;
				}
				else
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  SID $08 NAK response error!\n",
					      EcuId,
					      pSidReq->SID );
				}
			}
			break;

			case 0x09:
			{
//				// IF EOBD/IOBD/OBDBr and CVN
//				if ( pSidReq->Ids[0] == 0x06 &&
//				     ( gUserInput.eComplianceType != US_OBDII &&
//				       gUserInput.eComplianceType != HD_OBD ) )
//				{
//					break;
//				}
			}
			break;

			default :
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  SID $%02X NAK response error!\n",
				      EcuId,
				      pSidReq->SID );
				return FAIL;
			}
			break;
		}
	}

	// Check for response message
	else if ( pRxMsg->DataSize >= 5 )
	{
		if ( gEcuTimingData[EcuTimingIndex].NAKReceived == FALSE &&
		     gEcuTimingData[EcuTimingIndex].FFReceived  == FALSE )
		{
			// Tally up the response statistics
			gEcuTimingData[EcuTimingIndex].AggregateResponseTimeMsecs += ulResponseTimeMsecs;
			gEcuTimingData[EcuTimingIndex].AggregateResponses += ( ( ( pRxMsg->DataSize - 4 ) / ISO15765_MAX_BYTES_PER_FRAME ) + 1 );
			if ( ulResponseTimeMsecs > gEcuTimingData[EcuTimingIndex].LongestResponsesTime )
			{
				gEcuTimingData[EcuTimingIndex].LongestResponsesTime = ulResponseTimeMsecs;
			}
		}
		else
		{
			gEcuTimingData[EcuTimingIndex].NAKReceived = FALSE;
		}

		// Check if response was late (compensate for segmented responses)
		if ( (pRxMsg->Timestamp - *ulTxTimestamp) >
		     ((gOBDMaxResponseTimeMsecs + gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs) * 1000) )
		{
			// Exceeded maximum response time
			Log ( ERROR_FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  OBD Response was later than allowed (> %dmsec)\n",
			      EcuId,
			      (gOBDMaxResponseTimeMsecs + gEcuTimingData[EcuTimingIndex].ExtendResponseTimeMsecs) );
			eReturnCode = ERRORS;

			gEcuTimingData[EcuTimingIndex].RespTimeOutofRange++;
			gEcuTimingData[EcuTimingIndex].RespTimeTooLate++;
			Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Calculated OBD Response time: %2.4f msec\n",
			      (float)(pRxMsg->Timestamp - *ulTxTimestamp) / 1000 );
		}

		// clear wait flag
		if ( LookupEcuIndex ( pRxMsg, &EcuIndex ) == PASS )
		{
			ulEcuWaitFlags &= ~(1<<EcuIndex);
			gEcuTimingData[EcuTimingIndex].ResponsePendingDelay = 0;        // ECU response no longer pending
		}
		else
		{
			return FAIL;
		}

		// Check if all ECUs with response pending have responded
		if ( ulEcuWaitFlags == 0 )
		{
			ulResponsePendingDelay = 0;         // no ECU response is pending

			if ( *ulNumFirstFrames == 0 )
			{
				*ulExtendResponseTimeMsecs = 0;  // no First Frames, clear extended response time
			}
		}

		// Check if we can turn off the time extension for segmented frames
		if ( pRxMsg->DataSize > ( 4 + ISO15765_MAX_BYTES_PER_FRAME ) )
		{
			if ( *ulNumFirstFrames > 0 )
			{
				// If we have received as many segmented frames as FirstFrame indications,
				// restore 'response pending' time extension
				if ( --(*ulNumFirstFrames) == 0 )
				{
					*ulExtendResponseTimeMsecs = ulResponsePendingDelay;
				}
			}
		}

		// Check for proper SID response
		if ( pRxMsg->Data[4] != ( pSidReq->SID + OBD_RESPONSE_BIT ) )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Invalid SID response to request\n",
			      EcuId );
			return FAIL;
		}

		// If a PID was requested
		if ( pSidReq->NumIds != 0 )
		{
			// Check for proper PID/MID/TID/INF response
			for ( i = 0; i < pSidReq->NumIds; i++ )
			{
				if ( pRxMsg->Data[6] == (pSidReq->DID[i] & 0x00FF) )
				{
					break;
				}
				// if this is a SID $14 request, no DID in response
				if ( pSidReq->SID == 0x14 )
				{
					i++;
				}
			}
			if ( i == pSidReq->NumIds )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Invalid PID response to request\n",
				      EcuId );
				return FAIL;
			}
		}

		// Save the response information
		if ( SidSaveResponseData ( pRxMsg, pSidReq, pulNumResponses ) != PASS )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Cannot save SID response data\n" );
			return FAIL;
		}
	}

	// Ignore, invalid message

	return eReturnCode;
}


/*
*****************************************************************************
**
**  Function:  SetupRequestMSG
**
**  Purpose:   Purpose is to setup request message based on the protocol.
**             Routine will normalize the message data handling between
**             the regulated OBD protocol message structures.
**
*******************************************************************************
*/
STATUS SetupRequestMSG ( SID_REQ *pSidReq, PASSTHRU_MSG *pTxMsg )
{
	unsigned int TxIndex;
	unsigned int IdIndex;

	// Set message data information common to all protocols
	pTxMsg->ProtocolID  = gOBDList[gOBDListIndex].Protocol;
	pTxMsg->RxStatus    = TX_MSG_TYPE;
	pTxMsg->TxFlags     = 0x00;  // Change back to non-blocking

	// Configure protocol specific message data.
	switch ( gOBDList[gOBDListIndex].Protocol )
	{
		case ISO15765:
		{
			pTxMsg->TxFlags |= ISO15765_FRAME_PAD;

			if ( gOBDList[gOBDListIndex].InitFlags & CAN_29BIT_ID )
			{
				pTxMsg->TxFlags |= CAN_29BIT_ID;

				pTxMsg->Data[0]  = 0x18;
				pTxMsg->Data[1]  = 0xDB;
				pTxMsg->Data[2]  = 0x33;
				pTxMsg->Data[3]  = TESTER_NODE_ADDRESS;
			}
			else
			{
				pTxMsg->Data[0]  = 0x00;
				pTxMsg->Data[1]  = 0x00;
				pTxMsg->Data[2]  = 0x07;
				pTxMsg->Data[3]  = 0xDF;
			}

			pTxMsg->Data[4] = pSidReq->SID;

			// if this is check for J1979 classic support (SID $1-$A)
			if ( pSidReq->SID <= 0x0A )
			{
				if ( pSidReq->SID != 0x0A )
				{
					pTxMsg->Data[5] = 0x00;
					pTxMsg->DataSize = 6;
				}
				else
				{
					pTxMsg->DataSize = 5;
				}
			}
			// if this is a Clear DTC request
			else if ( pTxMsg->Data[4] == 0x14 )
			{
				pTxMsg->Data[5] = (unsigned char)((pSidReq->DID[0] & 0xFF0000) >> 16);
				pTxMsg->Data[6] = (unsigned char)((pSidReq->DID[0] & 0x00FF00) >> 8);
				pTxMsg->Data[7] = (unsigned char)(pSidReq->DID[0]  & 0x0000FF);

				pTxMsg->DataSize = 5 + 3;
			}
			// if this is a  DTC request
			else if ( pTxMsg->Data[4] == 0x19 )
			{
				for ( TxIndex = 5, IdIndex = 0; IdIndex < pSidReq->NumIds; TxIndex++, IdIndex++ )
				{
					pTxMsg->Data[TxIndex] = (unsigned char)pSidReq->DID[IdIndex];
				}

				pTxMsg->DataSize = 5 + pSidReq->NumIds;
			}
			else
			{
				for ( TxIndex = 5, IdIndex = 0; IdIndex < pSidReq->NumIds; IdIndex++ )
				{
					pTxMsg->Data[TxIndex++] = (unsigned char)((pSidReq->DID[IdIndex] & 0xFF00) >> 8);
					pTxMsg->Data[TxIndex++] = (unsigned char)(pSidReq->DID[IdIndex]  & 0x00FF);
				}
				pTxMsg->DataSize = 5 + ((pSidReq->NumIds)*2);
			}
		}
		break;
		default:
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Invalid protocol specified in SetupRequestMSG\n" );
			return FAIL;
		}
	}

	return PASS;
}
