/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


STATUS LookupEcuIndex ( PASSTHRU_MSG *RxMsg, unsigned long *pulEcuIndex );
void ChkIFRAdjust ( PASSTHRU_MSG *RxMsg );  // Logic ajustment for PID Group Reverse order request.
STATUS IsMessageUnique ( unsigned char *pucBuffer,  // pointer to buffer
                         unsigned short usSize,     // size of buffer
                         unsigned char *pucMsg,     // pointer to message to be checked
                         unsigned short usMsgSize   // size of message
                       );

extern STATUS VerifyEcuID ( unsigned char EcuId[] );


/*
*******************************************************************************
**  SidSaveResponseData -
**  Function to save SID response data
*******************************************************************************
*/
STATUS SidSaveResponseData ( PASSTHRU_MSG *RxMsg, SID_REQ *SidReq, unsigned long *pulNumResponses )
{
	unsigned long HeaderSize;
	unsigned long EcuIndex;
	unsigned char bElementOffset;
	unsigned long ulInx;  // coordinate mode 6 data respones for multiple data response.
	unsigned int  ExpectedDataSize = 0;

	BOOL          CheckDataSize = TRUE;
	BOOL          bPendingDTC = FALSE;
	BOOL          bConfirmedDTC = FALSE;
	DTC_HEADER   *pDTCHeader = NULL;
	DTC          *pDTC = NULL;


	// Set the response header size based on the protocol
	HeaderSize = gOBDList[gOBDListIndex].HeaderSize;

	// Get index into gOBDResponse struct
	if ( LookupEcuIndex ( RxMsg, &EcuIndex ) != PASS )
	{
		return FAIL;
	}

	ChkIFRAdjust ( RxMsg );

	if ( gOBDResponse[EcuIndex].bResponseReceived == TRUE )
	{
		Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  has already responded\n",
		      GetEcuId ( EcuIndex ) );
	}
	else
	{
		(*pulNumResponses)++;  // count the new response
	}

	// Save the data in the appropriate SID/PID/MID/TID/InfoType
	switch ( RxMsg->Data[HeaderSize] )
	{
		// OBD Services
		// Reserved Services
		case 0x40:
		case 0x4B:
		case 0x4C:
		case 0x4D:
		case 0x4E:
		case 0x4F:
		{
			// indicate that this ECU has responded
			gOBDResponse[EcuIndex].bResponseReceived = TRUE;

			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Response to Reserved Service $X.\n",
			      RxMsg->Data[HeaderSize] );
		}
		break;

		// Deprecated Services
		case 0x41:
		case 0x42:
		case 0x43:
		case 0x44:
		case 0x45:
		case 0x46:
		case 0x47:
		case 0x48:
		case 0x49:
		case 0x4A:
		{
			// indicate that this ECU has responded
			gOBDResponse[EcuIndex].bResponseReceived = TRUE;

			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Response to J1979 Classic Service $X.\n",
			      RxMsg->Data[HeaderSize] );
		}
		break;
		
		// SID 0x14  Clear DTCs
		case 0x54:
		{
			// indicate that this ECU has responded
			gOBDResponse[EcuIndex].bResponseReceived = TRUE;

			// Indicate positive response
			gOBDResponse[EcuIndex].ClrDTCData[0] = 0x54;
			gOBDResponse[EcuIndex].ClrDTCSize = 1;
		}
		break;

		// SID 19  Read DTC Information
		case 0x59:
		{
			// Process message based on subfunction
			switch ( RxMsg->Data[HeaderSize + 1] )
			{
				// Freeze Frame (Snapshot) Identification
				case 0x03:
				{
					// indicate that this ECU has responded
					gOBDResponse[EcuIndex].bResponseReceived = TRUE;

					// Determine if there is enough room in the buffer to store the data
					if ( (RxMsg->DataSize - HeaderSize - 1) > sizeof ( gOBDResponse[EcuIndex].FF ) )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $19 Subfunction $03 response data exceeded buffer size\n",
						      GetEcuId ( EcuIndex ) );
						return FAIL;
					}

					// check for Freeze Frame data
					if ( RxMsg->DataSize <= (HeaderSize + 5) )
					{
						// if no Freeze Frames, set to zero
						memset ( gOBDResponse[EcuIndex].FF, 0, 5 );
						gOBDResponse[EcuIndex].FFSize = 5;
					}
					else
					{
						// Otherwise copy the Freeze Frame data
						memcpy ( gOBDResponse[EcuIndex].FF,
						         &RxMsg->Data[HeaderSize + 2],
						         (RxMsg->DataSize - HeaderSize - 2) );
						gOBDResponse[EcuIndex].FFSize = (unsigned short)(RxMsg->DataSize - HeaderSize - 2);
					}
				}  // end case 0x03 freeze frame identification
				break;

				// Freeze Frame (Snapshot) Data
				case 0x04:
				{

				}
				break;

				// Confirmed or Pending DTCs
				case 0x42:
				{
					if ( (RxMsg->Data[HeaderSize + 3] & 0x08) != 0 )
					{
						bConfirmedDTC = TRUE;
						pDTCHeader = &gOBDResponse[EcuIndex].ConfDTCHeader;
						pDTC = (DTC*)&gOBDResponse[EcuIndex].ConfDTC[0];
					}
					else if ( (RxMsg->Data[HeaderSize + 3] & 0x04) != 0 )
					{
						bPendingDTC = TRUE;
						pDTCHeader = &gOBDResponse[EcuIndex].PendDTCHeader;
						pDTC = (DTC*)&gOBDResponse[EcuIndex].PendDTC[0];
					}
					pDTCHeader->DTCSupported = TRUE;

					// indicate that this ECU has responded
					gOBDResponse[EcuIndex].bResponseReceived = TRUE;

					// Determine if there is enough room in the buffer to store the data
					if ( (RxMsg->DataSize - HeaderSize - 1) > sizeof ( gOBDResponse[EcuIndex].PendDTC ) )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $19 Subfunction $42 response data exceeded buffer size\n",
						      GetEcuId ( EcuIndex ) );
						return FAIL;
					}

					// Save the DTCStatusAvailabiltyMask
					pDTCHeader->DTCStatus = RxMsg->Data[HeaderSize + 3];

					// Save the DTCSeverityMask
					pDTCHeader->DTCSeverity = RxMsg->Data[HeaderSize + 4];

					// Save the DTCFormatIdentifier
					pDTCHeader->DTCFormat = RxMsg->Data[HeaderSize + 5];

					// check for DTC data
					if ( RxMsg->DataSize <= (HeaderSize + 9) )
					{
						memset ( pDTC, 0, 5 );
						pDTCHeader->DTCSize = 5;
					}
					else
					{
						// Otherwise copy the DTC data
						memcpy ( pDTC,
						         &RxMsg->Data[HeaderSize + 6],
						         (RxMsg->DataSize - HeaderSize - 6) );
						pDTCHeader->DTCSize = (unsigned short)(RxMsg->DataSize - HeaderSize - 6);
					}
				}  //end case 0x42 Pending or Confirmed DTCs
				break;

				// Permanent DTCs
				case 0x55:
				{
					gOBDResponse[EcuIndex].PermDTCSupported = TRUE;

					// indicate that this ECU has responded
					gOBDResponse[EcuIndex].bResponseReceived = TRUE;

					// Determine if there is enough room in the buffer to store the data
					if ( (RxMsg->DataSize - HeaderSize - 1) > sizeof ( gOBDResponse[EcuIndex].PermDTC ) )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $19 Subfunction $55 response data exceeded buffer size\n",
						      GetEcuId ( EcuIndex ) );
						return FAIL;
					}

					// Save the DTCStatusAvailabiltyMask
					gOBDResponse[EcuIndex].PermDTCStatus = RxMsg->Data[HeaderSize + 3];

					// Save the DTCFormatIdentifier
					gOBDResponse[EcuIndex].PermDTCFormat = RxMsg->Data[HeaderSize + 4];

					// check for DTC data
					if ( RxMsg->DataSize <= (HeaderSize + 4) )
					{
						// If no DTCs, set to zero
						memset ( &gOBDResponse[EcuIndex].PermDTC[0], 0, 4 );
						gOBDResponse[EcuIndex].PermDTCSize = 4;
					}
					else
					{
						// Otherwise copy the DTC data
						memcpy ( &gOBDResponse[EcuIndex].PermDTC[0],
						         &RxMsg->Data[HeaderSize + 5],
						         (RxMsg->DataSize - HeaderSize - 5) );
						gOBDResponse[EcuIndex].PermDTCSize = (unsigned short)(RxMsg->DataSize - HeaderSize - 4);
					}
				}  //end case 0x55 Permanent DTCs
			}  // end switch ( RxMsg->Data[HeaderSize + 1] )
		}
		break;

		// SID 0x22
		case 0x62:
		{
			// indicate that this ECU has responded
			gOBDResponse[EcuIndex].bResponseReceived = TRUE;

			// Process message based on DID number MSB
			switch ( RxMsg->Data[HeaderSize + 1] )
			{
				// PID
				case 0xF4:
				{
					// Process message based on PID number (DID LSB)
					switch ( RxMsg->Data[HeaderSize + 2] )
					{
						case 0x00:
						{
							if ( gVerifyLink == TRUE ||
							     gDetermineProtocol == TRUE )
							{
								gOBDResponse[EcuIndex].LinkActive = TRUE;

								if ( gOBDResponse[EcuIndex].PidSupportSize != 0 )
								{
									break;
								}
							}
						}
						// 0x00 falls through

						case 0x20:
						case 0x40:
						case 0x60:
						case 0x80:
						case 0xA0:
						case 0xC0:
						case 0xE0:
						{
							// Make sure there is enough data in the message to process
							if ( RxMsg->DataSize < (HeaderSize + 1 + sizeof ( ID_SUPPORT )) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Not enough data in PID $%02X%02X support response message to process\n",
								      GetEcuId ( EcuIndex ),
								      RxMsg->Data[HeaderSize + 1],
								      RxMsg->Data[HeaderSize + 2] );
								return FAIL;
							}

							// Save the data in the buffer and set the new size
							for ( ulInx = HeaderSize + 2; ulInx < RxMsg->DataSize; ulInx += (sizeof ( ID_SUPPORT ) + 1) )
							{
								// find the element number (for the array of structures)
								bElementOffset = (unsigned char)(RxMsg->Data[ulInx]) >> 5;

								// if this ID is $F5x0 (from a group PID support PID response which started with a $F4x0 PID)
								if ( RxMsg->Data[ulInx-1] == 0xF5 )
								{
									// add the offset to the $F5x0 segment of the PidSupport array
									bElementOffset += 8;
								}

								// move the data into the element
								memcpy ( &gOBDResponse[EcuIndex].PidSupport[bElementOffset], &RxMsg->Data[ulInx], sizeof ( ID_SUPPORT ) );

								// increase the element count if the element in the array was previously unused
								if ( (bElementOffset + 1) > gOBDResponse[EcuIndex].PidSupportSize )
								{
									gOBDResponse[EcuIndex].PidSupportSize = (bElementOffset + 1);
								}
							}
						}

						/*
						 * FALL THROUGH AND TREAT LIKE OTHER PIDS
						 */

						default:
						{
							// If not PID 0, then check if this is a response to an unsupported PID
							if ( gIgnoreUnsupported == FALSE &&
							     RxMsg->Data[HeaderSize + 2] != 0x00 &&
							     (gOBDResponse[EcuIndex].PidSupport[
							       (RxMsg->Data[HeaderSize + 2] - 1) >> 5].IDBits[
							       ((RxMsg->Data[HeaderSize + 2] - 1) >> 3) & 0x03]
							       & (0x80 >> ((RxMsg->Data[HeaderSize + 2] - 1) & 0x07))) == 0 )
							{
								Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Unsupported PID $%02X%02X detected\n",
								      GetEcuId ( EcuIndex ),
								      RxMsg->Data[HeaderSize + 1],
								      RxMsg->Data[HeaderSize + 2] );
							}

							// Make sure there is enough data in the message to process
							if ( RxMsg->DataSize <= (HeaderSize + 2) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Not enough data in PID $%02X%02X response message to process\n",
								      GetEcuId ( EcuIndex ),
								      RxMsg->Data[HeaderSize + 1],
								      RxMsg->Data[HeaderSize + 2] );
								return FAIL;
							}

							// If only one PID requested,
							// make sure the PID has the correct amount of data
							if ( SidReq->NumIds == 1 )
							{
								CheckDataSize = TRUE;

								switch ( RxMsg->Data[HeaderSize + 2] )
								{
									// Single byte PIDs
									case 0x04:
									case 0x05:
									case 0x0A:
									case 0x0B:
									case 0x0D:
									case 0x0E:
									case 0x0F:
									case 0x11:
									case 0x12:
									case 0x13:
									case 0x1C:
									case 0x1D:
									case 0x1E:
									case 0x2C:
									case 0x2D:
									case 0x2E:
									case 0x2F:
									case 0x30:
									case 0x33:
									case 0x45:
									case 0x46:
									case 0x47:
									case 0x48:
									case 0x49:
									case 0x4A:
									case 0x4B:
									case 0x4C:
									case 0x51:
									case 0x52:
									case 0x5A:
									case 0x5B:
									case 0x5C:
									case 0x5F:
									case 0x61:
									case 0x62:
									case 0x7D:
									case 0x7E:
									case 0x84:
									case 0x8D:
									case 0x8E:
									{
										ExpectedDataSize = 1;
									}
									break;

									// Two byte PIDs
									case 0x02:
									case 0x03:
									case 0x0C:
									case 0x10:
									case 0x14:
									case 0x15:
									case 0x16:
									case 0x17:
									case 0x18:
									case 0x19:
									case 0x1A:
									case 0x1B:
									case 0x1F:
									case 0x21:
									case 0x22:
									case 0x23:
									case 0x31:
									case 0x32:
									case 0x3C:
									case 0x3D:
									case 0x3E:
									case 0x3F:
									case 0x42:
									case 0x43:
									case 0x44:
									case 0x4D:
									case 0x4E:
									case 0x53:
									case 0x54:
									case 0x59:
									case 0x5D:
									case 0x5E:
									case 0x63:
									case 0x65:
									case 0x92:
									case 0x9E:
									case 0xA2:
									{
										ExpectedDataSize = 2;
									}
									break;

									// Three byte PIDs
									case 0x67:
									case 0x6F:
									case 0x90:
									case 0x93:
									{
										ExpectedDataSize = 3;
									}
									break;

									// Four byte PIDs
									case 0x00:
									case 0x20:
									case 0x40:
									case 0x60:
									case 0x80:
									case 0xA0:
									case 0xC0:
									case 0xE0:
									case 0x01:
									case 0x24:
									case 0x25:
									case 0x26:
									case 0x27:
									case 0x28:
									case 0x29:
									case 0x2A:
									case 0x2B:
									case 0x34:
									case 0x35:
									case 0x36:
									case 0x37:
									case 0x38:
									case 0x39:
									case 0x3A:
									case 0x3B:
									case 0x41:
									case 0x4F:
									case 0x50:
									case 0x9B:
									case 0x9D:
									case 0xA4:
									case 0xA5:
									case 0xA6:
									{
										ExpectedDataSize = 4;

										// IF PID $01, save appropriate size data to allow the other ECU's to be saved and evaluated
										if ( (RxMsg->DataSize - (HeaderSize + 3)) != 4 &&
										     RxMsg->Data[HeaderSize + 2] == 0x01 )
										{
											if ( (RxMsg->DataSize - (HeaderSize + 3)) < 4 )
											{
												memset ( &RxMsg->Data[RxMsg->DataSize],
												         0x00,
												         4 - (RxMsg->DataSize - (HeaderSize + 3)) );
											}
											RxMsg->DataSize = 4 + (HeaderSize + 2);
										}
									}
									break;

									// Five byte PIDs
									case 0x64:
									case 0x66:
									case 0x6A:
									case 0x6B:
									case 0x6C:
									case 0x72:
									case 0x73:
									case 0x74:
									case 0x77:
									case 0x86:
									case 0x87:
									case 0x91:
									{
										ExpectedDataSize = 5;
									}
									break;

									// Six byte PIDs
									case 0x71:
									case 0x9A:
									{
										ExpectedDataSize = 6;
									}
									break;

									// Seven byte PIDs
									case 0x68:
									case 0x69:
									case 0x75:
									case 0x76:
									case 0x7A:
									case 0x7B:
									case 0x8B:
									case 0x8F:
									{
										ExpectedDataSize = 7;
									}
									break;

									// Nine byte PIDs
									case 0x6E:
									case 0x78:
									case 0x79:
									case 0x7C:
									case 0x83:
									case 0x98:
									case 0x99:
									case 0x9F:
									case 0xA1:
									case 0xA3:
									{
										ExpectedDataSize = 9;
									}
									break;

									// Ten byte PIDs
									case 0x70:
									case 0x85:
									{
										ExpectedDataSize = 10;
									}
									break;

									// Eleven byte PIDs
									case 0x6D:
									{
										ExpectedDataSize = 11;
									}
									break;

									// Twelve byte PIDs
									case 0x94:
									{
										ExpectedDataSize = 12;
									}
									break;

									// Thirteen byte PIDs
									case 0x7F:
									case 0x88:
									{
										ExpectedDataSize = 13;
									}
									break;

									// Seventeen byte PIDs
									case 0x8C:
									case 0x9C:
									{
										ExpectedDataSize = 17;
									}
									break;

									// Forty-One byte PIDs
									case 0x81:
									case 0x82:
									case 0x89:
									case 0x8A:
									{
										ExpectedDataSize = 41;
									}
									break;

									// 1 or 2 byte PIDs
									case 0x06:      // PID $06 1 / 2-Byte definition
									case 0x07:      // PID $07 1 / 2-Byte definition
									case 0x08:      // PID $08 1 / 2-Byte definition
									case 0x09:      // PID $09 1 / 2-Byte definition
									case 0x55:      // PID $55 1 / 2-Byte definition
									case 0x56:      // PID $56 1 / 2-Byte definition
									case 0x57:      // PID $57 1 / 2-Byte definition
									case 0x58:      // PID $58 1 / 2-Byte definition
									{
										// Only check if successfully determined the (variable) PID size
										if ( gSid1VariablePidSize > 0 )
										{
											ExpectedDataSize = gSid1VariablePidSize;
										}
										else
										{
											CheckDataSize = FALSE;
										}
									}
									default:
									{
										// Non-OBD PID
										CheckDataSize = FALSE;
									}
									break;
								}  // end switch ( RxMsg->Data[HeaderSize + 2] )

								// Check for correct number of data bytes
								if ( CheckDataSize == TRUE &&
								    (RxMsg->DataSize - (HeaderSize + 3)) != ExpectedDataSize )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $%02X%02X  has %d data bytes (should be %d)\n",
									      GetEcuId ( EcuIndex ),
									      RxMsg->Data[HeaderSize + 1],
									      RxMsg->Data[HeaderSize + 2],
									      (RxMsg->DataSize - (HeaderSize + 3)),
									      ExpectedDataSize );
									return FAIL;
								}
							}  // end if ( SidReq->NumIds == 1 )

							// Determine if there is enough room in the buffer to store the data
							if ( (RxMsg->DataSize - HeaderSize - 1) > sizeof ( gOBDResponse[EcuIndex].Pid ) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  SID $22 response data exceeded buffer size\n",
								      GetEcuId ( EcuIndex ) );
								return FAIL;
							}

							// Save the data in the buffer and set the new size
							memcpy ( &gOBDResponse[EcuIndex].Pid[0],
							         &RxMsg->Data[HeaderSize + 1],
							         (RxMsg->DataSize - HeaderSize - 1) );
							gOBDResponse[EcuIndex].PidSize = (unsigned short)(RxMsg->DataSize - HeaderSize - 1);
						}
						break;
					}  // end case 0xF4 switch (RxMsg->Data[HeaderSize + 2])
				}  // end case 0xF4
				break;

				// PID
				case 0xF5:
				{
					// Process message based on PID number (DID LSB)
					switch ( RxMsg->Data[HeaderSize + 2] )
					{
						case 0x00:
						case 0x20:
						case 0x40:
						case 0x60:
						case 0x80:
						case 0xA0:
						case 0xC0:
						case 0xE0:
						{
							// Make sure there is enough data in the message to process
							if ( RxMsg->DataSize < (HeaderSize + 1 + sizeof ( ID_SUPPORT )) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Not enough data in PID $%02X%02X support response message to process\n",
								      GetEcuId ( EcuIndex ),
								      RxMsg->Data[HeaderSize + 1],
								      RxMsg->Data[HeaderSize + 2] );
								return FAIL;
							}

							// Save the data in the buffer and set the new size
							for ( ulInx = HeaderSize + 2; ulInx < RxMsg->DataSize; ulInx += (sizeof ( ID_SUPPORT ) + 1) )
							{
								// find the element number (for the array of structures)
								bElementOffset = ((unsigned char)(RxMsg->Data[ulInx]) >> 5);

								// if this ID is $F5x0 (in case there is a $F4x0 PID in a reverse group PID support PID response which started with a $F5x0 PID)
								if ( RxMsg->Data[ulInx-1] == 0xF5 )
								{
									// add the offset to the $F5x0 segment of the PidSupport array
									bElementOffset += 8;
								}

								// move the data into the element
								memcpy ( &gOBDResponse[EcuIndex].PidSupport[bElementOffset],
								         &RxMsg->Data[ulInx],
								         sizeof ( ID_SUPPORT ) );

								// increase the element count if the element in the array was previously unused
								if ( (bElementOffset + 1) > gOBDResponse[EcuIndex].PidSupportSize )
								{
									gOBDResponse[EcuIndex].PidSupportSize = (bElementOffset + 1);
								}
							}
						}

						/*
						 * FALL THROUGH AND TREAT LIKE OTHER PIDS
						 */

						default:
						{
							// If not PID 0, then check if this is a response to an unsupported PID
							if ( gIgnoreUnsupported == FALSE &&
							     RxMsg->Data[HeaderSize + 2] != 0x00 &&
							     (gOBDResponse[EcuIndex].PidSupport[
							      8 + ((RxMsg->Data[HeaderSize + 2] - 1) >> 5)].IDBits[
							      ((RxMsg->Data[HeaderSize + 2] - 1) >> 3) & 0x03]
							      & (0x80 >> ((RxMsg->Data[HeaderSize + 2] - 1) & 0x07))) == 0 )
							{
								Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Unsupported PID $%02X%02X detected\n",
								      GetEcuId ( EcuIndex ),
								      RxMsg->Data[HeaderSize + 1],
								      RxMsg->Data[HeaderSize + 2] );
							}

							// Make sure there is enough data in the message to process
							if (RxMsg->DataSize <= (HeaderSize + 2))
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Not enough data in PID $%02X%02X response message to process\n",
								      GetEcuId ( EcuIndex ),
								      RxMsg->Data[HeaderSize + 1],
								      RxMsg->Data[HeaderSize + 2] );
								return FAIL;
							}

							// If only one PID requested,
							// make sure the PID has the correct amount of data
							if ( SidReq->NumIds == 1 )
							{
								CheckDataSize = TRUE;

								switch ( RxMsg->Data[HeaderSize + 2] )
								{
									// Four byte PIDs
									case 0x00:
									case 0x20:
									case 0x40:
									case 0x60:
									case 0x80:
									case 0xA0:
									case 0xC0:
									case 0xE0:
									{
										ExpectedDataSize = 4;
									}
									break;

									// Six byte PIDs
									case 0x01:
									{
										ExpectedDataSize = 6;
									}
									break;

									// Eight byte PIDs
									case 0x02:
									case 0x03:
									case 0x04:
									case 0x05:
									{
										ExpectedDataSize = 8;
									}
									break;

									default:
									{
										// Non-OBD PID
										CheckDataSize = FALSE;
									}
									break;
								}  // end switch ( RxMsg->Data[HeaderSize + 2] )

								// Check for correct number of data bytes
								if ( CheckDataSize == TRUE &&
								     (RxMsg->DataSize - (HeaderSize + 3)) != ExpectedDataSize )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $%02X%02X  has %d data bytes (should be %d)\n",
									      GetEcuId ( EcuIndex ),
									      RxMsg->Data[HeaderSize + 1],
									      RxMsg->Data[HeaderSize + 2],
									      (RxMsg->DataSize - (HeaderSize + 3)),
									      ExpectedDataSize );
									return FAIL;
								}
							}  // end if ( SidReq->NumIds == 1 )

							// Determine if there is enough room in the buffer to store the data
							if ( (RxMsg->DataSize - HeaderSize - 1) > sizeof ( gOBDResponse[EcuIndex].Pid ) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  SID $22 response data exceeded buffer size\n",
								      GetEcuId ( EcuIndex ) );
								return FAIL;
							}

							// Save the data in the buffer and set the new size
							memcpy ( &gOBDResponse[EcuIndex].Pid[0],
							         &RxMsg->Data[HeaderSize + 1],
							         (RxMsg->DataSize - HeaderSize - 1) );
							gOBDResponse[EcuIndex].PidSize = (unsigned short)(RxMsg->DataSize - HeaderSize - 1);
						}
						break;
					}  // end case 0xF4 switch ( RxMsg->Data[HeaderSize + 2] )
				}  // end case 0xF4
				break;

				// MID
				case 0xF6:
				{
					// indicate that this ECU has responded
					gOBDResponse[EcuIndex].bResponseReceived = TRUE;

					// Process message based on MID number
					switch ( RxMsg->Data[HeaderSize + 2] )
					{
						case 0x00:
						case 0x20:
						case 0x40:
						case 0x60:
						case 0x80:
						case 0xA0:
						case 0xC0:
						case 0xE0:
						{
							// Make sure there is enough data in the message to process
							if ( RxMsg->DataSize < (HeaderSize + 2 + sizeof ( ID_SUPPORT )) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Not enough data in SID $22 support response message to process\n",
								      GetEcuId ( EcuIndex ) );
								return FAIL;
							}

							// Save the data in the buffer and set the new size
							for ( ulInx = HeaderSize + 2; ulInx < RxMsg->DataSize; ulInx += (sizeof ( ID_SUPPORT ) + 1) )
							{
								// find the element number (for the array of structures)
								bElementOffset = (unsigned char)(RxMsg->Data[ulInx]) >> 5;

								// move the data into the element
								memcpy ( &gOBDResponse[EcuIndex].MidSupport[bElementOffset],
								         &RxMsg->Data[ulInx],
								         sizeof ( ID_SUPPORT ) );

								// increase the element count if the element in the array was previously unused
								if ( (bElementOffset + 1) > gOBDResponse[EcuIndex].MidSupportSize )
								{
									gOBDResponse[EcuIndex].MidSupportSize = (bElementOffset + 1);
								}
							}
						}

						/*
						 * FALL THROUGH AND TREAT LIKE OTHER PIDS
						 */

						default:
						{
							// Check if this is a response to an unsupported MID
							if ( IsDIDSupported ( EcuIndex, ((RxMsg->Data[HeaderSize + 1]<<8) + RxMsg->Data[HeaderSize + 2]) ) == FALSE )
							{
								Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Unsupported SID $22 MID $%02X%02X detected\n",
								      GetEcuId ( EcuIndex ),
								      RxMsg->Data[HeaderSize + 1],
								      RxMsg->Data[HeaderSize + 2] );
							}

							// Make sure there is enough data in the message to process
							if ( RxMsg->DataSize <= (HeaderSize + 2) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Not enough data in SID $22 response message to process\n",
								      GetEcuId ( EcuIndex ) );
								return FAIL;
							}

							// Save the data in the buffer and set the new size
							// Determine if there is enough room in the buffer to store the data
							if ( (RxMsg->DataSize - HeaderSize - 1) > sizeof ( gOBDResponse[EcuIndex].Mid ) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  SID $22 response data exceeded buffer size\n",
								      GetEcuId ( EcuIndex ) );
								return FAIL;
							}

							memcpy ( &gOBDResponse[EcuIndex].Mid[0],
							         &RxMsg->Data[HeaderSize + 1],
							         (RxMsg->DataSize - HeaderSize - 1));
							gOBDResponse[EcuIndex].MidSize = (unsigned short)(RxMsg->DataSize - HeaderSize - 1);
						}  // end default
						break;
					}  // end case 0xF6 switch ( RxMsg->Data[HeaderSize + 2] )
				}  // end case 0xF6
				break;

				// INF
				case 0xF8:
				{
					// indicate that this ECU has responded
					gOBDResponse[EcuIndex].bResponseReceived = TRUE;

					// Process message based on PID number
					switch ( RxMsg->Data[HeaderSize + 2] )
					{
						case 0x00:
						case 0x20:
						case 0x40:
						case 0x60:
						case 0x80:
						case 0xA0:
						case 0xC0:
						case 0xE0:
						{
							ulInx = HeaderSize + 2;

							// Make sure there is enough data in the message to process
							// Must contain 4-byte address, SID, INF MSB, INF LSB (FirstID), and 4 bytes of data
							if ( RxMsg->DataSize < (ulInx + sizeof ( ID_SUPPORT )) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Not enough data in SID $22 support response message to process.\n",
								      GetEcuId ( EcuIndex ) );
								return FAIL;
							}

							// Save the data in the buffer and set the new size
							for ( ; ulInx < RxMsg->DataSize; ulInx += (sizeof ( ID_SUPPORT ) + 1) )
							{
								// find the element number (for the array of structures)
								bElementOffset = (unsigned char)(RxMsg->Data[ulInx]) >> 5;

								// move the data into the element
								memcpy ( &gOBDResponse[EcuIndex].InfSupport[bElementOffset],
								         &RxMsg->Data[ulInx],
								         sizeof ( ID_SUPPORT ) );

								// increase the element count if the element in the array was previously unused
								if ( (bElementOffset + 1) > gOBDResponse[EcuIndex].InfSupportSize )
								{
									gOBDResponse[EcuIndex].InfSupportSize = (bElementOffset + 1);
								}
							}
						}

						/*
						 * FALL THROUGH AND TREAT LIKE OTHER PIDS
						 */

						default:
						{
							// Check if this is a response to an unsupported INF
							if ( IsDIDSupported ( EcuIndex, ((RxMsg->Data[HeaderSize + 1]<<8) + RxMsg->Data[HeaderSize + 2]) ) == FALSE )
							{
								Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Unsupported INF $%02X%02X detected\n",
								      GetEcuId ( EcuIndex ),
								      RxMsg->Data[HeaderSize + 1],
								      RxMsg->Data[HeaderSize + 2]);
							}

							// Make sure there is enough data in the message to process
							// Must contain at least 4-byte address, SID, INF MSB and INF LSB
							if ( RxMsg->DataSize <= (HeaderSize + 3) )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  Not enough data in SID $22 response message to process\n",
								      GetEcuId ( EcuIndex ) );
								return FAIL;
							}

							// Save the data in the buffer and set the new size

							// Determine if there is enough room in the buffer to store the data
							if ( (RxMsg->DataSize - HeaderSize) > sizeof ( gOBDResponse[EcuIndex].Inf )  )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  SID $22 response data exceeded buffer size\n",
								      GetEcuId ( EcuIndex ) );
								return FAIL;
							}

							memcpy ( &gOBDResponse[EcuIndex].Inf[0],
							         &RxMsg->Data[HeaderSize + 1],
							         (RxMsg->DataSize - (HeaderSize + 1)) );
							gOBDResponse[EcuIndex].InfSize = (unsigned short)(RxMsg->DataSize - (HeaderSize + 1));

							// Verify CVNs contain 4 HEX bytes
							if ( (0xF800 + RxMsg->Data[HeaderSize + 2]) == INF_TYPE_CVN )
							{
								// HeaderSize is 4; 0x62 0xF8 0x06 0x(CVN Count) 0x(DATA) ...
								if ( ( ( RxMsg->DataSize - ( HeaderSize + 4 ) ) % 4 ) != 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  INF $F806  All CVNs must contain 4 HEX bytes.\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							// Verify CALIDs contain 16 HEX bytes
							else if ( (0xF800 + RxMsg->Data[HeaderSize + 2]) == INF_TYPE_CALID )
							{
								// HeaderSize is 4; 0x62 0x04 0x(CALID Count) 0x(DATA) ...
								if ( ( ( RxMsg->DataSize - ( HeaderSize + 4 ) ) % 16 ) != 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  INF $F804  All CALIDs must contain 16 HEX bytes.\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
						}  // end default
						break;
					}
				}  // end case 0xF8
				break;

			}  // end case 0x62 (0x22 response) switch ( RxMsg->Data[HeaderSize + 1] )
		}  // end case 0x62 (0x22 response)
		break;


		default:
		{
			// indicate that this ECU has responded
			gOBDResponse[EcuIndex].bResponseReceived = TRUE;

			// Check for a negative response code
			if ( RxMsg->Data[HeaderSize] == NAK )
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Received negative response code\n",
				      GetEcuId ( EcuIndex ) );
				if ( RxMsg->Data[HeaderSize + 2] == NAK_RESPONSE_PENDING )
				{
					(*pulNumResponses)--;  // don't count the 0x78 response
				}
			}
			else
			{
				// Unexpected SID response
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Unexpected SID $%02X response\n",
				      GetEcuId ( EcuIndex ),
				      (RxMsg->Data[HeaderSize] - OBD_RESPONSE_BIT) );
				return FAIL;
			}
		}

	}  // end switch ( RxMsg->Data[HeaderSize] ) SID

	return PASS;
}


/*
********************************************************************************
**
**  FUNCTION    ChkIFRAdjust
**
**  PURPOSE     Account for checksum or IFR in data size.
**
********************************************************************************
*/
void ChkIFRAdjust ( PASSTHRU_MSG *RxMsg )
{
	// If extra data was returned (e.g. checksum, IFR, ...), remove it
	if ( RxMsg->ExtraDataIndex != 0 && RxMsg->DataSize != RxMsg->ExtraDataIndex )
	{
		// Adjust the data size to ignore the extra data
		RxMsg->DataSize = RxMsg->ExtraDataIndex;
	}
}


/*
********************************************************************************
**
**  FUNCTION    LookupEcuIndex
**
**  PURPOSE     Return index into gOBDResponse struct for this message
**
********************************************************************************
*/
STATUS LookupEcuIndex ( PASSTHRU_MSG *RxMsg, unsigned long *pulEcuIndex )
{
	unsigned long HeaderSize;
	unsigned long EcuIndex;
	unsigned long ByteIndex;
	unsigned long EcuId = 0;

	BOOL bTestFailed = FALSE;

	// Set the response header size based on the protocol
	HeaderSize = gOBDList[gOBDListIndex].HeaderSize;

	// Find the appropriate EcuIndex based on the header information
	for ( EcuIndex = 0; EcuIndex < gNumEcusForAlloc; EcuIndex++ )
	{
		// Check if we have a header match
		for ( ByteIndex = 0; ByteIndex < HeaderSize; ByteIndex++ )
		{
			if ( RxMsg->Data[ByteIndex] != gOBDResponse[EcuIndex].Header[ByteIndex] )
			{
				break;
			}
		}
		// If no match, check if EcuIndex is empty
		if ( ByteIndex != HeaderSize )
		{
			if ( gOBDResponse[EcuIndex].Header[0] == 0x00 &&
			     gOBDResponse[EcuIndex].Header[1] == 0x00 &&
			     gOBDResponse[EcuIndex].Header[2] == 0x00 &&
			     gOBDResponse[EcuIndex].Header[3] == 0x00 )
			{
				// if not currently in the process of determining the protocol
				// then check every Rx msg's ID
				if ( gOBDDetermined == TRUE )
				{
					if ( VerifyEcuID ( &RxMsg->Data[0] ) == FAIL )
					{
						// create the ECU ID for later use
						for ( ByteIndex = 0; ByteIndex < HeaderSize; ByteIndex++ )
						{
							EcuId = (EcuId << 8) + (RxMsg->Data[ByteIndex]);
						}

						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "Response from ECU %X not in initial list\n",
						      EcuId );
						bTestFailed = TRUE;
						break;
					}
				}

				// If empty, add the new response
				memcpy ( &gOBDResponse[EcuIndex].Header[0],
				         &RxMsg->Data[0],
				         HeaderSize );
				break;
			}
		}
		else
		{
			// We have a header match
			break;
		}
	}

	*pulEcuIndex = EcuIndex;

	if ( ( EcuIndex >= OBD_LEGACY_MAX_ECUS &&
	       gOBDList[gOBDListIndex].Protocol != ISO15765_29_BIT_I &&
	       gOBDList[gOBDListIndex].Protocol != ISO15765_250K_29_BIT_I ) ||
	     ( EcuIndex >= OBD_MAX_ECUS &&
	       ( gOBDList[gOBDListIndex].Protocol == ISO15765_29_BIT_I ||
	         gOBDList[gOBDListIndex].Protocol == ISO15765_250K_29_BIT_I ) ) )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Too many OBD ECU responses\n" );
		bTestFailed = TRUE;
	}


	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}

	return PASS;
}


/*
********************************************************************************
**
**  FUNCTION  IsMessageUnique
**
**  PURPOSE   Check to see if this is a duplicate message
**
**  RETURNS   PASS - message is unique
**            FAIL - message is a duplicate
**
********************************************************************************
*/
STATUS IsMessageUnique ( unsigned char *pucBuffer,   // pointer to buffer
                         unsigned short usSize,      // size of buffer
                         unsigned char *pucMsg,      // pointer to message to be checked
                         unsigned short usMsgSize    // size of message
                       )
{
	unsigned short i;
	unsigned short x;
	BOOL bDone = FALSE;

	for ( i = 0, x = 0; i < usSize && x < usMsgSize; i += usMsgSize )
	{
		x = 0;

		while ( pucBuffer[i + x] == pucMsg[x] && x < usMsgSize && (i + x) < usSize )
		{
			x++;
		}

		if ( x == usMsgSize )
		{
			bDone = TRUE;
		}
	}

	if ( bDone == TRUE )
	{
		return FAIL;
	}

	return PASS;
}
