/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


extern STATUS CheckMILLight ( void );


/*
*******************************************************************************
**  TestWithNoDtc - Tests 5.xx
**  Function to run test with no malfunction, no DTC set
*******************************************************************************
*/
STATUS TestWithNoDtc ( void )
{
	BOOL          bSubTestFailed = FALSE;
	unsigned long ulStartTimeMsecs;
	unsigned long ulTimeMsecs;


//*******************************************************************************
	// Test 5.1, Do MIL light check
	gTestSubsection = 1;
	gOBDgTestSubsectionFailed = FALSE;
	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(MIL bulb check, Engine Off, No DTC set)" );

	if ( CheckMILLight () != PASS || gOBDgTestSubsectionFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.2, Determine the OBD protocol to use
	gTestSubsection = 2;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Determine Protocol, Ignition On, Engine Off)" );

	if ( DetermineProtocol () != PASS )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Protocol determination unsuccessful.\n" );
		Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
		return FAIL;
	}

	if ( VehicleReport () != PASS )
	{
		bSubTestFailed = TRUE;
	}
	Log ( PROMPT, SCREENOUTPUTON, LOGOUTPUTON, CUSTOM_PROMPT,
	      "Review the above Vehicle Information.\n\n" );
	if ( Log ( PROMPT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_PROMPT,
	      "Is the information correct and would you like to continue the test?" ) != 'Y' )
	{
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Test Ended By User.\n" );
		return EXIT;
	}

	// check for response to OBD SID $1 PID $00
	if ( VerifyDeprecatedServices ( 0x01 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState (gOBDEngineRunning, gOBDHybridFlag) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.3, Clear codes
	gTestSubsection = 3;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Clear DTCs (Service $14), Engine Off)" );

	if ( ClearCodes () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.4, Verify MIL and DTC status is cleared
	gTestSubsection = 4;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify MIL status bit, Verify no Permanent DTCs, Engine Off)" );

	// check for response to OBD SID $1 PID $00
	if ( VerifyDeprecatedServices ( 0x01 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyMILData () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	// check for response to OBD SID $A
	if ( VerifyDeprecatedServices ( 0x0A ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyPermanentCodeSupport () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.5, Verify monitor test support and results
	gTestSubsection = 5;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Monitor Test Results (Service $22 MIDs), Engine Off)" );

	// check for response to OBD SID $6 PID $00
	if ( VerifyDeprecatedServices ( 0x06 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyMonitorTestSupportAndResults () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.6, Verify diagnostic support and data
	gTestSubsection = 6;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Diagnostic Data (Service $22 PIDs), Engine Off)" );

	// check for response to OBD SID $1 PID $00
	if ( VerifyDeprecatedServices ( 0x01 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyDiagnosticSupportAndData ( FALSE ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.7, Verify control support and data
	gTestSubsection = 7;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Control of System, Test or Component (Service $31), Engine Off)" );

	// check for response to OBD SID $8 PID $00
	if ( VerifyDeprecatedServices ( 0x08 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Request of Test Routine Identifers NOT CURRENTLY IMPLEMENTED" );
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );
//	if ( VerifyControlSupportAndData () != PASS )
//	{
//		bSubTestFailed = TRUE;
//	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.8, Tell user to start engine
	gTestSubsection = 8;
	gOBDgTestSubsectionFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Determine Protocol, Engine Running)" );

	clear_keyboard_buffer ();

	Log ( PROMPT, SCREENOUTPUTON, LOGOUTPUTON, ENTER_PROMPT,
	      "Start engine or activate propulsion system for Hybrid/PHEV.\n\n"

	      "Let idle for one (1) minute.\n\n"

	      "Note: Some powertrain control systems have engine controls that can start and\n"
	      "stop the engine without regard to ignition position.\n" );

	gOBDEngineRunning = TRUE;

	// Determine the OBD protocol to use
	if ( DetermineProtocol () != PASS )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Protocol determination unsuccessful.\n" );
		Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
		return FAIL;
	}

	// check for response to OBD SID $1 PID $00
	if ( VerifyDeprecatedServices ( 0x01 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS ||
	     gOBDgTestSubsectionFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.9, Clear codes
	gTestSubsection = 9;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Clear DTCs (Service $14), Engine Running)" );

	if ( ClearCodes () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.10, Verify diagnostic support and data
	gTestSubsection = 10;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Diagnostic Data (Service $22 PIDs), Engine Running)" );

	// check for response to OBD SID $1 PID $00
	if ( VerifyDeprecatedServices ( 0x01 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyDiagnosticSupportAndData ( FALSE ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.11, Verify freeze frame support and data
	gTestSubsection = 11;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Freeze Frame Data (Service $19 Subfunction $04), Engine Running)" );

	// check for response to OBD SID $2
	if ( VerifyDeprecatedServices ( 0x02 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Read Freeze Frame DTCs NOT CURRENTLY IMPLEMENTED" );
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );
//	if ( VerifyFreezeFrameSupportAndData () != PASS )
//	{
//		bSubTestFailed = TRUE;
//	}
//
	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.12, Verify stored DTC data
	gTestSubsection = 12;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Stored DTCs (Service $19 Subfunction $42 StatusMask $08), Engine Running)" );

	// check for response to OBD SID $3
	if ( VerifyDeprecatedServices ( 0x03 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyDTCStoredData () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyMILData () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.13, Verify monitor test support and results
	gTestSubsection = 13;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Monitor Test Results (Service $22 MIDs), Engine Running)" );

		// check for response to OBD SID $7
	if ( VerifyDeprecatedServices ( 0x07 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyMonitorTestSupportAndResults () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.14, Verify pending DTC data
	gTestSubsection = 14;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Pending DTCs (Service $19 Subfunction $42 StatusMask $04), Engine Running)" );

	// check for response to OBD SID $7
	if ( VerifyDeprecatedServices ( 0x07 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyDTCPendingData () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.15, Verify control support and data
	gTestSubsection = 15;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Control of System, Test or Component (Service $31), Engine Running)" );

	// check for response to OBD SID $8
	if ( VerifyDeprecatedServices ( 0x08 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Request Control of On-Board System NOT CURRENTLY IMPLEMENTED" );
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );

//	if ( VerifyControlSupportAndData () != PASS )
//	{
//		bSubTestFailed = TRUE;
//	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.16, Verify vehicle information support and data
	gTestSubsection = 16;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Vehicle Information (Service $22 INFs), Engine Running)" );

	// check for response to OBD SID $9
	if ( VerifyDeprecatedServices ( 0x09 ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleInformationSupportAndData () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.17, Verify reverse order support
	gTestSubsection = 17;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Diagnostic Data (Service $22 PIDs) in reverse order)" );

	if ( VerifyReversePIDSupport () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.18, Verify tester present support
	gTestSubsection = 18;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Diagnostic Data (Service $22) idle message timing)" );

	// Sleep for 15 seconds, then make sure the tester present messages kept the link alive
//	Sleep ( 15000 );
	ulStartTimeMsecs = GetTickCount ();
	while ( (ulTimeMsecs = (GetTickCount () - ulStartTimeMsecs)) < 15000)
	{
		printf ( "INFORMATION: Idle Time Remaining = %d sec  \r", (15000 - ulTimeMsecs)/1000 );
	}
	printf ( "\n" );

	// Link active test to verify communication remained active for ALL protocols
	if ( VerifyLinkActive () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	 {
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	 }
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.19, Verify diagnostic burst support
	gTestSubsection = 19;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Diagnostic Data (Service $22) burst message timing)" );

	if ( VerifyDiagnosticBurstSupport () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}


//*******************************************************************************
	// Test 5.20, Verify Reserved Services
	gTestSubsection = 20;
	gOBDgTestSubsectionFailed = FALSE;
	bSubTestFailed = FALSE;

	Log ( SUBSECTION_BEGIN, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "(Verify Reserved/Unused Services, Engine Running)" );

	if ( VerifyReservedServices () != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( VerifyVehicleState ( gOBDEngineRunning, gOBDHybridFlag ) != PASS )
	{
		bSubTestFailed = TRUE;
	}

	if ( gOBDgTestSubsectionFailed == TRUE || bSubTestFailed == TRUE )
	{
		if ( Log ( SUBSECTION_FAILED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT, "" ) == 'N' )
		{
			return FAIL;
		}
	}
	else
	{
		Log ( SUBSECTION_PASSED_RESULT, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "" );
	}

	return PASS;
}
