/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


STATUS ReadECUNAME    ( void );
STATUS ReadVIN        ( void );
STATUS ReadCALIDs     ( void );
STATUS ReadCVNs       ( void );
STATUS ReadOBDSUP     ( void );
STATUS ReadFuelType   ( void );
STATUS ReadHybridType ( void );

extern STATUS VerifyPID1C       ( PID *pPid, unsigned long SidIndex, unsigned long EcuIndex );
extern STATUS VerifyECUNameData ( unsigned long EcuIndex );
extern STATUS VerifyVINData     ( unsigned long EcuIndex );
extern STATUS VerifyCALIDData   ( unsigned long EcuIndex );
extern STATUS VerifyCVNData     ( unsigned long EcuIndex );
extern STATUS VerifyInf10Data   ( unsigned long EcuIndex );
extern STATUS VerifyInf12Data   ( unsigned long EcuIndex );
extern STATUS VerifyInf13Data   ( unsigned long EcuIndex );
extern STATUS VerifyInf1BData   ( unsigned long EcuIndex );
extern STATUS VerifyInf1CData   ( unsigned long EcuIndex );


extern const char* FUEL_TYPE[];


/*
********************************************************************************
**  VehicleReport
********************************************************************************
*/
STATUS VehicleReport ()
{
	STATUS        RetCode = PASS;


	// Request INF support data
	if ( RequestDIDSupportData ( INFREQUEST, FALSE ) != PASS )
	{
		return FAIL;
	}

	// Request PID support data
	if ( RequestDIDSupportData ( PIDREQUEST, FALSE) != PASS )
	{
		return FAIL;
	}


	// get ECUNAME
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Read ECUNAME(s)\n" );
	if ( ReadECUNAME () != PASS )
	{
		RetCode = FAIL;
	}
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );


	// get VIN
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Read VIN\n" );
	if ( ReadVIN () != PASS )
	{
		RetCode = FAIL;
	}
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );


	// get CALID
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Read CALID(s)\n" );
	if ( ReadCALIDs () != PASS )
	{
		RetCode = FAIL;
	}
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );


	// get CVNs
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Read CVN(s)\n" );
	if ( ReadCVNs () != PASS )
	{
		RetCode = FAIL;
	}
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );


	// get OBDSUP
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Read OBD Compliance\n" );
	if ( ReadOBDSUP () != PASS )
	{
		RetCode = FAIL;
	}
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );


	// get FUEL TYPE
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	     "Read Fuel_TYP\n" );
	if ( ReadFuelType () != PASS )
	{
		RetCode = FAIL;
	}
	Log ( BLANK, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT, "\n\n" );


	// get Hybrid Information
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	     "Read Hybrid Information\n" );
	if ( ReadHybridType () != PASS )
	{
		RetCode = FAIL;
	}


	return RetCode;
}




/*
*******************************************************************************
**  Function:  ReadECUNAME
**
**  Purpose:   Purpose of this function is to read the ReadECUNAME from the ECUs.
**             If an ECU returns a correctly formatted ReadECUNAME, return TRUE,
**             otherwise return FAIL.
*******************************************************************************
*/
STATUS ReadECUNAME ( void )
{
	unsigned long  EcuIndex;
	SID_REQ        SidReq;
	INF           *pINF;

	STATUS         RetCode = PASS;


	// If INF is supported by any ECU, request it
	if ( IsDIDSupported ( -1, INF_TYPE_ECUNAME ) == TRUE )
	{
		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = INF_TYPE_ECUNAME;
		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "INF $F80A (ECUNAME) request\n" );
			RetCode = FAIL;
		}

		// check responses
		for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
		{
			// If INF is not supported, skip to next ECU
			if ( IsDIDSupported ( EcuIndex, INF_TYPE_ECUNAME ) == FALSE )
			{
				continue;
			}

			if ( gOBDResponse[EcuIndex].InfSize > 0 )
			{
				// Check the data to see if it is valid
				pINF = (INF*)&gOBDResponse[EcuIndex].Inf[0];

				if ( pINF->INFLSB == (INF_TYPE_ECUNAME & 0x00FF) )
				{
					// Validate ECUNAME data
					if ( VerifyECUNameData ( EcuIndex ) == FAIL )
					{
						RetCode = FAIL;
					}
				}
			}
			else
			{
				// No ECUNAME data
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  INF $F80A (ECUNAME) no data\n",
				      GetEcuId ( EcuIndex ) );
				RetCode = FAIL;
			}
		}
	}  // end  if InfSupported
	else
	{
		// No ECUNAME support
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "INF $F80A (ECUNAME) not supported by any ECUs\n" );
		RetCode = FAIL;
	}

	return RetCode;
}



/*
*******************************************************************************
**  Function:  ReadVIN
**
**  Purpose:   Purpose of this function is to read the VIN from the ECUs.
**             If an ECU returns a correctly formatted VIN, return TRUE,
**             otherwise return FAIL.
*******************************************************************************
*/
STATUS ReadVIN ( void )
{
	unsigned long  EcuIndex;
	unsigned long  NumResponses;
	SID_REQ        SidReq;
	INF           *pINF;

	STATUS         RetCode = PASS;


	// If INF is supported by any ECU, request it
	if ( IsDIDSupported ( -1, INF_TYPE_VIN ) == TRUE )
	{
		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = INF_TYPE_VIN;
		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "INF $F802 (VIN) request\n" );
			RetCode = FAIL;
		}

		NumResponses = 0;

		// check responses
		for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
		{
			// If INF is supported
			if ( IsDIDSupported ( EcuIndex, INF_TYPE_VIN ) == TRUE )
			{
				// Check the data to see if it is valid
				pINF = (INF*)&gOBDResponse[EcuIndex].Inf[0];
	
				if ( pINF->INFLSB == (INF_TYPE_VIN & 0x00FF) )
				{
					NumResponses++;
	
					// Copy the VIN into the global array
					memcpy ( gEcuData[EcuIndex].VIN_String, &pINF[0].Data[0], 17 );
				}
	
				if ( VerifyVINData ( EcuIndex ) != PASS )
				{
					RetCode = FAIL;
				}
			}
		}

		// only 1 ECU allowed to support VIN
		if ( NumResponses != 1 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "%u ECUs responded to INF $F802 (VIN).  Only 1 is allowed.\n",
			      NumResponses );
			RetCode = FAIL;
		}
	}  // end  if InfSupported
	else
	{
		// No VIN support
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "INF $F802 (VIN) not reported by any ECU (must be reported by one ECU)\n" );
		RetCode = FAIL;
	}

	return RetCode;
}


/*
*****************************************************************************
**
**  Function:  ReadCALIDs
**
**  Purpose:   print CALIDs to the log file
**
*****************************************************************************
*/
STATUS ReadCALIDs ( void )
{
	unsigned long  EcuIndex;
	SID_REQ        SidReq;
	STATUS         RetCode = PASS;

	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = INF_TYPE_CALID;
	if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "INF $F804 (CALID) request\n" );
		return FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		if ( VerifyCALIDData ( EcuIndex ) == FAIL )
		{
			RetCode = FAIL;
		}
	}

	return RetCode;
}


/*
*****************************************************************************
**
**  Function:  ReadCVNs
**
**  Purpose:   print CVNs to the log file
**
*****************************************************************************
*/
STATUS ReadCVNs ( void )
{
	unsigned long  EcuIndex;
	SID_REQ        SidReq;
	STATUS         RetCode = PASS;

	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = INF_TYPE_CVN;
	if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "INF $F806 (CVN) request\n" );
		return FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		if ( VerifyCVNData ( EcuIndex ) == FAIL )
		{
			RetCode = FAIL;
		}
	}

	return RetCode;
}


/*
*****************************************************************************
**
**  Function:  ReadProtocolID
**
**  Purpose:   Read the PID $F810 (Protocol ID) data from each ECU
**
*****************************************************************************
*/
STATUS ReadProtocolID ( void )
{
	unsigned long  EcuIndex;
	SID_REQ        SidReq;
	INF           *pInf;
	STATUS         RetCode = PASS;

	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF810;
	if ( SidRequest ( &SidReq, SID_REQ_NO_PERIODIC_DISABLE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "PID $F810 (Protocol ID) request\n" );
		return FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		// If INF is not supported, skip to next ECU
		if ( IsDIDSupported ( EcuIndex, 0xF810 ) == FALSE )
		{
			continue;
		}

		if ( gOBDResponse[EcuIndex].InfSize == 0 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  PID $F810 (Protocol ID) no data\n" );
			RetCode = FAIL;
		}
		else
		{
			pInf = (INF*)&gOBDResponse[EcuIndex].Pid[0];

			if ( pInf->INFLSB == 0x10 )
			{
				if ( VerifyInf10Data ( EcuIndex ) == FAIL )
				{
					RetCode = FAIL;
				}
				else
				{
					gEcuData[EcuIndex].ProtocolID = pInf->Data[0];
				}
			}
		}
	}

	return RetCode;
}
/*
*****************************************************************************
**
**  Function:  ReadTGEFN
**
**  Purpose:   Read the PID $F813 (TG/EFN) data from each ECU
**
*****************************************************************************
*/
STATUS ReadTGEFN ( void )
{
	unsigned long  EcuIndex;
	SID_REQ        SidReq;
	INF           *pInf;
	STATUS         RetCode = PASS;

	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF813;
	if ( SidRequest ( &SidReq, SID_REQ_NO_PERIODIC_DISABLE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "INF $F813 (TG/EFN) request\n" );
		return FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		// If INF is not supported, skip to next ECU
		if ( IsDIDSupported ( EcuIndex, 0xF813 ) == FALSE )
		{
			continue;
		}

		if ( gOBDResponse[EcuIndex].InfSize > 0 )
		{
			pInf = (INF*)&gOBDResponse[EcuIndex].Pid[0];

			if ( pInf->INFLSB == 0x13 )
			{
				if ( VerifyInf13Data ( EcuIndex ) == FAIL )
				{
					RetCode = FAIL;
				}
				else
				{
					// Copy the TG/EFN into the global array
					memcpy ( gEcuData[EcuIndex].TGEFN_String, &pInf[0].Data[0], 13 );
				}
			}
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  INF $F813 (TG/EFN) no data\n" );
			return FAIL;
		}
	}

	return RetCode;
}


/*
*****************************************************************************
**
**  Function:  ReadOBDSUP
**
**  Purpose:   Read the PID $F41C data from each ECU
**
*****************************************************************************
*/
STATUS ReadOBDSUP ( void )
{
	unsigned long  EcuIndex;
	SID_REQ        SidReq;
	PID           *pPid;
	STATUS         RetCode = PASS;

	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF41C;
	if ( SidRequest ( &SidReq, SID_REQ_NO_PERIODIC_DISABLE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "PID $F41C (OBDSUP) request\n" );
		return FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		// If INF is not supported, skip to next ECU
		if ( IsDIDSupported ( EcuIndex, 0xF41C ) == FALSE )
		{
			continue;
		}
	
		if ( gOBDResponse[EcuIndex].PidSize > 0 )
		{
			pPid = (PID*)&gOBDResponse[EcuIndex].Pid[0];
	
			if ( pPid->PIDLSB == 0x1c )
			{
				if ( VerifyPID1C ( pPid, 0, EcuIndex ) == FAIL )
				{
					RetCode = FAIL;
				}
				else
				{
					gEcuData[EcuIndex].ODB_Compliance_Level = pPid->Data[0];
				}
			}
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  PID $F41C (OBD_Comp) no data\n" );
			return FAIL;
		}
	}

	return RetCode;
}


/*
*****************************************************************************
**
**  Function:  ReadFuelType
**
**  Purpose:   Read the PID $F451 data from each ECU
**
*****************************************************************************
*/
STATUS ReadFuelType ( void )
{
	unsigned long  EcuIndex;
	SID_REQ        SidReq;
	PID           *pPid;
	STATUS         RetCode = PASS;
	unsigned int   PID51_SupCnt = 0;
	unsigned int   PID51_Cnt    = 0;
	unsigned int   PID51_Saved  = 0;

	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF451;
	if ( SidRequest ( &SidReq, SID_REQ_NO_PERIODIC_DISABLE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "PID $F451 (FUEL_TYP) request\n" );
		return FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		// If INF is not supported, skip to next ECU
		if ( IsDIDSupported ( EcuIndex, 0xF451 ) == FALSE )
		{
			continue;
		}
	
		if ( gOBDResponse[EcuIndex].PidSize > 0 )
		{
			pPid = (PID*)&gOBDResponse[EcuIndex].Pid[0];
	
			if ( pPid->PIDLSB == 0x51 )
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  PID $F451  FUEL_TYP = $%02X (%s)\n",
				      GetEcuId ( EcuIndex ),
				      pPid->Data[0],
				      (pPid->Data[0] <= 0x1F ? FUEL_TYPE[pPid->Data[0]] : FUEL_TYPE[0x20]) );
	
				gEcuData[EcuIndex].PID51 = pPid->Data[0];
	
				if ( PID51_Saved == 0x00 )
				{
					PID51_Saved = pPid->Data[0];
				}
	
				PID51_SupCnt++;
	
				if ( pPid->Data[0] >= 0x20 )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  Invalid FUEL_TYP ($20-$FF reserved)\n",
					      GetEcuId ( EcuIndex ) );
	
					RetCode = FAIL;
				}
//				else if ( (pPid->Data[0] >= 0x0F && pPid->Data[0] <= 0x17) )
//				{
//					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//					     "ECU %X  Invalid FUEL_TYP ($0F-$17 not to be used)\n",
//					     GetEcuId ( EcuIndex ) );
//
//					RetCode = FAIL;
//				}
	
				// if the value from the current ECU matches the value from the 1st responding ECU, increment the number of matching values
				if ( PID51_Saved == pPid->Data[0] )
				{
					PID51_Cnt++;
				}
			}
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  PID $F451 (FUEL_TYP) no data\n" );
			RetCode = FAIL;
		}
	}

	// if the number of matching values doesn't match the number ECUs which support FUEL_TYP, this is a failure
	if ( PID51_Cnt != PID51_SupCnt )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		     "Not all ECUs had the same FUEL_TYP\n" );

		RetCode = FAIL;
	}

	return RetCode;
}


/*
*****************************************************************************
**
**  Function:  ReadHybridType
**
**  Purpose:   Read the PID $F45B, INF $F812, $F81B, $F81C data from each ECU
**
*****************************************************************************
*/
STATUS ReadHybridType ( void )
{
	unsigned long  EcuIndex;
	SID_REQ        SidReq;
	PID           *pPid;
	STATUS         RetCode = PASS;

	SidReq.SID = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF45B;
	if ( SidRequest ( &SidReq, SID_REQ_NO_PERIODIC_DISABLE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "PID $F45B BAT_PWR request\n" );
		RetCode = FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		// If INF is not supported, skip to next ECU
		if ( IsDIDSupported ( EcuIndex, 0xF45B ) == FALSE )
		{
			continue;
		}

		if ( gOBDResponse[EcuIndex].PidSize > 0 )
		{
			pPid = (PID*)&gOBDResponse[EcuIndex].Pid[0];

			if ( pPid->PIDLSB == 0x5B )
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  PID $F45B  BAT_PWR = %f %%\n",
				      GetEcuId ( EcuIndex ),
				      (float)pPid[0].Data[0] * (float)(100.0 / 255.0));

				gEcuData[EcuIndex].HEV = TRUE;
			}
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  PID $F45B BAT_PWR no data\n" );
			RetCode = FAIL;
		}
	}


	SidReq.SID = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF812;
	if ( SidRequest ( &SidReq, SID_REQ_NO_PERIODIC_DISABLE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "INF $F812 (FEOCNTR) request\n" );
		RetCode = FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		// If INF is not supported, skip to next ECU
		if ( IsDIDSupported ( EcuIndex, 0xF812 ) == FALSE )
		{
			continue;
		}

		VerifyInf12Data ( EcuIndex );
	}


	SidReq.SID = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF81B;
	if ( SidRequest ( &SidReq, SID_REQ_NO_PERIODIC_DISABLE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "INF $F81B (Plug-in Hybrid Vehicle Fuel Data) request\n" );
		RetCode = FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		// If INF is not supported, skip to next ECU
		if ( IsDIDSupported ( EcuIndex, 0xF81B ) == FALSE )
		{
			continue;
		}

		VerifyInf1BData ( EcuIndex );
	}


	SidReq.SID = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF81C;
	if ( SidRequest ( &SidReq, SID_REQ_NO_PERIODIC_DISABLE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "INF $F81C (Plug-in Hybrid Vehicle Grid Data) request\n" );
		RetCode = FAIL;
	}

	for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
	{
		// If INF is not supported, skip to next ECU
		if ( IsDIDSupported ( EcuIndex, 0xF81C ) == FALSE )
		{
			continue;
		}

		VerifyInf1CData ( EcuIndex );
	}

	return RetCode;
}
