/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**  Copyright (C) 2002 Drew Technologies. http://j1699-3.sourceforge.net/
**
** ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


/*
*******************************************************************************
**  VerifyDTCStoredData - Function to verify SID3 DTC stored data
**
*******************************************************************************
*/
STATUS VerifyDTCStoredData ( void )
{
	SID_REQ        SidReq;
	unsigned long  EcuIndex;
	unsigned long  NumDTCs;
	unsigned long  DataOffset;
	unsigned long  ulInit_FailureCount = 0;
	DTC           *pDTC;

	ulInit_FailureCount = GetFailureCount ();

	// Request SID $19 $42 $33 $08 data
	SidReq.SID    = 0x19;
	SidReq.NumIds = 4;
	SidReq.DID[0] = 0x42;
	SidReq.DID[1] = 0x33;
	SidReq.DID[2] = 0x08;
	SidReq.DID[3] = 0xFF;
	if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
	{
		// If DTC stored or historical DTC or ISO15765 protocol, there must be a response
		if ( gOBDDTCStored == TRUE ||
		     gOBDDTCHistorical == TRUE ||
		     gOBDList[gOBDListIndex].Protocol == ISO15765 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Sid $19 $42 $33 $08 request failed\n" );
			return FAIL;
		}
	}

	// if a DTC is not supposed to be present (Test 5.12, 6.4, 10.8, 11.8)
	if ( gOBDDTCStored == FALSE && gOBDDTCHistorical == FALSE )
	{
		// Verify that SID $19 $42 $33 $08 reports no DTCs confirmed
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			// warn for ECUs which don't respond
			if ( gOBDResponse[EcuIndex].ConfDTCHeader.DTCSize == 0 )
			{
				Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  No response to SID $19 $42 $33 $08 request\n",
				      GetEcuId ( EcuIndex ) );
				continue;
			}

			pDTC = (DTC*)&gOBDResponse[EcuIndex].ConfDTC[0];
			// check that DTCs are $00
			if ( pDTC->DTC[0] != 0x00 || pDTC->DTC[1] != 0x00 || pDTC->DTC[2] != 0x00 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  SID $19 $42 $33 $08 indicates DTC stored\n",
				      GetEcuId ( EcuIndex ) );
				return FAIL;
			}
		}
	}

	// if a DTC is supposed to be present (Test 7.4, 8.4, 9.4)
	else
	{
		// Verify that SID $19 $42 $33 $08 reports DTCs confirmed
		NumDTCs = 0;
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			// warn for ECUs which don't respond
			if ( gOBDResponse[EcuIndex].ConfDTCHeader.DTCSize == 0 )
			{
				Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  No response to SID $19 $42 $33 $08 request\n",
				      GetEcuId ( EcuIndex ) );
				continue;
			}

			// Print out all the DTCs
			for ( DataOffset = 0; DataOffset < gOBDResponse[EcuIndex].ConfDTCHeader.DTCSize; DataOffset += sizeof ( DTC ) )
			{
				pDTC = (DTC*)&gOBDResponse[EcuIndex].ConfDTC[DataOffset];
				if ( pDTC->DTC[0] != 0 || pDTC->DTC[1] != 0  || pDTC->DTC[2] != 0)
				{
					Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  Confirmed DTC $%02X%02X%02X (Severity = $%02X  Status = $%02X) detected\n",
					      GetEcuId ( EcuIndex ),
					      pDTC->DTC[0],    // High Byte of DTC
					      pDTC->DTC[1],    // Middle Byte of DTC
					      pDTC->DTC[2],    // Low Byte of DTC
					      pDTC->Severity,  // Severity of DTC
					      pDTC->Status );  // Status of DTC
					// Process based on the type of DTC
//					switch ( gOBDResponse[EcuIndex].ConfDTC[DataOffset] & 0xC0 )
//					{
//						case 0x00:
//						{
//							Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "ECU %X  Stored DTC P%02X%02X detected\n",
//							      GetEcuId ( EcuIndex ),
//							      gOBDResponse[EcuIndex].ConfDTC[DataOffset] & 0x3F,
//							      gOBDResponse[EcuIndex].ConfDTC[DataOffset + 1] );
//						}
//						break;
//						case 0x40:
//						{
//							Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "ECU %X  Stored DTC C%02X%02X detected\n",
//							      GetEcuId ( EcuIndex ),
//							      gOBDResponse[EcuIndex].ConfDTC[DataOffset] & 0x3F,
//							      gOBDResponse[EcuIndex].ConfDTC[DataOffset + 1] );
//						}
//						break;
//						case 0x80:
//						{
//							Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "ECU %X  Stored DTC B%02X%02X detected\n",
//							      GetEcuId ( EcuIndex ),
//							      gOBDResponse[EcuIndex].ConfDTC[DataOffset] & 0x3F,
//							      gOBDResponse[EcuIndex].ConfDTC[DataOffset + 1] );
//						}
//						break;
//						case 0xC0:
//						{
//							Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "ECU %X  Stored DTC U%02X%02X detected\n",
//							      GetEcuId ( EcuIndex ),
//							      gOBDResponse[EcuIndex].ConfDTC[DataOffset] & 0x3F,
//							      gOBDResponse[EcuIndex].ConfDTC[DataOffset + 1] );
//						}
//						break;
//					}
					NumDTCs++;
				}
			}
		}

		// If no confirmed DTCs found, then fail
		if ( NumDTCs == 0 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "No ECU indicates DTC confirmed\n" );
			return FAIL;
		}
	}

	// Link active test to verify communication remained active for ALL protocols
	if ( VerifyLinkActive () != PASS )
	{
		return FAIL;
	}

	if ( ulInit_FailureCount != GetFailureCount () )
	{
		// There could have been early/late responses that weren't treated as FAIL
		return FAIL;
	}
	else
	{
		return PASS;
	}

}
