/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**  Copyright (C) 2002 Drew Technologies. http://j1699-3.sourceforge.net/
**
** ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


/*
*******************************************************************************
**  VerifyDiagnosticBurstSupport -
**  Function to verify PID diagnostic burst support
*******************************************************************************
*/
STATUS VerifyDiagnosticBurstSupport ( void )
{
	SID_REQ       SidReq;
	unsigned long StartTimeMsecs;
	unsigned long ulTimeMsecs;
	unsigned long ulReqTotal = 0;   // the number of requests
	unsigned long ulRespTotal = 0;  // the number of responses without fault(s)
	float         fRespReq = 0;     // the proportion of responses to requests
	unsigned long nPid0EcuCount, nPid1EcuCount;

	const unsigned long NumEcusSave = gOBDNumEcus;       // save gOBDNumEcus

	STATUS RetCode = PASS;

	// Count number of ECUs responding to PID $F400
	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF400;
	if ( SidRequest ( &SidReq, SID_REQ_ALLOW_NO_RESPONSE ) != PASS )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "PID $F400 burst support failed\n" );
		return FAIL;
	}

	nPid0EcuCount = gOBDNumEcusResp;

	// Count number of ECUs responding to PID $F501
	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF501;

	if ( SidRequest ( &SidReq, SID_REQ_ALLOW_NO_RESPONSE ) != PASS )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "PID $F501 burst support failed\n" );
		return FAIL;
	}

	nPid1EcuCount = gOBDNumEcusResp;

	// If no ECU supports both PID 0 and PID 1, then cannot run test
	if ( nPid0EcuCount == 0 && nPid1EcuCount == 0 )
	{
		return PASS;
	}

	// Stop the tester present message before burst test
	// since we are using the message in the burst test
	if ( StopPeriodicMsg ( TRUE ) != PASS )
	{
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "StopPeriodicMsg failed\n");
	}

	// Wait for possible race conditions.
	// SidRequest will flush the queue.
	Sleep ( gOBDRequestDelay );

	gSuspendScreenOutput = TRUE;
	// Request PID $F400 and PID $F501 in alternating order for 30 seconds
	SidReq.DID[0] = 0xF501;
	StartTimeMsecs = GetTickCount ();
	while ( (ulTimeMsecs = (GetTickCount () - StartTimeMsecs)) < 30000)
	{
		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		if ( SidReq.DID[0] == 0xF400 )
		{
			SidReq.DID[0] = 0xF501;
		}
		else
		{
			SidReq.DID[0] = 0xF400;
		}

		// set the number of expected responses
		gOBDNumEcus = (SidReq.DID[0] == 0xF400) ? nPid0EcuCount : nPid1EcuCount;

		// if no ECU supports this PID, skip to next PID
		if ( gOBDNumEcus == 0 )
		{
			continue;
		}

		ulReqTotal++;

		if ( SidRequest ( &SidReq, (SID_REQ_ALLOW_NO_RESPONSE|SID_REQ_RETURN_AFTER_ALL_RESPONSES) ) != PASS )
		{
			gOBDNumEcus = NumEcusSave;
			Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $%04X request failed in burst test\n",
			      SidReq.Ids[0] );
		}
		else if (gOBDNumEcusResp != gOBDNumEcus)
		{
			gOBDNumEcus = NumEcusSave;
			Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Missing PID $%04X response in burst test\n",
			      SidReq.Ids[0] );
		}
		else
		{
			ulRespTotal++;
		}

		printf ( "INFORMATION: Burst Time Remaining = %d sec  Requests = %d  Responses = %d  \r",
		         (30000 - ulTimeMsecs)/1000,
		         ulReqTotal,
		         ulRespTotal );
	}
	gSuspendScreenOutput = FALSE;

	printf ( "\n" );

	// restore gOBDNumEcus
	gOBDNumEcus = NumEcusSave;

	// Restart periodic message when done
	if ( StartPeriodicMsg () != PASS )
	{
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "StartPeriodicMsg failed\n");
	}

	// Link active test to verify communication remained active for ALL protocols
	if ( VerifyLinkActive () != PASS )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "VerifyLinkActive failed\n" );
	}

	fRespReq = (float)(ulReqTotal - ulRespTotal)/(float)(ulReqTotal);

	// Log total number of response messages received from burst test
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Burst test result:  Total Number of Requests =  %d,  Total Number of Errors =  %d,  Error Percentage = %2.2f%%\n",
	      ulReqTotal,
	      ulReqTotal - ulRespTotal,
	      (float)(100)*fRespReq );

	// Verify that less than 10% of requests had responses with failures
	if ( fRespReq > (float)(.10) )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Burst test result exceeded maximum error of 10%%\n" );
		RetCode = FAIL;
	}

	return RetCode;
}
