/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


/* Funtion prototypes */
STATUS RequestFFSupportData ( unsigned char *pFreezeFrameDTC, unsigned int FFDTCCount );
STATUS RequestFFIDData ( unsigned char *pFreezeFrameDTC, unsigned int *pFFDTCSupportCount, unsigned int *pFFDTCCount, BOOL *FFSupOnPendFault );
STATUS VerifyFFSupportData (unsigned char *pFreezeFrameDTC);
STATUS VerifyFFDTCinDTCList ( unsigned char *pFreezeFrameDTC );
STATUS VerifyAllFFData ( unsigned char *pFreezeFrameDTC, unsigned int FFDTCSupportCount, unsigned int FFDTCCount );
unsigned int IsFFSupported  ( unsigned int EcuIndex, unsigned int PidIndex );


/*
*******************************************************************************
**  VerifyFreezeFrameSupportAndData - Function to verify freeze frame support and data
*******************************************************************************
*/
STATUS VerifyFreezeFrameSupportAndData ( void )
{
	static unsigned long FreezeFramePendingDTC[OBD_MAX_ECUS] = {0};   // array of pending Freeze Frame DTCs
	unsigned char FreezeFrameDTC[OBD_MAX_ECUS][MAX_RESP_BUF_SIZE];  // array of the Freeze Frame data
	unsigned int  FFDTCCount[OBD_MAX_ECUS];        // the count of ECUs which report a Freeze Frame DTC
	BOOL          FFSupOnPendFault;                // Set if Freeze Frame stored on pending fault
	unsigned int  FFDTCSupportCount;               // the count of ECUs which support SID $19 subfunction $03

	STATUS bReturn = PASS;
	unsigned long ulInit_FailureCount = 0;

	// Initialize data
	memset ( &FFDTCCount[0], 0, sizeof ( FFDTCCount ) );
	FFDTCSupportCount = 0;
	FFSupOnPendFault = FALSE;
	memset ( &FreezeFrameDTC[0], 0, sizeof ( FreezeFrameDTC ) );

	ulInit_FailureCount = GetFailureCount ();

	// Request SID $19 Subfunc $03 data
	if ( RequestFFIDData ( &FreezeFrameDTC[0][0], &FFDTCSupportCount, &FFDTCCount[0], &FFSupOnPendFault ) != PASS )
	{
		bReturn = FAIL;
	}

//	// Request SID $19 subfunction $03 support data
//	if ( RequestFFSupportData ( FreezeFrameDTC, FFDTCCount ) != PASS )
//	{
//		bReturn = FAIL;
//	}

	// Check if a DTC is stored
	// If FF stored on pending fault then verify that it is stored in SID $07
//	if ( gOBDDTCStored    == TRUE ||
//	     FFSupOnPendFault == TRUE )  // If FF stored on pending fault
//	{
//		// Verify the DTC in the freeze frame is in the list of stored DTCs
//		if ( VerifyFFDTCinDTCList ( &FreezeFrameDTC[0][0] ) != PASS )
//		{
//			bReturn = FAIL;
//		}
//
//		// Verify that all SID $19 subfunction $03 PID data is valid
//		if ( VerifyAllFFData ( &FreezeFrameDTC[0][0], FFDTCSupportCount, &FFDTCCount[0] ) != PASS )
//		{
//			bReturn = FAIL;
//		}
//	}
//	else
//	{
//		// Verify that all SID $19 subfunctoin $03 PID data is valid
//		if ( VerifyAllFFData ( &FreezeFrameDTC[0][0], FFDTCSupportCount, &FFDTCCount[0] ) != PASS )
//		{
//			bReturn = FAIL;
//		}
//	}

//	// test 6.5.1 - save FF DTC to compare during next test phase (7.5.1)
//	if ( gTestPhase == eTestPendingDTC )
//	{
//		unsigned long  EcuIndex;
//
//		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
//		{
//			FreezeFramePendingDTC[EcuIndex] = FreezeFrameDTC[EcuIndex];
//		}
//	}
//
//	// test 7.5.1 - compare FF DTC to previous test phase (6.5.1)
//	if ( gTestPhase == eTestConfirmedDTC )
//	{
//		unsigned long  EcuIndex;
//
//		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
//		{
//			if ( FreezeFramePendingDTC[EcuIndex] != 0 &&
//			     FreezeFramePendingDTC[EcuIndex] != FreezeFrameDTC[EcuIndex] )
//			{
//				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT,
//				      "ECU %X  Confirmed Freeze Frame DTC doesn't match Pending FF DTC\n",
//				      GetEcuId ( EcuIndex ));
//				bReturn = FAIL;
//			}
//		}
//	}

	// Try group support
//	if ( FFDTCCount != 0 )
//	{
//		STATUS ret_code = VerifyGroupFreezeFrameSupport ( FreezeFrameDTC );
//		if ( ret_code != PASS )
//		{
//			bReturn = ret_code;
//		}
//
//		// Tests 6.x - 9.x only
//		if ( gTestPhase == eTestPendingDTC    || gTestPhase == eTestConfirmedDTC        ||
//		     gTestPhase == eTestFaultRepaired || gTestPhase == eTestNoFault3DriveCycle )
//		{
//			ret_code = VerifyGroupFreezeFrameResponse ( FreezeFrameDTC );
//			if ( ret_code != PASS )
//			{
//				bReturn = ret_code;
//			}
//		}
//	}

	if ( ulInit_FailureCount != GetFailureCount () )
	{
		// There could have been early/late responses that weren't treated as FAIL
		// or the test failed earlier
		return FAIL;
	}
	else
	{
		return bReturn;
	}

}


/*
******************************************************************************
**
**  Function:  RequestFFSupportData
**
**  Purpose:   Purpose of this routine is to verify that SID $19 subfunction $04
**             returns a support record.  Continue requesting support
**             PIDS thru highest supported group.
**
********************************************************************************
*/
STATUS RequestFFSupportData ( unsigned char *pFreezeFrameDTC, unsigned int FFDTCCount )
{
	STATUS bReturn = PASS;
	unsigned long EcuIndex;
	unsigned long IdIndex;
	unsigned long ulPIDSupport;
	SID_REQ SidReq;

	ulPIDSupport = 0;  // Used to verify that if no PID support, failure may be flagged

	// Request SID $19 subfunction $04 support data
	for ( IdIndex = 0x00; IdIndex < 0x100; IdIndex += 0x20 )
	{
		SidReq.SID    = 0x19;
		SidReq.NumIds = 5;
		SidReq.DID[0] = 0x04;                    // Subfunction: ReportDTCSnapshotRecordByDTCNumber
		SidReq.DID[1] = 0x00;                    // DTCMaskRecord.DTCHighByte
		SidReq.DID[2] = 0x00;                    // DTCMaskRecord.DTCMiddleByte
		SidReq.DID[3] = (unsigned char)IdIndex;  // DTCMaskRecord.DTCLowByte
		SidReq.DID[4] = 0x00;                    // DTCMaskRecord.DTCSnapshotRecordNumber

		if ( SidRequest ( &SidReq, SID_REQ_ALLOW_NO_RESPONSE ) == FAIL )
		{
			// If there were any Freeze Frame DTCs, there must be a response to PID 0x00
			if ( (IdIndex == 0x00 && FFDTCCount != 0 ) ||
			     IdIndex != 0x00 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "No Response to SID $19 PID-Supported PID $%02X request\n",
				      IdIndex );
				return FAIL;
			}
		}

		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			// If the next group is supported, request it
			if ( gOBDResponse[EcuIndex].FFSupport[IdIndex >> 5].IDBits[3] & 0x01 )
			{
				break;
			}
		}

		// If no ECUs support the next group, stop requests
		if ( EcuIndex >= gOBDNumEcus )
		{
			break;
		}
	}


	// Flag error if ECU indicates no support
	if ( VerifyFFSupportData ( pFreezeFrameDTC ) == FAIL )
	{
		bReturn = FAIL;
	}


	// Request next unsupported PID-support PID
	if ( IdIndex != 0xE0 )
	{
		SidReq.SID    = 0x19;
		SidReq.NumIds = 5;
		SidReq.Ids[1] = 0x04;
		SidReq.Ids[0] = (unsigned char)(IdIndex += 0x20);

		if ( SidRequest ( &SidReq, SID_REQ_NORMAL|SID_REQ_IGNORE_NO_RESPONSE ) != FAIL )
		{
			for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
			{
				if ( gOBDResponse[EcuIndex].FFSize != 0 )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  Unexpected response to unsupported SID $19 PID $%02X!\n",
					      GetEcuId ( EcuIndex ),
					      IdIndex );
					bReturn = FAIL;
				}
			}
		}
	}


	// Verify ECU did not drop out
	if ( VerifyLinkActive () != PASS )
	{
		bReturn = FAIL;
	}
	return bReturn;
}


/*
********************************************************************************
**
**  Function: RequestFFIDData
**
**  Purpose: Purpose of this routine is to verify that SID $19 subfunction $03
**           returns stored DTC.  Assuming SID $19 subfunction $03 was supported
**           by the controller, data should be returned.
**
********************************************************************************
*/
STATUS RequestFFIDData ( unsigned char *pFreezeFrameDTC, unsigned int *pFFDTCSupportCount, unsigned int *pFFDTCCount, BOOL *FFSupOnPendFault )
{
	STATUS bReturn = PASS;
	unsigned long EcuIndex;
	SID_REQ SidReq;
	unsigned int DataOffset;
	FreezeFrame *pFF;

	// Request SID $19 subfunction $03 data
	SidReq.SID    = 0x19;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0x03;  // Subfunction:  reportdtcsnapshotidentification

	gIgnoreUnsupported = TRUE;
	if ( SidRequest ( &SidReq, SID_REQ_ALLOW_NO_RESPONSE ) == FAIL )
	{
		gIgnoreUnsupported = FALSE;
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "SID $19 Subfunction $03 request error\n" );
		return FAIL;
	}
	gIgnoreUnsupported = FALSE;

	// Check for freeze frame DTCs
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		// Clear Freeze Frame Array entry for this ECU
		pFreezeFrameDTC[EcuIndex] = 0;

		// Skip if no response or no DTCs stored
		if ( gOBDResponse[EcuIndex].FFSize != 0 )
		{
			// Increment the count of ECUs which support SID $19 subfunction $03
			*pFFDTCSupportCount = *pFFDTCSupportCount + 1;

			if ( gOBDResponse[EcuIndex].FF[1] != 0 ||
			     gOBDResponse[EcuIndex].FF[2] != 0 ||
			     gOBDResponse[EcuIndex].FF[3] != 0 )
			{
				// Increment the count of ECUs which report a Freeze Frame DTC
				*pFFDTCSupportCount = *pFFDTCSupportCount + 1;

				// Save the DTC
				memcpy ( pFreezeFrameDTC, &gOBDResponse[EcuIndex].FF, gOBDResponse[EcuIndex].FFSize );
//				pFreezeFrameDTC[EcuIndex] = ((gOBDResponse[EcuIndex].FF[1] * 65536) +
//				                             (gOBDResponse[EcuIndex].FF[2] * 256) +
//				                              gOBDResponse[EcuIndex].FF[3]);

				for ( DataOffset = 0; DataOffset <gOBDResponse[EcuIndex].FFSize; DataOffset += sizeof ( FreezeFrame ))
				{
					pFF = (FreezeFrame*)&gOBDResponse[EcuIndex].FF[DataOffset];
					pFFDTCCount[EcuIndex] = pFFDTCCount[EcuIndex] + 1;

					// Check if a DTC is not supposed to be present
					if ( gOBDDTCStored     == FALSE &&
					     gOBDDTCPending    == FALSE &&
					     gOBDDTCHistorical == FALSE )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $19 subfunction $03 FFDTC $%06X stored\n",
						      GetEcuId ( EcuIndex ),
						      pFreezeFrameDTC[EcuIndex] );
						bReturn = FAIL;
					}

					else if ( gOBDDTCPending == TRUE  &&
					          gOBDDTCStored  == FALSE )
					{
						Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $19 subfunction $03 FFDTC $%06X stored on pending fault\n",
						      GetEcuId ( EcuIndex ),
						      pFreezeFrameDTC[EcuIndex] );
						*FFSupOnPendFault = TRUE;
					}
					else
					{
						Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $19 subfunction $03 FFDTC $%06X stored\n",
						      GetEcuId ( EcuIndex ),
						      pFreezeFrameDTC[EcuIndex] );
						*FFSupOnPendFault = TRUE;
					}
				}
			}
		}
	}

	// Test 5.11.1 - If all ECUs report Freeze Frame DTCs stored, it is a Failure
	if ( gTestPhase == eTestNoDTC && gTestSubsection == 11 &&
	     *pFFDTCCount == *pFFDTCSupportCount )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "All ECUs reported Freeze Frame DTCs stored\n" );
		bReturn = FAIL;
	}

	return bReturn;
}


/*
*******************************************************************************
**  Function: VerifyFFSupportData
**
**  Purpose: Verify each controller supports at a minimum one PID.
**           Any ECU that responds that does not support at least
**           one PID is flagged as an error.
**
*******************************************************************************
*/
STATUS VerifyFFSupportData (unsigned char *pFreezeFrameDTC)
{
	STATUS          bReturn = PASS;
	int             bEcuResult;
	unsigned long   EcuIndex;
	unsigned long   Index;

	// For each ECU
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		if ( pFreezeFrameDTC[EcuIndex] != 0x00 )
		{
			bEcuResult = FAIL;
			for ( Index = 0; Index < gOBDResponse[EcuIndex].FFSupportSize; Index++ )
			{
				// If MID is supported, keep looking
				if ( ( gOBDResponse[EcuIndex].FFSupport[Index].IDBits[0]  ||
				       gOBDResponse[EcuIndex].FFSupport[Index].IDBits[1]  ||
				       gOBDResponse[EcuIndex].FFSupport[Index].IDBits[2]  ||
				     ( gOBDResponse[EcuIndex].FFSupport[Index].IDBits[3] & 0xFE ) ) != 0x00)
				{
					bEcuResult = PASS;
					break;
				}
			}

			if ( bEcuResult == FAIL && gOBDResponse[EcuIndex].FFSupportSize > 0 )
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  SID $19 PID-Supported PIDs indicate no PIDs supported",
				      GetEcuId ( EcuIndex ) );
				bReturn = FAIL;
			}
		}
	}

	return bReturn;
}


/*
******************************************************************************
**
**  Function:  VerifyFFDTCinDTCList
**
**  Purpose:   Purpose of this routine is to verify the DTC in the freeze
**             frame is in the list of stored DTCs.
**
******************************************************************************
*/
STATUS VerifyFFDTCinDTCList ( unsigned char *pFreezeFrameDTC )
{
	unsigned long  FFEcuIndex;
	unsigned long  DTCEcuIndex;
	unsigned long  DTCIndex;
	unsigned long  DTC;
	unsigned long  StrdDTCCount = 0;
	unsigned long  FFDTCCount = 0;
	unsigned long  MtchdDTCCount = 0;
	unsigned long  MtchdDTC = FALSE;
	STATUS         ReturnCode = PASS;


	// Count all DTCs
	for ( DTCEcuIndex = 0; DTCEcuIndex < gOBDNumEcus; DTCEcuIndex++ )
	{
		if ( gDTCList[DTCEcuIndex].Size != 0 )
		{
			for ( DTCIndex = 0; DTCIndex < gDTCList[DTCEcuIndex].Size; DTCIndex += 2 )
			{
				if ( gDTCList[DTCEcuIndex].DTC[DTCIndex] != 0 ||
				     gDTCList[DTCEcuIndex].DTC[DTCIndex+1] != 0 )
				{
					StrdDTCCount++;
				}
			}
		}
	}

	// Count all Freeze Frames
	for ( FFEcuIndex = 0; FFEcuIndex < gOBDNumEcus; FFEcuIndex++ )
	{
		if ( pFreezeFrameDTC[FFEcuIndex] != 0x0000 )
		{
			FFDTCCount++;
		}
	}

	// Verify each Freeze DTCs is in the list of Confirmed or Pending DTCs
	for ( FFEcuIndex = 0; FFEcuIndex < gOBDNumEcus; FFEcuIndex++ )
	{
		MtchdDTC = FALSE;

		if ( pFreezeFrameDTC[FFEcuIndex] == 0x0000 )
		{
			continue;
		}

		for ( DTCEcuIndex = 0; DTCEcuIndex < gOBDNumEcus; DTCEcuIndex++ )
		{
			for ( DTCIndex = 0; DTCIndex < gDTCList[DTCEcuIndex].Size; DTCIndex += 2 )
			{
				DTC = (gDTCList[DTCEcuIndex].DTC[DTCIndex] * 256) +
				       gDTCList[DTCEcuIndex].DTC[DTCIndex + 1];

				if ( DTC == 0x0000 )
				{
					continue;
				}

				if ( pFreezeFrameDTC[FFEcuIndex] == DTC )
				{
					MtchdDTC = TRUE;
					MtchdDTCCount++;
					break;
				}
			}

			if ( MtchdDTC == TRUE )
			{
				break;
			}
		}

		if ( !MtchdDTC )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Freeze Frame DTC $%02X02X02X not found in list of stored or pending DTCs on any ECU\n",
			      GetEcuId (FFEcuIndex),
			      pFreezeFrameDTC[FFEcuIndex] );
			ReturnCode = FAIL;
		}
	}


	if ( StrdDTCCount !=0  &&
	     MtchdDTCCount == 0 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Stored DTC with no Freeze Frame DTC on any ECU\n" );
		ReturnCode = FAIL;
	}


	return ReturnCode;
}


/*
********************************************************************************
**
**  Function: VerifyAllFFData
**
**  Purpose: Purpose of this routine is to Verify that all SID $19 subfunction $04
**           data is valid.  Verification of minimum support list is
**           performed here.
**
********************************************************************************
*/
STATUS VerifyAllFFData ( unsigned char *pFreezeFrameDTC, unsigned int FFDTCSupportCount, unsigned int FFDTCCount )
{
	STATUS bReturn = PASS;
	STATUS eResult = PASS;

	unsigned long EcuIndex;
	unsigned long IdIndex;
	unsigned long ulPIDSupport = 0;

	SID_REQ SidReq;

	unsigned char fPid02Supported = FALSE;      // Set if SID $19 PID $02 is supported
	unsigned char fPid04Supported = FALSE;      // Set if SID $19 PID $04 is supported
	unsigned char fPid05Supported = FALSE;      // Set if SID $19 PID $05 is supported
	unsigned char fPid0CSupported = FALSE;      // Set if SID $19 PID $0C is supported
	unsigned char fPid0DSupported = FALSE;      // Set if SID $19 PID $0D is supported
	unsigned char fPid11Supported = FALSE;      // Set if SID $19 PID $11 is supported
	unsigned char fPid45Supported = FALSE;      // Set if SID $19 PID $45 is supported
	unsigned char fPid49Supported = FALSE;      // Set if SID $19 PID $49 is supported
	unsigned char fPid67Supported = FALSE;      // Set if SID $19 PID $67 is supported


	// Verify that all SID $19 subfunction $04 data is valid
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		// If more than one ECU supports SID $19 Frame $00 PID $02,
		// clear the PID flags to check for support by each of those ECUs
		if ( FFDTCSupportCount > 1 )
		{
			fPid02Supported = FALSE;
			fPid04Supported = FALSE;
			fPid05Supported = FALSE;
			fPid0CSupported = FALSE;
			fPid0DSupported = FALSE;
			fPid11Supported = FALSE;
			fPid45Supported = FALSE;
			fPid49Supported = FALSE;
			fPid67Supported = FALSE;
		}


		// For each PID group
		for ( IdIndex = 0x01; IdIndex < 0x100; IdIndex++ )
		{
			// skip PID Supported PIDs
			if ( IdIndex == 0x00 || IdIndex == 0x20 || IdIndex == 0x40 || IdIndex == 0x60 ||
			     IdIndex == 0x80 || IdIndex == 0xA0 || IdIndex == 0xC0 || IdIndex == 0xE0 )
			{
				continue;
			}

			// If PID is supported, request it
			if ( IsFFSupported ( EcuIndex, IdIndex ) == TRUE )
			{
				// Set flags if PID supported
				if ( IdIndex == 0x02 )
				{
					fPid02Supported = TRUE;
				}
				else if ( IdIndex == 0x04 )
				{
					fPid04Supported = TRUE;
				}
				else if ( IdIndex == 0x05 )
				{
					fPid05Supported = TRUE;
				}
				else if ( IdIndex == 0x0C )
				{
					fPid0CSupported = TRUE;
				}
				else if ( IdIndex == 0x0D )
				{
					fPid0DSupported = TRUE;
				}
				else if ( IdIndex == 0x11 )
				{
					fPid11Supported = TRUE;
				}
				else if ( IdIndex == 0x45 )
				{
					fPid45Supported = TRUE;
				}
				else if ( IdIndex == 0x49 )
				{
					fPid49Supported = TRUE;
				}
				else if ( IdIndex == 0x67 )
				{
					fPid67Supported = TRUE;
				}

				SidReq.SID    = 19;
				SidReq.NumIds = 2;
				SidReq.Ids[0] = (unsigned char)IdIndex;
				SidReq.Ids[1] = 0; // Frame #0
				eResult = SidRequest ( &SidReq, SID_REQ_NORMAL|SID_REQ_IGNORE_NO_RESPONSE );
				if ( eResult == FAIL )
				{
					// There must be a response when a freeze frame DTC is set
					if ( pFreezeFrameDTC[EcuIndex] != 0x00 )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  No response to SID $19 PID $%02X request\n",
						      GetEcuId ( EcuIndex ),
						      IdIndex );
						bReturn = FAIL;
					}
				}

				else
				{
					if ( eResult == ERRORS )
					{
						// message was early/late; FAIL but keep processing
						bReturn = FAIL;
					}

					//  If PID $02 is not = $0000, it is a Failure
					if ( gTestPhase == eTestNoDTC && gTestSubsection == 11 &&
					     IdIndex == 0x02 &&
					     ( gOBDResponse[EcuIndex].FF[2] != 0x00 ||
					       gOBDResponse[EcuIndex].FF[3] != 0x00 ) )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  Response to SID $19 PID $02 request indicates DTC present\n",
						      GetEcuId ( EcuIndex ),
						      IdIndex );
						bReturn = FAIL;
					}

					// If PID $02 == 0, there should be no response to PID request
					if ( gOBDResponse[EcuIndex].FFSize != 0 &&
					     pFreezeFrameDTC[EcuIndex] == 0x00 &&
					     IdIndex != 0x02 )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  Response to SID $19 PID $%02X request when PID $02 is 0\n",
						      GetEcuId ( EcuIndex ),
						      IdIndex );
						bReturn = FAIL;
					}

					//  If PID is supported and there is a Freeze Frame DTC, there should be a response
					if ( gOBDResponse[EcuIndex].FFSize == 0 &&
					      pFreezeFrameDTC[EcuIndex] != 0x00 )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  No SID $19 PID $%02X data\n",
						      GetEcuId ( EcuIndex ),
						      IdIndex );
						bReturn = FAIL;
					}
				}

			} // if (IsFFSupported...

		} // end for (IdIndex...


		// Test If this ECU has a Freeze Frame DTCs, it must support at least one PID
		if ( pFreezeFrameDTC[EcuIndex] != 0 )
		{
			for ( IdIndex = 0; IdIndex < gOBDResponse[EcuIndex].FFSupportSize; IdIndex++ )
			{
				// If this ECU supports a PID, no need to check anymore PIDs on this ECU
				if ( ( gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[0]  ||
				       gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[1]  ||
				       gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[2]  ||
				     ( gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[3] & 0xFE ) ) != 0x00 )
				{
					ulPIDSupport = 1;
					break;
				}
			}

			// There must be a supported PID
			if ( ulPIDSupport == 0 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  SID $19 no PIDs are supported",
				      GetEcuId ( EcuIndex ) );
				bReturn = FAIL;
			}
		}


		// If more than one ECU supports SID $19 Frame $00 PID $02 and
		// the current ECU has a Freeze Frame DTC, check that the required PIDs are supported
		if ( FFDTCSupportCount > 1 &&
		     pFreezeFrameDTC[EcuIndex] )
		{
			if ( ( fPid02Supported == FALSE )   ||
			     ( fPid04Supported == FALSE )   ||
			     ( ( fPid05Supported == FALSE ) && ( fPid67Supported == FALSE ) ) ||
			     ( fPid0CSupported == FALSE )   ||
			     ( ( gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD ) &&
			       ( ( fPid0DSupported == FALSE ) ||
			         ( ( fPid11Supported == FALSE ) && ( fPid45Supported == FALSE ) && ( fPid49Supported == FALSE ) ) ) ) )
			{
				if ( gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  One or more required SID $19 PIDs (02,04,either 05 or 67,0C,0D and either 11, 45 or 49) are not supported\n",
					      GetEcuId ( EcuIndex ) );
				}
				else
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  One or more required SID $19 PIDs (02,04,either 05 or 67,0C) are not supported\n",
					      GetEcuId ( EcuIndex ) );
				}
				bReturn = FAIL;
			}
		}
	} // end for (EcuIndex...

	// If only one ECU supports SID $2 Frame $00 PID $02
	// and has a Freeze Frame DTC,
	// check that the required PIDs are supported by the vehicle
	if ( FFDTCSupportCount == 1 &&
	     FFDTCCount == 1 )
	{
		if ( ( fPid02Supported == FALSE )   ||
		     ( fPid04Supported == FALSE )   ||
		     ( ( fPid05Supported == FALSE ) && ( fPid67Supported == FALSE ) ) ||
		     ( fPid0CSupported == FALSE )   ||
		     ( ( gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD ) &&
		       ( fPid0DSupported == FALSE ) ) ||
		     ( ( gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD ) &&
		       ( fPid11Supported == FALSE ) && ( fPid45Supported == FALSE ) && ( fPid49Supported == FALSE ) ) )
		{
			if ( gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  One or more required SID $19 PIDs (02,04, either 05 or 67, 0C, 0D, and either 11, 45 or 49) are not supported on this vehicle\n" ,
				      GetEcuId ( EcuIndex ) );
			}
			else
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  One or more required SID $19 PIDs (02,04,either 05 or 67,0C) are not supported\n",
				      GetEcuId ( EcuIndex ) );
			}
			bReturn = FAIL;
		}
	}

	return bReturn;
}


/*
*******************************************************************************
**  VerifyGroupFreezeFrameSupport -
**  Function to verify SID2 group support
*******************************************************************************
*/
STATUS VerifyGroupFreezeFrameSupport ( unsigned long *pFreezeFrameDTC )
{
	unsigned long EcuIndex;
	unsigned long IdIndex;
	unsigned long IdBitIndex;
	unsigned long BitIndex;
	unsigned long BitMask;
	SID_REQ SidReq;

	BOOL bTestFailed = FALSE;

	int bReqNextGroup;  // Binary logic for request for next group!

	// Save the original response data for comparison
	memcpy ( &gOBDCompareResponse[0], &gOBDResponse[0], sizeof ( OBD_DATA ) * gNumEcusForAlloc );

	// Request Freeze Frame (Snapshot) support data as a group
	SidReq.SID = 0x19;
	SidReq.NumIds = 6;
	SidReq.Ids[0] = 0x00;
	SidReq.Ids[1] = 0x00;
	SidReq.Ids[2] = 0x20;
	SidReq.Ids[3] = 0x00;
	SidReq.Ids[4] = 0x40;
	SidReq.Ids[5] = 0x00;
	if ( SidRequest ( &SidReq, SID_REQ_ALLOW_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "SID $2 group ($00-$40) support request failed\n" );
		bTestFailed = TRUE;
	}
	else
	{
		// Set next group select to false
		bReqNextGroup = 0;

		// Determine if next group support request is required
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			// Index 0x02 corresponds to 0x40
			if ( (gOBDResponse[EcuIndex].FFSupport[2].IDBits[3] & 0x01) &&
			     pFreezeFrameDTC[EcuIndex] != 0x00 )
			{
				bReqNextGroup = 1;
			}
		}

		// Request next group if supported by previous request!
		if ( bReqNextGroup == 1 )
		{
			// Request Freeze Frame (Snapshot) support data as a group
			SidReq.SID = 0x19;
			SidReq.NumIds = 6;
			SidReq.Ids[0] = 0x60;
			SidReq.Ids[1] = 0x00;
			SidReq.Ids[2] = 0x80;
			SidReq.Ids[3] = 0x00;
			SidReq.Ids[4] = 0xA0;
			SidReq.Ids[5] = 0x00;
			if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "SID $2 group ($60-$A0) support request failed\n" );
				bTestFailed = TRUE;
			}

			// Set next group select to false
			bReqNextGroup = 0;

			// Determine if next group support request is required
			for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
			{
				// Check 0xA0 for suppport of 0xC0
				if ( (gOBDResponse[EcuIndex].FFSupport[5].IDBits[3] & 0x01) &&
				     pFreezeFrameDTC[EcuIndex] != 0x00 )
				{
					bReqNextGroup = 1;
				}
			}

			// Request next group if supported by previous request!
			if ( bReqNextGroup == 1 )
			{
				// Request Freeze Frame (Snapshot) support data as a group
				SidReq.SID = 0x19;
				SidReq.NumIds = 4;
				SidReq.Ids[0] = 0xC0;
				SidReq.Ids[1] = 0x00;
				SidReq.Ids[2] = 0xE0;
				SidReq.Ids[3] = 0x00;
				if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "SID $2 group ($C0,$E0) support request failed\n" );
					bTestFailed = TRUE;
				}
			}
		}

		// if the ECU has a FF DTC, verify that all Freeze Frame (Snapshot) support data compares
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( pFreezeFrameDTC[EcuIndex] != 0x00 )
			{
				if ( gOBDResponse[EcuIndex].FFSupportSize !=
				     gOBDCompareResponse[EcuIndex].FFSupportSize )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  SID $2 group/individual support response size mismatch\n"
					      "                               group response size = %d\n"
					      "                          individual response size = %d\n",
					      GetEcuId ( EcuIndex ),
					      gOBDResponse[EcuIndex].FFSupportSize,
					      gOBDCompareResponse[EcuIndex].FFSupportSize );
					bTestFailed = TRUE;
				}

				for ( IdIndex = 0; IdIndex < 8; IdIndex++ )
				{
					if ( gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].FirstID != gOBDResponse[EcuIndex].FFSupport[IdIndex].FirstID )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $2 PID $%02X group/individual response Support ID mismatch\n",
						      GetEcuId ( EcuIndex ),
						      IdIndex * 0x20,
						      gOBDResponse[EcuIndex].FFSupport[IdIndex].FirstID );
						bTestFailed = TRUE;

						if ( gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].FirstID != IdIndex * 0x20 )
						{
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "ECU %X  SID $2 PID $%02X invalid individual response Support ID ($%02X)\n",
							      GetEcuId ( EcuIndex ),
							      IdIndex * 0x20,
							      gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].FirstID );
							bTestFailed = TRUE;
						}
						if ( gOBDResponse[EcuIndex].FFSupport[IdIndex].FirstID != IdIndex * 0x20 )
						{
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "ECU %X  SID $2 PID $%02X invalid group response Support ID ($%02X)\n",
							      GetEcuId ( EcuIndex ),
							      IdIndex * 0x20,
							      gOBDResponse[EcuIndex].FFSupport[IdIndex].FirstID );
							bTestFailed = TRUE;
						}
					}

					if ( (gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[0] != gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[0] ) ||
					     (gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[1] != gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[1] ) ||
					     (gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[2] != gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[2] ) ||
					     (gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[3] != gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[3] ) )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $2 PID $%02X group/individual support response mismatch\n"
						      "                              group response = %02X  %02X  %02X  %02X\n"
						      "                         individual response = %02X  %02X  %02X  %02X\n",
						      GetEcuId ( EcuIndex ),
						      IdIndex * 0x20,
						      gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[0],
						      gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[1],
						      gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[2],
						      gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[3],
						      gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[0],
						      gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[1],
						      gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[2],
						      gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[3] );
						bTestFailed = TRUE;

						for ( IdBitIndex = 0; IdBitIndex < 4; IdBitIndex++ )
						{
							if ( (BitMask = gOBDResponse[EcuIndex].FFSupport[IdIndex].IDBits[IdBitIndex] ^ gOBDCompareResponse[EcuIndex].FFSupport[IdIndex].IDBits[IdBitIndex]) != 0 )
							{
								for ( BitIndex = 8; BitIndex != 0; BitIndex-- )
								{
									if ( (BitMask & 0x80) != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  SID $2 PID $%02X Byte %d:  PID $%02X support response mismatch\n",
										      GetEcuId ( EcuIndex ),
										      IdIndex * 0x20,
										      IdBitIndex,
										      (IdIndex * 0x20) + (IdBitIndex * 8) + (8 - (BitIndex - 1)) );
									}
									BitMask <<= 1;
								}
							}
						}
					}
				}
			}
		}
	}

	if ( bTestFailed == TRUE )
	{
		// Something was wrong, restore the original response data
		memcpy ( &gOBDResponse[0], &gOBDCompareResponse[0], sizeof ( OBD_DATA ) * gNumEcusForAlloc );
		return FAIL;
	}

	return PASS;
}


/*
*******************************************************************************
**  VerifyGroupFreezeFrameResponse -
**  Function to verify SID2 group response
*******************************************************************************
*/
STATUS VerifyGroupFreezeFrameResponse ( unsigned long *pFreezeFrameDTC )
{
	unsigned long  EcuIndex;
	unsigned long  IdIndex;
	unsigned long  GroupIndex;
	unsigned long  SingleIndex;
	unsigned long  ByteIndex;
	unsigned long  Index;
	unsigned long  NumGroupIds;

	BOOL bTestFailed = FALSE;

	SID_REQ SidReq, SidGroupReq;

	// alloc enough memory for three (max size) Sid2 responses
//	unsigned char  Sid2Single[sizeof ( SID2 ) * 3];
	unsigned char  Sid2Single[sizeof ( FreezeFrameData ) * 3];
//	unsigned char  Sid2Group[sizeof ( SID2 ) * 3];
	unsigned char  Sid2Group[sizeof ( FreezeFrameData ) * 3];
	unsigned short Sid2Size[3], Sid2GroupSize;

	// test each ECU seperately
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		if ( pFreezeFrameDTC[EcuIndex] != 0x00 )
		{
			SidGroupReq.SID = 2;
			SidGroupReq.NumIds = 0;

			// collect first 3 supported PIDs for current ECU
			for ( IdIndex = 1, NumGroupIds = 0; IdIndex < 256; IdIndex++ )
			{
				// skip PID Supported PIDs
				if ( IdIndex == 0x00 || IdIndex == 0x20 || IdIndex == 0x40 || IdIndex == 0x60 ||
				     IdIndex == 0x80 || IdIndex == 0xA0 || IdIndex == 0xC0 || IdIndex == 0xE0 )
				{
					continue;
				}

				if ( IsFFSupported ( EcuIndex, IdIndex ) == TRUE )
				{
					SidGroupReq.Ids[SidGroupReq.NumIds++] = (unsigned char)IdIndex;
					SidGroupReq.Ids[SidGroupReq.NumIds++] = 0x00;
					if ( ++NumGroupIds == 3 )
					{
						break;
					}
				}
			}

			if ( NumGroupIds == 0 )
			{
				continue;
			}

			// do group request
			if ( SidRequest ( &SidGroupReq, SID_REQ_NORMAL ) == FAIL )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  SID $2 group request failed\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
			else
			{
				// save the original group response data for comparison
				Sid2GroupSize = gOBDResponse[EcuIndex].FFSize;
				memcpy ( Sid2Group, gOBDResponse[EcuIndex].FF, Sid2GroupSize );

				// request each PID seperately
				for ( IdIndex = 0, ByteIndex = 0; IdIndex < NumGroupIds; IdIndex++ )
				{
					SidReq.SID    = 0x19;
					SidReq.NumIds = 2;
					SidReq.Ids[0] = SidGroupReq.Ids[IdIndex * 2];
					SidReq.Ids[1] = 0; // Frame #0

					if ( SidRequest ( &SidReq, SID_REQ_ALLOW_NO_RESPONSE ) == FAIL )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $2 PID %02X request failed\n",
						      GetEcuId ( EcuIndex ),
						      SidGroupReq.Ids[IdIndex * 2] );
						bTestFailed = TRUE;
					}

					// save single response data
					Sid2Size[IdIndex] = gOBDResponse[EcuIndex].FFSize;
					memcpy ( &Sid2Single[ByteIndex], gOBDResponse[EcuIndex].FF, Sid2Size[IdIndex] );
					ByteIndex += Sid2Size[IdIndex];
				}

				// check total data
				if ( ByteIndex != Sid2GroupSize )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  SID $2 PID group request length (%d) / individual request length (%d) mismatch\n",
					      GetEcuId ( EcuIndex ),
					      Sid2GroupSize,
					      ByteIndex );
					bTestFailed = TRUE;
				}

				// compare PID message data
				for ( Index = 0, GroupIndex = 0; (Index < NumGroupIds) && (GroupIndex < Sid2GroupSize); Index++ )
				{
					for ( IdIndex = 0, SingleIndex = 0; (IdIndex < NumGroupIds) && (SingleIndex < Sid2GroupSize); IdIndex++ )
					{
						if ( Sid2Group[GroupIndex] == Sid2Single[SingleIndex] )
						{
							if ( memcmp ( &Sid2Group[GroupIndex], &Sid2Single[SingleIndex], Sid2Size[IdIndex] ) != 0 )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  SID $2 PID $%02X group data / individual data mismatch\n",
								      GetEcuId ( EcuIndex ),
								      Sid2Single[SingleIndex]);
								bTestFailed = TRUE;
							}
							break;
						}

						SingleIndex += Sid2Size[IdIndex];
					}

					if ( IdIndex >= NumGroupIds )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  SID $2 PID group response / individual response mismatch\n",
						      GetEcuId ( EcuIndex ) );
						bTestFailed = TRUE;
						break;
					}

					GroupIndex += Sid2Size[IdIndex];
				}
			}
		} // end if (ECU has FF DTC)
	} // end for (EcuIndex ...

	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}

	return PASS;
}


/*
*****************************************************************************
**
**  Function:  IsFFSupported
**
**  Purpose:   Determine if Freeze Frame XX is supported on specific ECU.
**             Need to have obtained supported PIDs previously.
**             If EcuIndex < 0 then check all ECUs.
**
*****************************************************************************
*/
unsigned int IsFFSupported ( unsigned int EcuIndex, unsigned int PidIndex )
{
	int index1;
	int index2;
	int mask;

	if ( PidIndex == 0 )
	{
		return TRUE;            // all modules must support SID 02 PID 00
	}

	PidIndex--;

	index1 =  PidIndex >> 5;
	index2 = (PidIndex >> 3) & 0x03;
	mask   = 0x80 >> (PidIndex & 0x07);

	if ( (signed int)EcuIndex < 0 )
	{
		for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
		{
			if ( gOBDResponse[EcuIndex].FFSupport[index1].IDBits[index2] & mask )
			{
				return TRUE;
			}
		}
	}
	else
	{
		if ( gOBDResponse[EcuIndex].FFSupport[index1].IDBits[index2] & mask )
		{
			return TRUE;
		}
	}

	return FALSE;
}
