/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
*******************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


unsigned char gPermDTCSupported = 0;
unsigned int  gPermDTCSupECU[OBD_MAX_ECUS];


/*
*******************************************************************************
**  VerifyPermanentCodeSupport -
**  Function to verify SID $19 subfunction $55 support
*******************************************************************************
*/
STATUS VerifyPermanentCodeSupport ( void )
{
	SID_REQ SidReq;
	unsigned long EcuIndex;
	unsigned long NumDTCs;
	unsigned long DataOffset;
	unsigned long ulInit_FailureCount = 0;
	unsigned char TestFailed = FALSE;
	unsigned char i;


	ulInit_FailureCount = GetFailureCount ();

	// Count number of ECUs responding to SID $19 $55
	SidReq.SID = 0x19;
	SidReq.NumIds = 2;
	SidReq.DID[0] = 0x55;  // subfunction reportWWHOBDDTCWithPermanentStatus
	SidReq.DID[1] = 0x33;  // Emissions-system group
	if ( SidRequest ( &SidReq, SID_REQ_NORMAL|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		// An ECU supports Service $19 subfunction $55
		for ( EcuIndex = 0, i = 0, gPermDTCSupported = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( gOBDResponse[EcuIndex].PermDTCSupported )
			{
				gPermDTCSupported++;
				gPermDTCSupECU[i++] = GetEcuId ( EcuIndex );
			}
		}

		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "SID $19 Subfunction $55 request failed\n" );
		return FAIL;
	}


	for ( EcuIndex = 0, i = 0, gPermDTCSupported = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		if ( gOBDResponse[EcuIndex].PermDTCSupported )
		{
			gPermDTCSupported++;
			gPermDTCSupECU[i++] = GetEcuId ( EcuIndex );
		}
	}


	// If a permanent DTC is not supposed to be present (Test 5.4, 9.6, 9.22)
	if ( gOBDDTCPermanent == FALSE )
	{
		// Verify that SID $19 SF $55 reports no DTCs stored
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			// check the Permanent DTCs count
			if ( gOBDResponse[EcuIndex].PermDTC[0] != 0x00 || gOBDResponse[EcuIndex].PermDTC[1] != 0x00 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  SID $19 Subfunction $55 reports Permanent DTC stored\n",
				      GetEcuId ( EcuIndex ) );
				TestFailed = TRUE;
			}
		}
	}

	// If a permanent DTC is supposed to be present (Test 8.6, 9.11, 9.15, 9.18)
	else
	{
		// Verify that SID $19 SF $55 reports DTCs stored
		NumDTCs = 0;
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			// Print out all the DTCs
			for ( DataOffset = 0; DataOffset < gOBDResponse[EcuIndex].PermDTCSize; DataOffset += 4 )
			{
				if ( gOBDResponse[EcuIndex].PermDTC[DataOffset] != 0 ||
				     gOBDResponse[EcuIndex].PermDTC[DataOffset + 1] != 0 )
				{
					Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  Permanent DTC $%02X%02X%02X (Status = $%02X) detected\n",
					      GetEcuId ( EcuIndex ),
					      gOBDResponse[EcuIndex].PermDTC[DataOffset],        // High Byte of DTC
					      gOBDResponse[EcuIndex].PermDTC[DataOffset + 1],    // Middle Byte of DTC
					      gOBDResponse[EcuIndex].PermDTC[DataOffset + 2],    // Low Byte of DTC
					      gOBDResponse[EcuIndex].PermDTC[DataOffset + 3] );  // Status of DTC
//					// Process based on the type of DTC
//					switch ( gOBDResponse[EcuIndex].PermDTC[DataOffset] & 0xC0 )
//					{
//						case 0x00:
//						{
//							Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "ECU %X  Permanent DTC P%02X%02X detected\n",
//							      GetEcuId ( EcuIndex ),
//							      gOBDResponse[EcuIndex].PermDTC[DataOffset] & 0x3F,
//							      gOBDResponse[EcuIndex].PermDTC[DataOffset + 1]);
//						}
//						break;
//						case 0x40:
//						{
//							Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "ECU %X  Permanent DTC C%02X%02X detected\n",
//							      GetEcuId ( EcuIndex ),
//							      gOBDResponse[EcuIndex].PermDTC[DataOffset] & 0x3F,
//							      gOBDResponse[EcuIndex].PermDTC[DataOffset + 1]);
//						}
//						break;
//						case 0x80:
//						{
//							Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "ECU %X  Permanent DTC B%02X%02X detected\n",
//							       GetEcuId ( EcuIndex ),
//							       gOBDResponse[EcuIndex].PermDTC[DataOffset] & 0x3F,
//							       gOBDResponse[EcuIndex].PermDTC[DataOffset + 1]);
//						}
//						break;
//						case 0xC0:
//						{
//							Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "ECU %X  Permanent DTC U%02X%02X detected\n",
//							      GetEcuId ( EcuIndex ),
//							      gOBDResponse[EcuIndex].PermDTC[DataOffset] & 0x3F,
//							      gOBDResponse[EcuIndex].PermDTC[DataOffset + 1]);
//						}
//						break;
//					}
					NumDTCs++;
				}
			}
		}

		// If no stored DTCs found, then fail
		if ( NumDTCs == 0 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "All ECUs  SID $19 Subfunction $55 report no Permanent DTC stored\n" );
			TestFailed = TRUE;
		}
	}

//	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
//	{
//		if ( ( gOBDResponse[EcuIndex].ConfDTCSupported == TRUE ||
//		       gOBDResponse[EcuIndex].PendDTCSupported == TRUE ) &&
//		     gOBDResponse[EcuIndex].PermDTCSupported == FALSE )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supports SID $19 Subfunction 42 Status Mask $04 and $08 but not SID $19 Subfunction $55
//			      GetEcuId ( EcuIndex ) );
//			TestFailed = TRUE;
//		}
//	}

	if ( TestFailed == TRUE || ulInit_FailureCount != GetFailureCount () )
	{
		// There could have been early/late responses that weren't treated as FAIL
		return FAIL;
	}
	else
	{
		return PASS;
	}
}
