/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**  Copyright (C) 2002 Drew Technologies. http://j1699-3.sourceforge.net/
**
** ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


STATUS VerifyReverseGroupDiagnosticSupport ( void );
STATUS VerifySingleAndGroupPIDRequest ( void );
BOOL IsPidInRequest ( SID_REQ SidReqList, unsigned short ubPidId );


extern STATUS DetermineVariablePidSize ( void );
extern STATUS VerifyDidSupportData ( DIDTYPE DidType );


unsigned char gReverseOrderState[OBD_MAX_ECUS];


/*
*******************************************************************************
**  VerifyReversePIDSupport - Function to verify SID 22 PID reverse order support
*******************************************************************************
*/
STATUS VerifyReversePIDSupport ( void )
{
	unsigned long EcuIndex;
	unsigned int  IdIndex;
	unsigned long ulInit_FailureCount = 0;
	SID_REQ SidReq;

	BOOL bTestFailed = FALSE;

	// Save the original response data for comparison
	memcpy ( &gOBDCompareResponse[0], &gOBDResponse[0], sizeof ( OBD_DATA ) * gNumEcusForAlloc );

	ulInit_FailureCount = GetFailureCount ();

	// Ignore unsupported responses
	// Support info is cleared on receive of first response to reverse order request
	gIgnoreUnsupported = TRUE;

	// reverse order request started
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		gReverseOrderState[EcuIndex] = REVERSE_ORDER_REQUESTED;
	}

	// Request SID $22 PID support data in reverse order
	for ( IdIndex = 0x100; IdIndex <= 0x100; IdIndex -= 0x20 )
	{
		SidReq.SID = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = 0xF400+IdIndex;
		if ( SidRequest ( &SidReq, SID_REQ_NORMAL|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
		{
			// There must be a response to PID 0x00
			if ( IdIndex == 0x00 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "No response to PID $F400 request\n" );
				bTestFailed = TRUE;
			}
			else
			{
				for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
				{
					// Check if PID is supported
					if ( gOBDResponse[EcuIndex].PidSupportSize != 0 &&
					     ((gOBDCompareResponse[EcuIndex].PidSupport[(IdIndex - 1) >> 5].IDBits[((IdIndex - 1) >> 3) & 0x03] & (0x80 >> ((IdIndex - 1) & 0x07))) != 0x00) )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  No response to PID $%04X request\n",
						      GetEcuId ( EcuIndex ),
						      0xF400 + IdIndex );
						bTestFailed = TRUE;
					}
				}
			}
		}
		else if ( IdIndex != 0x00 )
		{
			for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
			{
				// Check if data received and PID is not supported
				if ( gOBDResponse[EcuIndex].PidSupportSize != 0 &&
				     (gOBDCompareResponse[EcuIndex].PidSupport[(IdIndex - 1) >> 5].IDBits[((IdIndex - 1) >> 3) & 0x03] & (0x80 >> ((IdIndex - 1) & 0x07))) == 0 )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  Unsupported PID $%04X response\n",
					      GetEcuId ( EcuIndex ),
					      0xF400 + IdIndex );
					bTestFailed = TRUE;
				}
			}
		}
	}


	if ( VerifyDidSupportData ( PIDREQUEST ) == FAIL )
	{
		bTestFailed = TRUE;
	}


	// No longer ignore unsupported response
	gIgnoreUnsupported = FALSE;

	// reverse order request finished
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		gReverseOrderState[EcuIndex] = NOT_REVERSE_ORDER;
	}

	// Verify that all PID reverse/individual support data compares
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		if ( CompareDIDSupport ( EcuIndex,
		                         PIDREQUEST,
		                         gOBDResponse[EcuIndex].PidSupportSize,
		                         gOBDCompareResponse[EcuIndex].PidSupportSize,
		                         &gOBDResponse[EcuIndex].PidSupport[0],
		                         &gOBDCompareResponse[EcuIndex].PidSupport[0],
		                         FALSE ) ==FAIL )
		{
			bTestFailed = TRUE;
		}
	}


	// Determine size of PIDs $06, $07, $08, $09
	if ( DetermineVariablePidSize () != PASS )
	{
		bTestFailed = TRUE;
	}

	// Warn the user that this could take a while
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTOFF, NO_PROMPT,
	      "Please wait (press any key to abort)...\n" );
	clear_keyboard_buffer ();

	// Request PID data in reverse
	for ( IdIndex = 0xF505; IdIndex >= 0xF400; IdIndex -= 0x01 )
	{
		// Bypass PID request if it is a support PID (already done above)
		if ( (IdIndex & 0x001F) != 0 )
		{
			printf ( "INFORMATION: Checking PID $%04X\r", IdIndex );
			SidReq.SID    = 0x22;
			SidReq.NumIds = 1;
			SidReq.DID[0] = IdIndex;
			SidRequest ( &SidReq, SID_REQ_NORMAL|SID_REQ_IGNORE_NO_RESPONSE );
			for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
			{
				// Check if data received and PID is not supported
				if ( gOBDResponse[EcuIndex].PidSize != 0 &&
				     IsDIDSupported ( EcuIndex, IdIndex ) == FALSE )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  Unsupported PID $%04X response\n",
					      GetEcuId ( EcuIndex ),
					      IdIndex );
				}

				// Check if no data received and PID is supported
				if ( gOBDResponse[EcuIndex].PidSize == 0 &&
				     IsDIDSupported (EcuIndex, IdIndex) == TRUE )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  PID $%04X in reverse order failed\n",
					      GetEcuId ( EcuIndex ),
					      IdIndex );
				}
			}
		}

		// Abort test if user presses a key
		if ( _kbhit () )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "User abort\n" );
			clear_keyboard_buffer ();
			return FAIL;
		}
	}

	printf ( "\n" );

	// Make sure reverse group PID support responses match reverse single PID support responses
	if ( VerifyReverseGroupDiagnosticSupport () == FAIL )
	{
		bTestFailed = TRUE;
	}

	// Make sure the single requested PIDs responses match group requested PIDs response.
	if ( VerifySingleAndGroupPIDRequest () == FAIL )
	{
		bTestFailed = TRUE;
	}

	// Link active test to verify communication remained active for ALL protocols
	if ( VerifyLinkActive () != PASS )
	{
		bTestFailed = TRUE;
	}

	if ( bTestFailed == TRUE ||
	     ulInit_FailureCount != GetFailureCount () )
	{
		// There could have been early/late responses that weren't treated as FAIL
		// or the test failed earlier
		return FAIL;
	}
	else
	{
		return PASS;
	}

}


/*
*******************************************************************************
**  VerifyReverseGroupDiagnosticSupport -
**  Function to verify SID $22 PID group support
*******************************************************************************
*/
STATUS VerifyReverseGroupDiagnosticSupport ( void )
{
#define GROUP_PID_REQUEST_SIZE 3
	unsigned long  EcuIndex;
	SID_REQ        GroupSidReq;                             // group requests
	BOOL           bTestFailed = FALSE;


	// Save the original response data for comparison
	memcpy ( &gOBDCompareResponse[0], &gOBDResponse[0], sizeof ( OBD_DATA ) * gNumEcusForAlloc );

	// reverse order request started
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		gReverseOrderState[EcuIndex] = REVERSE_ORDER_REQUESTED;
		gOBDResponse[EcuIndex].PidSupportSize = 0;
		memset ( &gOBDResponse[EcuIndex].PidSupport[0], 0x00, sizeof ( ID_SUPPORT ) * 16 );
	}

	// Request SID 22 PID support data as a reverse (high to low) group
	GroupSidReq.SID    = 0x22;
	GroupSidReq.NumIds = GROUP_PID_REQUEST_SIZE;
	GroupSidReq.DID[0] = 0xF500;
	GroupSidReq.DID[1] = 0xF4E0;
	GroupSidReq.DID[2] = 0xF4C0;
	if ( SidRequest ( &GroupSidReq, SID_REQ_ALLOW_NO_RESPONSE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "SID $22 reverse group $F500-$F4C0 support request failed\n" );
		bTestFailed = TRUE;
	}

	// Request SID 22 PID support data as a reverse (high to low) group
	GroupSidReq.SID    = 0x22;
	GroupSidReq.NumIds = GROUP_PID_REQUEST_SIZE;
	GroupSidReq.DID[0] = 0xF4A0;
	GroupSidReq.DID[1] = 0xF480;
	GroupSidReq.DID[2] = 0xF460;
	if ( SidRequest ( &GroupSidReq,  SID_REQ_ALLOW_NO_RESPONSE|SID_REQ_IGNORE_NO_RESPONSE ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "SID $22 reverse group $F4A0-$F460 support request failed\n" );
		bTestFailed = TRUE;
	}

	// Request PID support data as a group
	GroupSidReq.SID    = 0x22;
	GroupSidReq.NumIds = GROUP_PID_REQUEST_SIZE;
	GroupSidReq.DID[0] = 0xF440;
	GroupSidReq.DID[1] = 0xF420;
	GroupSidReq.DID[2] = 0xF400;
	if ( SidRequest ( &GroupSidReq, SID_REQ_NORMAL ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "SID $22 reverse group $F440-$F400 support request failed\n" );
		bTestFailed = TRUE;
	}

	// Verify that all PID reverse/individual group support data compares
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		if ( CompareDIDSupport ( EcuIndex,
		                         PIDREQUEST,
		                         gOBDResponse[EcuIndex].PidSupportSize,
		                         gOBDCompareResponse[EcuIndex].PidSupportSize,
		                         &gOBDResponse[EcuIndex].PidSupport[0],
		                         &gOBDCompareResponse[EcuIndex].PidSupport[0],
		                         FALSE ) ==FAIL )
		{
			bTestFailed = TRUE;
		}
	}

	// Restore the original response data for comparison
	memcpy ( &gOBDResponse[0], &gOBDCompareResponse[0], sizeof ( OBD_DATA ) * gNumEcusForAlloc );

	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}

	return PASS;
}


/*
*******************************************************************************
**  VerifyReverseGroupDiagnosticSupport -
**  Function to verify SID $22 PID group support
*******************************************************************************
*/
STATUS VerifySingleAndGroupPIDRequest ( void )
{
#define GROUP_PID_REQUEST_SIZE 3
	unsigned long  EcuIndex;
	unsigned long  IdIndex;
	unsigned short usByteIndex;
	unsigned short usPidCount = 0;                          // temp. count of PIDs for current ECU
	unsigned short usNumberOfPIDsRequested = 0;
	unsigned short usDesiredResponseSize = 0;               // expected response size for current ECU
	unsigned short rgusEcuPidSize[GROUP_PID_REQUEST_SIZE];  // the size for each PID the ECU supports
	unsigned short rgusSinglePidSize[OBD_MAX_ECUS][GROUP_PID_REQUEST_SIZE];  // for each ECU, the size for each PID in the group
	SID_REQ        rgsEcuSpecificSidReq[OBD_MAX_ECUS];      // the last 6 PIDs for each ECU
	SID_REQ        SingleSidReq;                            // single requests
	SID_REQ        GroupSidReq;                             // group requests
	BOOL           bTestFailed = FALSE;


	// find the list of the last three PIDs supported by each ECU
	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		for ( IdIndex = 0xF505, usPidCount = 0; (IdIndex > 0xF400) && (usPidCount < GROUP_PID_REQUEST_SIZE); IdIndex-- )
		{
			// Don't use PID supported PIDs
			if ( IdIndex != 0xF400 && IdIndex != 0xF420 &&
			     IdIndex != 0xF440 && IdIndex != 0xF460 &&
			     IdIndex != 0xF480 && IdIndex != 0xF4A0 &&
			     IdIndex != 0xF4C0 && IdIndex != 0xF4E0 &&
			     IdIndex != 0xF500 )
			{
				// If PID is supported, add it to the list
				if ( IsDIDSupported ( EcuIndex, IdIndex ) == TRUE )
				{
					rgsEcuSpecificSidReq[EcuIndex].DID[usPidCount] = IdIndex;
					usPidCount++;
				}
			}
		}
		rgsEcuSpecificSidReq[EcuIndex].NumIds = (unsigned char)usPidCount;
	}


	// combine the list of PIDs from each ECU into one list with at least one PID from each ECU
	GroupSidReq.SID    = 0x22;
	GroupSidReq.NumIds = 0;
	IdIndex = 0;
	EcuIndex = 0;

	do
	{
		// add one PID at a time to the actual request
		if ( IdIndex < rgsEcuSpecificSidReq[EcuIndex].NumIds )
		{
			// still some PIDs in the specific ECU list
			if ( IsPidInRequest ( GroupSidReq, rgsEcuSpecificSidReq[EcuIndex].DID[IdIndex] ) == FALSE )
			{
				// not already in the list, so add it
				GroupSidReq.DID[GroupSidReq.NumIds] = rgsEcuSpecificSidReq[EcuIndex].DID[IdIndex];
				GroupSidReq.NumIds++;
			}
		}

		// point to next ECU
		EcuIndex++;

		if ( EcuIndex >= gOBDNumEcus )
		{
			// at the end of the ECU list, start over with next PID for each ECU
			EcuIndex = 0;
			IdIndex++;
		}

	} while ( GroupSidReq.NumIds < GROUP_PID_REQUEST_SIZE && IdIndex < GROUP_PID_REQUEST_SIZE );


	if ( GroupSidReq.NumIds != 0 )
	{
		// request the PIDs individually, to get the size information
		for ( usPidCount = 0; usPidCount < GroupSidReq.NumIds; usPidCount++ )
		{
			SingleSidReq.SID    = 0x22;
			SingleSidReq.NumIds = 1;
			SingleSidReq.DID[0] = GroupSidReq.DID[usPidCount];

			if ( SidRequest ( &SingleSidReq, SID_REQ_NORMAL ) == FAIL )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "Single PID $%04X request failed\n",
				      SingleSidReq.DID[0] );
				bTestFailed = TRUE;
			}

			// save the response PID size
			for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
			{
				rgusSinglePidSize[EcuIndex][usPidCount] = gOBDResponse[EcuIndex].PidSize;
			}
		}


		// request the PIDs as a group and check the responses
		if ( SidRequest ( &GroupSidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "SID $22 group PID request failed\n" );
			bTestFailed = TRUE;
		}
		else
		{
			for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
			{
				// rebuild ECU specific PID list, based on final group request
				for ( IdIndex = 0, usPidCount = 0, usDesiredResponseSize = 0;
				      IdIndex < GROUP_PID_REQUEST_SIZE;
				      IdIndex++ )
				{
					if ( IsDIDSupported ( EcuIndex, (GroupSidReq.DID[IdIndex] ) ) == TRUE )
					{
						rgsEcuSpecificSidReq[EcuIndex].DID[usPidCount] = GroupSidReq.DID[IdIndex];
						rgusEcuPidSize[usPidCount] = rgusSinglePidSize[EcuIndex][IdIndex];
						usDesiredResponseSize += rgusSinglePidSize[EcuIndex][IdIndex];
						usPidCount++;
					}
				}
				rgsEcuSpecificSidReq[EcuIndex].NumIds = (unsigned char)usPidCount;

				// check the response size
				if ( gOBDResponse[EcuIndex].PidSize != usDesiredResponseSize )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  SID $22 group/single PID size mismatch (Response = %d bytes  Expected = %d bytes)\n",
					      GetEcuId ( EcuIndex ),
					      gOBDResponse[EcuIndex].PidSize,
					      usDesiredResponseSize );
					bTestFailed = TRUE;
				}
				else if ( gOBDResponse[EcuIndex].PidSize != 0 )
				{
					// size is good, check message content
					usNumberOfPIDsRequested = rgsEcuSpecificSidReq[EcuIndex].NumIds;

					for ( usByteIndex = 0;
					      usNumberOfPIDsRequested > 0 && usByteIndex < gOBDResponse[EcuIndex].PidSize;
					      usNumberOfPIDsRequested-- )
					{
						for ( IdIndex = 0; IdIndex < rgsEcuSpecificSidReq[EcuIndex].NumIds; IdIndex++ )
						{
							if ( ((gOBDResponse[EcuIndex].Pid[usByteIndex] << 8) + gOBDResponse[EcuIndex].Pid[usByteIndex+1]) == rgsEcuSpecificSidReq[EcuIndex].DID[IdIndex] )
							{
								break;
							}
						}

						if ( IdIndex < rgsEcuSpecificSidReq[EcuIndex].NumIds )
						{
							// found the PID, get the offset to the next PID
							usByteIndex += rgusEcuPidSize[IdIndex];
						}
						else
						{
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "ECU %X  group/single PID mismatch (PID $%04X)\n",
							      GetEcuId ( EcuIndex ),
							      rgsEcuSpecificSidReq[EcuIndex].DID[IdIndex] );
							bTestFailed = TRUE;
							break;
						}
					} // end for (each PID)
				} // end else (size is good)
			} // end for (each ECU)
		}
	}

	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}

	return PASS;
}


/*
*******************************************************************************
**  IsPidInRequest -
**  Function to check if the designated PID is already in the request
*******************************************************************************
*/
BOOL IsPidInRequest ( SID_REQ SidReqList, unsigned short ubPidId )
{
	unsigned short usPidIndex;


	for ( usPidIndex = 0; usPidIndex < SidReqList.NumIds; usPidIndex ++ )
	{
		if ( SidReqList.DID[usPidIndex] == ubPidId )
		{
			return TRUE;
		}
	}

	return FALSE;
}
