/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


void   AbortTest ( void );
STATUS LogRPM ( unsigned short *pusRPM );

extern void   StopTest ( STATUS ExitCode, TEST_PHASE egTestPhase );


/*
*****************************************************************************
**
**  Function:  VerifyVehicleState
**
**  Purpose:   Purpose is desired vehicle state to actual vehicle state.
**
*****************************************************************************
*/
STATUS VerifyVehicleState ( unsigned char ucEngineRunning, unsigned char ucHybrid )
{
	STATUS RetCode = PASS;
	unsigned short usRPM;


	if ( LogRPM ( &usRPM ) == PASS )
	{
		Log ( INFORMATION, SCREENOUTPUTOFF, LOGOUTPUTON, NO_PROMPT,
		      "VerifyVehicleState - Expected State: %s, RPM: %u, Hybrid: %s\n",
		      ((ucEngineRunning == TRUE) ? "Running" : "Off" ),
		      usRPM,
		      ((ucHybrid == TRUE) ? "Yes" : "No" ) );

		if ( ucEngineRunning == FALSE && usRPM > 0 )
		{
			// Engine is running, but it should be off
			if ( Log ( USER_ERROR, SCREENOUTPUTON, LOGOUTPUTON, QUIT_CONTINUE_PROMPT,
			           "Engine is running, when it should be off.\n" ) == 'Q' )
			{
				RetCode = ABORT;
			}
		}
		else if ( ucEngineRunning == TRUE && usRPM < 300 )
		{
			// Engine is off, but it should be running
			if ( ucHybrid == TRUE )
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "Engine is off, but it should be running.\n" );
			}
			else
			{
				if ( Log ( USER_ERROR, SCREENOUTPUTON, LOGOUTPUTON, QUIT_CONTINUE_PROMPT,
				           "Engine is off, when it should be running.\n" ) == 'Q' )
				{
					RetCode = ABORT;
				}
			}
		}
	}
	else
	{
		if ( Log ( PROMPT, SCREENOUTPUTON, LOGOUTPUTON, QUIT_CONTINUE_PROMPT,
		           "Cannot determine engine state.\n" ) == 'Q' )
		{
			RetCode = ABORT;
		}
		else
		{
			RetCode = FAIL;
		}
	}


	if ( RetCode == ABORT )
	{
		AbortTest ();  // this function never returns
	}

	return RetCode;
}


/*
*****************************************************************************
**
**  Function:  LogRPM
**
**  Purpose:   Purpose is to read current vehicle engine speed.
**
*****************************************************************************
*/
STATUS LogRPM ( unsigned short *pusRPM )
{
	SID_REQ         SidReq;
	PID            *pPid;
	unsigned short  RPM = 0;
	unsigned long   EcuIndex;
	STATUS          eRetCode = FAIL;

	// request RPM - PID $F40C
	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF40C;

	if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "No response to PID $F40C (RPM) request.\n" );
		eRetCode = FAIL;
	}

	else
	{
		for ( EcuIndex = 0; EcuIndex < gUserNumEcus; EcuIndex++ )
		{
			if ( gOBDResponse[EcuIndex].PidSize > 0 )
			{
				pPid = (PID *)&gOBDResponse[EcuIndex].Pid[0];

				if ( pPid->PIDLSB == 0x0C )
				{
					RPM = pPid->Data[0];
					RPM = RPM << 8 | pPid->Data[1];

					// convert from 1 cnt = 1/4 RPM to 1 cnt = 1 RPM
					RPM >>= 2;

					Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  PID $F40C  RPM = %d\n",
					      GetEcuId ( EcuIndex ),
					      RPM );
					eRetCode = PASS;
				}
			}
		}
	}

	*pusRPM = RPM;
	return eRetCode;
}


/*
********************************************************************************
**  AbortTest
********************************************************************************
*/
void AbortTest ( void )
{

	// terminate the test
	Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "User Aborted :: terminating application\n\n" );

	gOBDTestAborted = TRUE;
	LogStats ();
	StopTest ( ABORT, gTestPhase );

	exit ( ABORT );
}
