/*
********************************************************************************
**  SAE J1699-5 Test Source Code
**
**   Copyright (C) 2021
**
**  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**
**  This source code, when compiled and used with an SAE J2534-compatible pass
**  thru device, is intended to run the tests described in the SAE J1699-5
**  document in an automated manner.
**
**  This computer program is based upon SAE Technical Report J1699,
**  which is provided "AS IS"
**
**  See j1699.c for details of how to build and run this test.
**
********************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <windows.h>
#include "j2534.h"
#include "j1699.h"


// PID $01 Bit defines - for tracking support when
// 'at least one ECU' must have these bits set
#define DATA_B_BIT_0 0x00010000
#define DATA_B_BIT_1 0x00020000
#define DATA_B_BIT_2 0x00040000
#define DATA_B_BIT_3 0x00080000
#define DATA_C_BIT_0 0x00000100
#define DATA_C_BIT_1 0x00000200
#define DATA_C_BIT_2 0x00000400
#define DATA_C_BIT_3 0x00000800
#define DATA_C_BIT_5 0x00002000
#define DATA_C_BIT_6 0x00004000
#define DATA_C_BIT_7 0x00008000


// PID $F501 Bit defines - for tracking support when
// 'at least one ECU' must have these bits set
#define MIS_SUP   0x00000001
#define FUEL_SUP  0x00000002
#define COOL_SUP  0x00000004
#define CAT_SUP   0x00000008
#define HCCAT_SUP 0x00000010
#define SCR_SUP   0x00000020
#define EVAP_SUP  0x00000040
#define PM_SUP    0x00000080
#define DATA_C_BIT_5 0x00002000
#define DATA_C_BIT_6 0x00004000
#define DATA_C_BIT_7 0x00008000


/************************
**  OBD type definitions
************************/
#define MAX_OBD_TYPES            43
const char *OBD_TYPE[MAX_OBD_TYPES] = {
"UNKNOWN",                      // 0x00
"OBD II",                       // 0x01
"OBD",                          // 0x02
"OBD and OBDII",                // 0x03
"OBD I",                        // 0x04
"NO OBD",                       // 0x05
"EOBD",                         // 0x06
"EOBD and OBD II",              // 0x07
"EOBD and OBD",                 // 0x08
"EOBD, OBD, and OBD II",        // 0x09
"JOBD",                         // 0x0A
"JOBD and OBD II",              // 0x0B
"JOBD and EOBD",                // 0x0C
"JOBD, EOBD, and OBD II",       // 0x0D
"OBD, EOBD, and KOBD",          // 0x0E
"OBD, OBDII, EOBD, and KOBD",   // 0x0F
"reserved",                     // 0x10
"EMD",                          // 0x11
"EMD+",                         // 0x12
"HD OBD-C",                     // 0x13
"HD OBD",                       // 0x14
"WWH OBD",                      // 0x15
"reserved",                     // 0x16
"HD EOBD-I",                    // 0x17
"HD EOBD-I N",                  // 0x18
"HD EOBD-II",                   // 0x19
"HD EOBD-II N",                 // 0x1A
"reserved",                     // 0x1B
"OBDBr-1",                      // 0x1C
"OBDBr-2 and 2+",               // 0x1D
"KOBD",                         // 0x1E
"IOBD I",                       // 0x1F
"IOBD II",                      // 0x20
"HD EOBD-VI",                   // 0x21
"OBD, OBDII, and HD OBD",       // 0x22
"OBDBr-3",                      // 0x23
"Motorcycle, Euro OBD-I",       // 0x24
"Motorcycle, Euro OBD-II",      // 0x25
"Motorcycle, China OBD-I",      // 0x26
"Motorcycle, Taiwan OBD-I",     // 0x27
"Motorcycle, Japan OBD-I",      // 0x28
"China Nationwide Stage 6",     // 0x29
"Brazil OBD Diesel"             // 0x2A
};


/************************
**  Fuel type definitions
************************/
#define MAX_FUELTYPE_LENGTH 10
const char* FUEL_TYPE[] =
{
	"NONE",      // 00 Not Available
	"GAS",       // 01 Gasoline
	"METH",      // 02 Methanol
	"ETH",       // 03 Ethanol
	"DSL",       // 04 Diesel
	"LPG",       // 05 Liquefied Petroleum Gas
	"CNG",       // 06 Compressed Natural Gas
	"PROP",      // 07 Propane
	"ELEC",      // 08 Battery/Electric
	"BI_GAS",    // 09 Bi-fuel vehicle using gasoline
	"BI_METH",   // 0A Bi-fuel vehicle using methanol
	"BI_ETH",    // 0B Bi-fuel vehicle using ethanol
	"BI_LPG",    // 0C Bi-fuel vehicle using LPG
	"BI_CNG",    // 0D Bi-fuel vehicle using CNG
	"BI_PROP",   // 0E Bi-fuel vehicle using propane
	"BI_ELEC",   // 0F Bi-fuel vehicle using battery [Not to be used]
	"BI_MIX",    // 10 Bi-fuel vehicle using battery and combustion engine [Not to be used]
	"HYB_GAS",   // 11 Hybrid vehicle using gasoline engine [Not to be used]
	"HYB_ETH",   // 12 Hybrid vehicle using gasoline engine on ethanol [Not to be used]
	"HYB_DSL",   // 13 Hybrid vehicle using diesel engine [Not to be used]
	"HYB_ELEC",  // 14 Hybrid vehicle using battery [Not to be used]
	"HYB_MIX",   // 15 Hybrid vehicle using battery and combustion engine [Not to be used]
	"HYB_REG",   // 16 Hybrid vehicle in regeneration mode [Not to be used]
	"NOT_USED",  // 17 (Bi-fuel vehicle using diesel) or (Natural Gas) (as defined in J1979) [Not to be used]
	"BI_NG",     // 18 Bi-fuel vehicle using Natural Gas
	"BI_DSL",    // 19 Bi-fuel vehicle using diesel
	"NG",        // 1A Natural Gas (Compressed or Liquefied Natural Gas)
	"DSL_CNG",   // 1B Dual Fuel  Diesel and CNG
	"DSL_LNG",   // 1C Dual Fuel  Diesel and LNG
	"FC_H2",     // 1D Fuel cell utilizing hydrogen
	"HICE_HHO",  // 1E Hydrogen Internal Combustion Engine
	"KERO",      // 1F Kerosene
	"RESERVED"   // 20 -FF RESERVED
};


/************************
**  Gear definitions
************************/
const char* GEAR[] =
{
	"NEUTRAL",      // 00 Neutral
	"GEAR 1",
	"GEAR 2",
	"GEAR 3",
	"GEAR 4",
	"GEAR 5",
	"GEAR 6",
	"GEAR 7",
	"GEAR 8",
	"GEAR 9",
	"GEAR 10",
	"GEAR 11",
	"GEAR 12",
	"GEAR 13",
	"GEAR 14",
	"GEAR 15"
};


STATUS VerifyPID01   ( PID *pPid, unsigned long EcuIndex, unsigned long *PidSupported );
STATUS VerifyPID1C   ( PID *pPid, unsigned long EcuIndex );
STATUS VerifyPID41   ( PID *pPid, unsigned long EcuIndex, unsigned char OBDEngineDontCare );
STATUS VerifyPIDF501 ( PID *pPid, unsigned long EcuIndex, unsigned long *PidSupported );

STATUS DetermineVariablePidSize ( void );

STATUS GetPID4FArray ( void );
STATUS GetPID50Array ( void );

void GetHoursMinsSecs ( unsigned long time, unsigned long *hours, unsigned long *mins, unsigned long *secs );

unsigned char Pid4F[OBD_MAX_ECUS][4];
unsigned char Pid50[OBD_MAX_ECUS][4];

//PID *PID01;    // capture the response from SID01 PID01 response.


/*
*******************************************************************************
**  VerifyDiagnosticSupportAndData -
**  Function to verify PID diagnostic support and data
*******************************************************************************
*/
STATUS VerifyDiagnosticSupportAndData ( unsigned char OBDEngineDontCare )
{
	unsigned int EcuIndex;
	unsigned int IdIndex;
	SID_REQ SidReq;
	PID    *pPid;
	unsigned long ulInit_FailureCount = 0;
	unsigned long ulTemp_FailureCount = 0;
	long  temp_data_long;
	float temp_data_float;
	float temp_EGR_CMD;
	float temp_EGR_ACT;
	unsigned int  BSIndex;
	unsigned char fPid13Supported = FALSE;      // set if the current ECU supports PID $F413
	unsigned char fPid1DSupported = FALSE;      // set if the current ECU supports PID $F41D
	unsigned char fPid4FSupported = FALSE;      // set if the current ECU supports PID $F44F
	unsigned char fPid50Supported = FALSE;      // set if the current ECU supports PID $F450
	unsigned char fReqPidNotSupported = FALSE;  // set if a required PID is not supported
	BOOL          fPidSupported[MAX_PIDS+1];    // an array of PIDs (TRUE if PID is supported)

	                                            // However, PID $01 is different as we must track
	                                            // support for specific bits. In this case, the ULONG
	                                            // will be structured so that DATA_A is the most
	                                            // significant byte and DATA_D is the least significant
	                                            // byte (i.e., AABBCCDD)
	unsigned long ulIMBitsForVehicle = 0;       // holds the PID $01 IM bits supported by the vehicle
	unsigned char fPid9FSuccess = FALSE;
	unsigned long hours;
	unsigned long mins;
	unsigned long secs;


	ulInit_FailureCount = GetFailureCount ();

	// Read PID support PIDs
	if ( RequestDIDSupportData ( PIDREQUEST, FALSE ) != PASS )
	{
		if ( Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT,
		           "Problems verifying PID supported data.\n" ) == 'N' )
		{
			return FAIL;
		}
	}

	// Verify ECU did not drop out
	if ( VerifyLinkActive () != PASS )
	{
		if ( Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT,
		           "Problems verifying link active.\n" ) == 'N' )
		{
			return FAIL;
		}
	}

	// Determine size of PIDs $F406, $F407, $F408, $F409
	if ( DetermineVariablePidSize () != PASS )
	{
		if ( Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT,
		           "Problems determining the size of variable PIDs.\n" ) == 'N' )
		{
			return FAIL;
		}
	}

	if ( GetPID4FArray () != PASS )
	{
		if ( Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT,
		           "Problems getting PID $F44F.\n" ) == 'N' )
		{
			return FAIL;
		}
	}

	if ( GetPID50Array () != PASS )
	{
		if ( Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, YES_NO_ALL_PROMPT,
		           "Problems getting PID $F450.\n" ) == 'N' )
		{
			return FAIL;
		}
	}


	// For each PID
	for ( IdIndex = 0xF401; IdIndex <= (0XF400 + MAX_PIDS); IdIndex++ )
	{
		// clear PID supported indicator
		fPidSupported[IdIndex-0xF400] = FALSE;

		// skip PID supported PIDs
		if ( IdIndex == 0xF420 || IdIndex == 0xF440 || IdIndex == 0xF460 || IdIndex == 0xF480 ||
		     IdIndex == 0xF4A0 || IdIndex == 0xF4C0 || IdIndex == 0xF4E0 || IdIndex == 0xF500 ||
		     IdIndex == 0xF520 || IdIndex == 0xF540 || IdIndex == 0xF560 || IdIndex == 0xF580 ||
		     IdIndex == 0xF5A0 || IdIndex == 0xF5C0 || IdIndex == 0xF5E0 )
		{
			continue;
		}

		if ( IsDIDSupported ( -1, IdIndex ) == FALSE )
		{
			continue;
		}

		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = IdIndex;
		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $%04X request\n",
			      IdIndex );
			continue;
		}

		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( gOBDResponse[EcuIndex].PidSize != 0 )
			{
				break;
			}
		}

		if ( EcuIndex >= gOBDNumEcus )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "No PID $%04X data\n",
			      IdIndex );
			continue;
		}

		// get Failure Count to allow for FAILURE checks from this point on
		ulTemp_FailureCount = GetFailureCount ();

		// Verify that all PID data is valid
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			// Set ECU dependent PID flags
			fPid13Supported = IsDIDSupported ( EcuIndex, 0xF413 );
			fPid1DSupported = IsDIDSupported ( EcuIndex, 0xF41D );
			fPid4FSupported = IsDIDSupported ( EcuIndex, 0xF44F );
			fPid50Supported = IsDIDSupported ( EcuIndex, 0xF450 );

			// If PID is supported, check it
			if ( IsDIDSupported ( EcuIndex, IdIndex ) == TRUE )
			{
				// Check the data to see if it is valid
				pPid = (PID *)&gOBDResponse[EcuIndex].Pid[0];
				if ( gOBDResponse[EcuIndex].PidSize == 0 )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  PID $%04X supported but no data\n",
					      GetEcuId ( EcuIndex ),
					      IdIndex );
				}

				else
				{
					// save indication that PID is supported
					fPidSupported[IdIndex-0xF400] = TRUE;

					if ( pPid[0].PIDMSB == 0xF4 )
					{
						// Check various PID values for validity based on vehicle state
						switch ( pPid[0].PIDLSB )
						{
							case 0x01:
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  ECUs on a UDS vehicle should not respond to a PID $F401 requests.\n",
								      GetEcuId ( EcuIndex ) );

								// Capture the response from PID $01!
								// for use with PID $41 and
								// MID verification logic
								memcpy ( &gEcuData[EcuIndex].PID01, &pPid[0], sizeof ( PID ) );

								if ( OBDEngineDontCare == FALSE ) // NOT Test 10.2
								{
									if ( VerifyPID01 ( pPid, EcuIndex, &ulIMBitsForVehicle ) != PASS )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "PID $F401 failures detected.\n" );
									}
								}
							}
							break;
							case 0x02:
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  ECUs on a UDS vehicle should not respond to a PID $F402 requests.\n",
								      GetEcuId ( EcuIndex ) );

								temp_data_long = (pPid[0].Data[0] * 256) +
								                  pPid[0].Data[1];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F402  DTCFRZF  = $%04X\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == FALSE )  // test 5.6 & 5.10
								{
									if ( pPid[0].Data[0] != 0x0 ||
									     pPid[0].Data[1] != 0x00 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  DTCFRZF = $%04X, Freeze frames available\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long );
									}
								}
							}
							break;
							case 0x03:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F403  FUELSYSA = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// Fuel System A Status bits
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL          = %s\n",
								      (pPid[0].Data[0] & 0x01) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    CL          = %s\n",
								      (pPid[0].Data[0] & 0x02) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL-Drive    = %s\n",
								      (pPid[0].Data[0] & 0x04) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL-Fault    = %s\n",
								      (pPid[0].Data[0] & 0x08) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    CL-Fault    = %s\n",
								      (pPid[0].Data[0] & 0x10) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL_B2       = %s\n",
								      (pPid[0].Data[0] & 0x20) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL-Drive_B2 = %s\n",
								      (pPid[0].Data[0] & 0x40) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL-Fault_B2 = %s\n",
								      (pPid[0].Data[0] & 0x80) ? "YES" : "NO" );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F403  FUELSYSB = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[1] );
								
								// Fuel System B Status bits
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL          = %s\n",
								      (pPid[0].Data[0] & 0x01) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    CL          = %s\n",
								      (pPid[0].Data[0] & 0x02) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL-Drive    = %s\n",
								      (pPid[0].Data[0] & 0x04) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL-Fault    = %s\n",
								      (pPid[0].Data[0] & 0x08) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    CL-Fault    = %s\n",
								      (pPid[0].Data[0] & 0x10) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL_B2       = %s\n",
								      (pPid[0].Data[0] & 0x20) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL-Drive_B2 = %s\n",
								      (pPid[0].Data[0] & 0x40) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    OL-Fault_B2 = %s\n",
								      (pPid[0].Data[0] & 0x80) ? "YES" : "NO" );


								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == FALSE &&  // if Test 5.6  AND
								     ( (pPid[0].Data[0] & 0x02) != 0 ||                   // FUELSYSA Closed Loop  OR
								       (pPid[0].Data[1] & 0x02) != 0 ) )                  // FUELSYSB Closed Loop
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FUELSYSA Bit 1 and FUELSYSB Bit 1 must both be 0 (not Closed Loop)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x04:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F404  LOAD_PCT = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
									{
										if ( gOBDHybridFlag != TRUE )
										{
											// don't range check hybrids at idle, as LOAD_PCT could be up to 100%
											if ( temp_data_float == 0 || temp_data_float > 60 )
											{
												// Load should be 60% or less with engine running
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  LOAD_PCT should be 60 %% or less with engine running\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
									else  // test 5.6
									{
										if ( temp_data_float != 0 )
										{
											// There should be no load with the engine OFF
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  LOAD_PCT > 0 %% (There should be no load with the engine OFF)\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x05:
							{
								temp_data_long = pPid[0].Data[0] - 40;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F405  ECT      = %d C\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( temp_data_long < -20 || temp_data_long > 120 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  ECT outside expected range (-20 to 120)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x06:
							{
								temp_data_float = (float)(pPid[0].Data[0]) * (float)(100.0/128.0) - (float)100.0;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F406  SHRTFT1  = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
								{
									if ( temp_data_float < -50.0 || temp_data_float > 50.0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  SHRTFT1 outside expected range (-50 to 50)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( gSid1VariablePidSize == 2 )
								{
									temp_data_float = (float)(pPid[0].Data[1]) * (float)(100.0/128.0) - (float)100.0;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F406  SHRTFT3  = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
									{
										if ( temp_data_float < -50.0 || temp_data_float > 50.0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  SHRTFT3 outside expected range (-50 to 50)\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}

							}
							break;
							case 0x07:
							{
								temp_data_float = (float)(pPid[0].Data[0]) * (float)(100.0/128.0) - (float)100.0;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F407  LONGFT1  = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
								{
									if ( temp_data_float < -50.0 || temp_data_float > 50.0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  LONGFT1 outside expected range (-50 to 50)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( gSid1VariablePidSize == 2 )
								{
									temp_data_float = (float)(pPid[0].Data[1]) * (float)(100.0/128.0) - (float)100.0;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F407  LONGFT3  = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
									{
										if ( temp_data_float < -50.0 || temp_data_float > 50.0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  LONGFT3 outside expected range (-50 to 50)\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}

							}
							break;
							case 0x08:
							{
								temp_data_float = (float)(pPid[0].Data[0]) * (float)(100.0/128.0) - (float)100.0;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F408  SHRTFT2  = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
								{
									if ( temp_data_float < -50.0 || temp_data_float > 50.0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  SHRTFT2 outside expected range (-50 to 50)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( gSid1VariablePidSize == 2 )
								{
									temp_data_float = (float)(pPid[0].Data[1]) * (float)(100.0/128.0) - (float)100.0;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F407  SHRTFT4  = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
									{
										if ( temp_data_float < -50.0 || temp_data_float > 50.0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  SHRTFT4 outside expected range (-50 to 50)\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x09:
							{
								temp_data_float = (float)(pPid[0].Data[0]) * (float)(100.0/128.0) - (float)100.0;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F409  LONGFT2  = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
								{
									if ( temp_data_float < -50.0 || temp_data_float > 50.0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  LONGFT2 outside expected range (-50 to 50)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( gSid1VariablePidSize == 2 )
								{
									temp_data_float = (float)(pPid[0].Data[1]) * (float)(100.0/128.0) - (float)100.0;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F407  LONGFT4  = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
									{
										if ( temp_data_float < -50.0 || temp_data_float > 50.0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  LONGFT4 outside expected range (-50 to 50)\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x0A:
							{
								temp_data_long = pPid[0].Data[0] * 3;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F40A  FRP      = %u kPa\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
								{
									if ( temp_data_long == 0 && gOBDHybridFlag != TRUE )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FRP must be greater than 0\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;
							case 0x0B:
							{
								temp_data_float = (float)(pPid[0].Data[0]);
								if ( fPid4FSupported == TRUE && Pid4F[EcuIndex][3]!=0 )
								{
									temp_data_float = temp_data_float * ((float)(Pid4F[EcuIndex][3]*10)/255);
								}
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F40B  MAP      = %.1f kPa\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
								{
									if ( temp_data_float == 0.0 && gOBDHybridFlag != TRUE )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  MAP must be greater than 0\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;
							case 0x0C:
							{
								temp_data_long = (((pPid[0].Data[0] * 256) + pPid[0].Data[1]) / 4);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F40C  RPM      = %d rpm\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
									{
										// if idle RPM is outside the reasonable range
										if ( temp_data_long > 2000 || temp_data_long < 300 )
										{
											// Gas, Deisel vehicle
											if ( gOBDHybridFlag == FALSE )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  RPM outside expected range (300 - 2000)\n",
												      GetEcuId ( EcuIndex ) );
											}

											// HEV, PHEV, S/S vehicle
											else if ( temp_data_long != 0 )
											{
												if (  temp_data_long >= 1 && temp_data_long < 300  )
												{
													Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
													      "ECU %X  RPM outside expected range (1 - 300)\n",
													      GetEcuId ( EcuIndex ) );
												}
												else
												{
													Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
													      "ECU %X  RPM outside expected range (1 - 300)\n",
													      GetEcuId ( EcuIndex ) );
												}
											}
										}
									}
									else  // test 5.6
									{
										if ( temp_data_long != 0 )
										{
											// There should be no RPM with the engine OFF
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  RPM > 0 rpm (There should be no RPM with the engine OFF)\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x0D:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F40D  VSS      = %d km/h\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								if ( pPid[0].Data[0] != 0x00 )
								{
									// There should be no vehicle speed when not moving
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  VSS > 0 km/h (There should be no vehicle speed when not moving)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x0E:
							{
								temp_data_float = ((float)pPid[0].Data[0] / (float)2.0) - (float)64.0;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F40E  SPARKADV = %.1f deg\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
								{
									if ( temp_data_float < -35 || temp_data_float > 55 )
									{
										if ( gOBDHybridFlag == FALSE )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  SPARKADV out of range (-35 to 55)\n",
											      GetEcuId ( EcuIndex ) );
										}

										// HEV, PHEV, S/S vehicle
										else
										{
											Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  SPARKADV out of range (-35 to 55)\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x0F:
							{
								temp_data_long = pPid[0].Data[0] - 40;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F40F  IAT      = %d C\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

 								if ( temp_data_long < -20 || temp_data_long > 120 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  IAT out of range (-20 to 120)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x10:
							{
								temp_data_float = (float)((pPid[0].Data[0] * 256) +
								                           pPid[0].Data[1]);
								if ( fPid50Supported == TRUE && Pid50[EcuIndex][0] != 0 )
								{
									temp_data_float = temp_data_float * ((float)(Pid50[EcuIndex][0] * 10) / 65535);
								}
								else
								{
									temp_data_float = temp_data_float / (float)100.0;
								}
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F410  MAF      = %.2f g/s\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
									{
										if ( temp_data_float == 0.0 && gOBDHybridFlag != TRUE )
										{
											// MAF should not be zero with the engine running
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  MAF should not be zero with the engine running\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else  // test 5.6, engine off
									{
										if ( temp_data_float > 5.0 )
										{
											// MAF should be zero with the engine OFF
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  MAF should be zero with the engine OFF\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x11:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F411  TP       = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDDieselFlag == FALSE )
								{
									// non-diesel
									if ( temp_data_float > 50 )
									{
										// Throttle position should be 40% or less when not driving
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TP > 50 %% (should be 40 %% or less when not driving)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
								else
								{
									// diesel
									if ( temp_data_float > 100 )
									{
										// Throttle position should be 100% or less when not driving
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TP > 100 %% (should be 100 %% or less when not driving)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x12:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F412  AIR_STAT = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );
							}
							break;

							/*
							case 0x13:  See 0x1D
							*/

							case 0x14:
							case 0x15:
							case 0x16:
							case 0x17:
							case 0x18:
							case 0x19:
							case 0x1A:
							case 0x1B:
							{
								switch ( pPid[0].PIDLSB )
								{
									case 0x14:
										BSIndex = 0x11;
									break;
									case 0x15:
										BSIndex = 0x12;
									break;
									case 0x16:
										BSIndex = fPid13Supported ? 0x13 : 0x21;
									break;
									case 0x17:
										BSIndex = fPid13Supported ? 0x14 : 0x22;
									break;
									case 0x18:
										BSIndex = fPid13Supported ? 0x21 : 0x31;
									break;
									case 0x19:
										BSIndex = fPid13Supported ? 0x22 : 0x32;
									break;
									case 0x1A:
										BSIndex = fPid13Supported ? 0x23 : 0x41;
									break;
									case 0x1B:
										BSIndex = fPid13Supported ? 0x24 : 0x42;
									break;
								}

								temp_data_float = (float)pPid[0].Data[0] * (float).005;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $%04X  O2S%x    = %.3f V\n",
								      GetEcuId ( EcuIndex ),
								      IdIndex,
								      BSIndex,
								      temp_data_float );

								temp_data_float = ((float)(pPid[0].Data[1]) * (float)(100.0 / 128.0)) - (float)100.0;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $%04X  SHRTFT%x = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      IdIndex,
								      BSIndex,
								      temp_data_float );
							}	  
							break;

							case 0x1C:
							{
								if ( VerifyPID1C ( pPid, EcuIndex ) != PASS )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "PID $F41C failures detected.\n" );
								}
							}
							break;

							case 0x13:
							case 0x1D:
							{
								unsigned short O2Bit;
								unsigned short O2Count;

								// Identify support for PID 0x13 / 0x1D
								if ( pPid[0].PIDLSB == 0x13 )
								{
									fPid13Supported = TRUE;
								}
								else
								{
									fPid1DSupported = TRUE;
								}

								// Evaluate for dual PID / Spark engine support
								if ( fPid13Supported == TRUE &&
								     fPid1DSupported == TRUE )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F413 & $F41D both indicated as supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// Count the number of O2 sensors
								for ( O2Bit = 0x01, O2Count = 0; O2Bit != 0x100; O2Bit = O2Bit << 1 )
								{
									if ( pPid[0].Data[0] & O2Bit )
									{
										O2Count++;
									}
								}
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $%04X  O2SLOC   = $%02X (%d O2 Sensors)\n",
								      GetEcuId ( EcuIndex ),
								      IdIndex,
								      pPid[0].Data[0],
								      O2Count );

								// At least 2 O2 sensors required for spark ignition enges
								if ( gOBDDieselFlag == FALSE &&
								     O2Count < 2 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  O2S < 2 (At least 2 O2 sensors required for spark ignition enges)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x1E:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F41E  PTO_STAT = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );
							}
							break;

							case 0x1F:
							{
								temp_data_long = ((pPid[0].Data[0] * 256) +
								                   pPid[0].Data[1]);
								GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F41F  RUNTM    = %d sec (%d hrs, %d mins, %d secs)\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long,
								      hours, mins, secs );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineRunning == TRUE )
									{
										if ( gOBDEngineWarm == TRUE )  // test 10.13
										{
											if ( temp_data_long <= 300 )
											{
												// Run time should be greater than 300 seconds
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  RUNTM should be greater than 300 seconds\n",
												      GetEcuId ( EcuIndex ),
												      temp_data_long );
											}
										}
										else  // test 5.10
										{
											if ( temp_data_long == 0 && gOBDHybridFlag != TRUE )
											{
												// Run time should not be zero if engine is running
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  RUNTM should not be zero if engine is running\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
									else  // test 5.6
									{
										if ( temp_data_long != 0 )
										{
											// Run time should be zero if engine is OFF
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  RUNTM should be zero if engine is OFF\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x21:
							{
								temp_data_long = ((pPid[0].Data[0] * 256) +
								                   pPid[0].Data[1]);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F421  MIL_DIST = %d km\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( ( gOBDEngineWarm == TRUE     ||            // test 10.12
									       gOBDEngineRunning == FALSE ||            // test 5.6
									       ( gOBDEngineRunning == TRUE  &&          // test 5.10
									         gOBDEngineWarm == FALSE    &&
									         gOBDResponse[EcuIndex].ClrDTCSize > 0 &&
									         gOBDResponse[EcuIndex].ClrDTCData[0] == 0x54 ) ) &&
									     temp_data_long != 0x00 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  MIL_DIST > 0 after clearing DTCs\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;
							case 0x22:
							{
								temp_data_float = (float)((pPid[0].Data[0] * 256) +
								                           pPid[0].Data[1]) * (float)0.079;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F422  FRP (relative to manifold) = %.1f kPa\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE &&  // NOT Test 10.2
								     gOBDEngineRunning == TRUE  &&  // test 5.10 & 10.13
								     gOBDHybridFlag != TRUE     &&  // not a hybrid vehicle
								     temp_data_float == 0.0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FRP should be greater than 0\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x23:
							{
								temp_data_long = ((pPid[0].Data[0] * 256) +
								                   pPid[0].Data[1]) * 10;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F423  FRP      = %ld kPa\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE &&  // NOT Test 10.2
								     gOBDEngineRunning == TRUE  &&  // test 5.10 & 10.12
								     gOBDHybridFlag != TRUE     &&  // not a hybrid vehicle
								     temp_data_long == 0  )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FRP should be greater than 0\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x24:
							case 0x25:
							case 0x26:
							case 0x27:
							case 0x28:
							case 0x29:
							case 0x2A:
							case 0x2B:
							{
								switch ( pPid[0].PIDLSB )
								{
									case 0x24:
										BSIndex = 0x11;
									break;
									case 0x25:
										BSIndex = 0x12;
									break;
									case 0x26:
										BSIndex = fPid13Supported ? 0x13 : 0x21;
									break;
									case 0x27:
										BSIndex = fPid13Supported ? 0x14 : 0x22;
									break;
									case 0x28:
										BSIndex = fPid13Supported ? 0x21 : 0x31;
									break;
									case 0x29:
										BSIndex = fPid13Supported ? 0x22 : 0x32;
									break;
									case 0x2A:
										BSIndex = fPid13Supported ? 0x23 : 0x41;
									break;
									case 0x2B:
										BSIndex = fPid13Supported ? 0x24 : 0x42;
									break;
								}

								temp_data_float = (float)(((unsigned long)(pPid[0].Data[0] << 8) |
								                                           pPid[0].Data[1]));
								if ( fPid4FSupported == FALSE || Pid4F[EcuIndex][0] == 0 )
								{
									temp_data_float = temp_data_float * (float)0.0000305;
								}
								else
								{
									temp_data_float = temp_data_float * ((float)(Pid4F[EcuIndex][0])/65535);
								}
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $%04X  EQ_RAT%x = %.3f %%\n",
								      GetEcuId ( EcuIndex ),
								      IdIndex,
								      BSIndex,
								      temp_data_float );

								temp_data_float = (float)(((unsigned long)(pPid[0].Data[2] << 8) |
								                                           pPid[0].Data[3]));
								if ( fPid4FSupported == FALSE || Pid4F[EcuIndex][1] == 0 )
								{
									temp_data_float = temp_data_float * (float)0.000122;
								}
								else
								{
									temp_data_float = temp_data_float * ((float)(Pid4F[EcuIndex][1])/65535);
								}
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $%04X  O2S%x    = %.3f V\n",
								      GetEcuId ( EcuIndex ),
								      IdIndex,
								      BSIndex,
								      temp_data_float );
							}
							break;
							case 0x2C:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F42C  EGR_PCT  = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDEngineRunning == FALSE )  // test 5.6
								{
									if ( temp_data_float > 10 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  EGR_PCT should be less than 10 %%\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
								else  // test 5.10 & 10.13
								{
									if ( gOBDDieselFlag == TRUE )  // compression ignition
									{
										if ( temp_data_float > 100)
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  EGR_PCT should be less than 100 %%\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else  // spark ignition engine
									{
										if ( gOBDHybridFlag == FALSE && temp_data_float > 10 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  EGR_PCT should be less than 10 %%\n",
											      GetEcuId ( EcuIndex ) );
										}
										else if ( gOBDHybridFlag == TRUE && temp_data_float > 50 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  EGR_PCT should be less than 50 %%\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;

							case 0x2D:
							{
								temp_data_float = ( (float)pPid[0].Data[0] * (float)(100.0/128.0) ) - (float)100.0;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F42D  EGR_ERR  = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
							}
							break;

							case 0x2E:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0/255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F42E  EVAP_PCT = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
							}
							break;

							case 0x2F:
							{
								temp_data_float = ((float)pPid[0].Data[0]) * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F42F  FLI      = %.1f\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( temp_data_float < 1.0 || temp_data_float > 100.0 )
								{
									Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FLI should be between 1 and 100 %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x30:
							{
								temp_data_long = pPid[0].Data[0];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F430  WARM_UPS = %d\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineWarm == TRUE )  // test 10.13
									{
										if ( temp_data_long > 4)
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  WARM_UPS should be less than or equal to 4\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else if ( gOBDEngineRunning == FALSE )          // test 5.6
									{
										if ( temp_data_long > 0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  WARM_UPS should be less than or equal to 0\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else if ( gOBDEngineRunning == TRUE  &&       // test 5.10
									          gOBDResponse[EcuIndex].ClrDTCSize > 0 &&
									          gOBDResponse[EcuIndex].ClrDTCData[0] == 0x54 )
									{
										if ( temp_data_long > 1 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  WARM_UPS should be less than or equal to 1\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x31:
							{
								temp_data_long = ((pPid[0].Data[0] * 256) +
								                   pPid[0].Data[1]);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F431  CLR_DIST = %d km\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineWarm == TRUE )                     // test 10.13
									{
										if ( temp_data_long >= 50 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  CLR_DIST should be less than 50 km after CARB drive cycle\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else if ( gOBDEngineRunning == FALSE ||       // test 5.6
									          (gOBDEngineRunning == TRUE &&       // test 5.10
									           gOBDResponse[EcuIndex].ClrDTCSize > 0 &&
									           gOBDResponse[EcuIndex].ClrDTCData[0] == 0x54) )
									{
										if ( temp_data_long != 0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  CLR_DIST should be 0 km after clearing DTCs\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;

							case 0x32:
							{
								temp_data_float = (float)( (signed short)( (pPid[0].Data[0] * 256) +
								                                            pPid[0].Data[1] ) ) * (float).25;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F432  EVAP_VP  = %.1f Pa\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
							}
							break;

							case 0x33:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F433  BARO     = %d kPa\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								if ( (pPid[0].Data[0] < 71) || (pPid[0].Data[0] > 110))
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  BARO outside expected range (71 - 110)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x34:
							case 0x35:
							case 0x36:
							case 0x37:
							case 0x38:
							case 0x39:
							case 0x3A:
							case 0x3B:
							{
								switch ( pPid[0].PIDLSB )
								{
									case 0x34:
										BSIndex = 0x11;
									break;
									case 0x35:
										BSIndex = 0x12;
									break;
									case 0x36:
										BSIndex = fPid13Supported ? 0x13 : 0x21;
									break;
									case 0x37:
										BSIndex = fPid13Supported ? 0x14 : 0x22;
									break;
									case 0x38:
										BSIndex = fPid13Supported ? 0x21 : 0x31;
									break;
									case 0x39:
										BSIndex = fPid13Supported ? 0x22 : 0x32;
									break;
									case 0x3A:
										BSIndex = fPid13Supported ? 0x23 : 0x41;
									break;
									case 0x3B:
										BSIndex = fPid13Supported ? 0x24 : 0x42;
									break;
								}

								// scale Equivalence Ratio
								temp_data_float = (float)(((unsigned long)(pPid[0].Data[0] << 8) |
								                                           pPid[0].Data[1]));
								if ( fPid4FSupported == FALSE || Pid4F[EcuIndex][0] == 0 )
								{
									temp_data_float = temp_data_float * (float)0.0000305;
								}
								else
								{
									temp_data_float = temp_data_float * ((float)(Pid4F[EcuIndex][0])/65535);
								}
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $%04X  EQ_RAT%x = %.3f %%\n",
								      GetEcuId ( EcuIndex ),
								      IdIndex,
								      BSIndex,
								      temp_data_float );

								// offset and scale Oxygen Sensor Current
								temp_data_float = (float)((((signed short)(pPid[0].Data[2]) << 8) |
								                                           pPid[0].Data[3]) - 0x8000);
								if ( fPid4FSupported == FALSE || Pid4F[EcuIndex][2] == 0 )
								{
									temp_data_float = temp_data_float * (float)0.00390625;
								}
								else
								{
									temp_data_float = temp_data_float * ((float)(Pid4F[EcuIndex][2])/32768);
								}
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $%04X  O2S%x    = %.2f mA\n",
								      GetEcuId ( EcuIndex ),
								      IdIndex,
								      BSIndex,
								      temp_data_float );
							}
							break;

							case 0x3C:
							case 0x3D:
							case 0x3E:
							case 0x3F:
							{
								switch ( pPid[0].PIDLSB )
								{
									case 0x3C:
										BSIndex = 0x11;
									break;
									case 0x3D:
										BSIndex = 0x21;
									break;
									case 0x3E:
										BSIndex = 0x12;
									break;
									case 0x3F:
										BSIndex = 0x22;
									break;
								}
								temp_data_float = ( (float)( (pPid[0].Data[0] * 256) +
								                              pPid[0].Data[1] ) * (float).1 ) - (float)(40.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $%04X  CATEMP%X = %.1f C\n",
								      GetEcuId ( EcuIndex ),
								      IdIndex,
								      BSIndex,
								      temp_data_float );
							}
							break;

							case 0x41:
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  ECUs on a UDS vehicle should not respond to a PID $F441 requests.\n",
								      GetEcuId ( EcuIndex ) );

								if ( VerifyPID41 ( pPid, EcuIndex, OBDEngineDontCare ) != PASS )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "PID $F441 failures detected.\n" );
								}
							}
							break;

							case 0x42:
							{
								temp_data_float = (float)( (pPid[0].Data[0] * 256) +
								                            pPid[0].Data[1] ) * (float).001;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F442  VPWR     = %.3f V\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
							}
							break;

							case 0x43:
							{
								temp_data_float = (float)( (pPid[0].Data[0] * 256) +
								                            pPid[0].Data[1] ) * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F443  LOAD_ABS = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
									{
										if ( (temp_data_float == 0.0) && (gOBDHybridFlag != TRUE) )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  LOAD_ABS should be greater than 0 %% when engine is running\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else  // test 5.6
									{
										if ( temp_data_float != 0.0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  LOAD_ABS should be 0 %% when engine is not running\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;
							case 0x44:
							{
								temp_data_float = (float)(((unsigned long)(pPid[0].Data[0]) << 8) |
								                                           pPid[0].Data[1]);
								if ( fPid4FSupported == FALSE || Pid4F[EcuIndex][0] == 0 )
								{
									temp_data_float = temp_data_float * (float)0.0000305;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F444  EQ_RAT   = %.3f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDEngineWarm == TRUE )  // test 10.13
									{
										if ( gOBDDieselFlag == FALSE && gOBDHybridFlag != TRUE )
										{
											if ( temp_data_float < 0.5 || temp_data_float > 1.5  )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EQ_RAT outside expected range (0.5 - 1.5)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}
								else
								{
									temp_data_float = temp_data_float * ((float)(Pid4F[EcuIndex][0])/65535);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F444  EQ_RAT   = %.3f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;
							case 0x45:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F445  TP_R     = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDDieselFlag == FALSE && gOBDHybridFlag == FALSE && temp_data_float > 50 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TP_R should be less than 50 %%\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x46:
							{
								temp_data_long = pPid[0].Data[0] - 40;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F446  AAT      = %d C\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( temp_data_long < -20 || temp_data_long > 85 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  AAT outside expected range (-20 to 85)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x47:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F447  TP_B     = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDDieselFlag == FALSE && gOBDHybridFlag == FALSE && temp_data_float > 60 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TP_B should be less than 60 %%\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x48:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F448  TP_C     = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDDieselFlag == FALSE && gOBDHybridFlag == FALSE && temp_data_float > 60 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TP_C should be less than 60 %%\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x49:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F449  APP_D    = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( temp_data_float > 40 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  APP_D should be less than 40 %%\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x4A:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44A  APP_E    = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( temp_data_float > 40 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  APP_E should be less than 40 %%\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x4B:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44B  APP_F    = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( temp_data_float > 40 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  APP_F should be less than 40 %%\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x4C:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44C  TAC_PCT  = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
							}
							break;

							case 0x4D:
							{
								temp_data_long = ( pPid[0].Data[0] * 256 ) +
								                   pPid[0].Data[1];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44D  MIL_TIME = %d min\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( temp_data_long != 0x00 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  MIL_TIME should be 0\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;
							case 0x4E:
							{
								temp_data_long = ( pPid[0].Data[0] * 256 ) +
								                   pPid[0].Data[1];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44E  CLR_TIME = %d min\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineWarm == TRUE )                   // test 10.13
									{
										if ( temp_data_long < 5 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  CLR_TIME should be greater than 5 min\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else if ( gOBDEngineRunning == FALSE ||       // test 5.6
									          ( gOBDEngineRunning == TRUE  &&     // test 5.10
									            gOBDResponse[EcuIndex].ClrDTCSize > 0 &&
									            gOBDResponse[EcuIndex].ClrDTCData[0] == 0x54 ) )
									{
										if ( temp_data_long != 0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  CLR_TIME should be 0 min\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;

							case 0x4F:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44F  EQ_RAT scaling      = %d / 65535\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44F  O2S Voltage scaling = %d * (1 V / 65535)\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[1] );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44F  O2S Current scaling = %d * (1 mA / 32768)\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[2] );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44F  MAP scaling         = %d * (10 kPa / 255)\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[3] );
							}
							break;

							case 0x50:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F450  MAF scaling         = %d * (10 g/s / 65535)\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );
							}
							break;

							case 0x51:
							{
								temp_data_long = pPid[0].Data[0];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F451  FUEL_TYP = $%02X (%s)\n",
								      GetEcuId ( EcuIndex ),
								      pPid->Data[0],
								      (pPid->Data[0] <= 0x1F ? FUEL_TYPE[pPid->Data[0]] : FUEL_TYPE[0x20]) );
				
								if ( temp_data_long == 0x00 ||
								     (temp_data_long >= 0x0F && temp_data_long <= 0x17) )
								{
									Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FUEL_TYPE must be $01 to $0E or $18 to $1C\n",
									      GetEcuId ( EcuIndex ) );
								}
								else if ( temp_data_long >= 0x1D )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FUEL_TYPE must be $01 to $0E or $18 to $1C\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x52:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F452  ALCH_PCT = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
							}
							break;

							case 0x53:
							{
								temp_data_float = ( (pPid[0].Data[0] * 256) +
								                     pPid[0].Data[1] ) * (float)0.005;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F453  EVAP_VPA = %.3f kPa\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
							}
							break;

							case 0x54:
							{
								temp_data_long = (long)((signed short)( (pPid[0].Data[0] * 256) +
								                                         pPid[0].Data[1] ) );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F454  EVAP_VP  = %d Pa\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );
							}
							break;

							case 0x55:
							{
								temp_data_float = (float)( pPid[0].Data[0]  - 128 ) * (float)(100.0 / 128.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F455  STSO2FT1 = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
								if ( temp_data_float < -100.0 || temp_data_float > 99.2 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  STSO2FT1 out of range (-100 to 99.2)\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( gSid1VariablePidSize == 2 )
								{
									temp_data_float =  (float)( pPid[0].Data[1] - 128 ) * (float)(100.0 / 128.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F455  STSO2FT3 = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
									if ( temp_data_float < -100.0 || temp_data_float > 99.2 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  STSO2FT3 out of range (-100 to 99.2)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x56:
							{
								temp_data_float = (float)( pPid[0].Data[0] - 128 ) * (float)(100.0 / 128.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F456  LGSO2FT1 = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
								if ( temp_data_float < -100.0 || temp_data_float > 99.2 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  LGSO2FT1 out of range (-100 to 99.2)\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( gSid1VariablePidSize == 2 )
								{
									temp_data_float = (float)( pPid[0].Data[1] - 128 ) * (float)(100.0 / 128.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F456  LGSO2FT3 = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
									if ( temp_data_float < -100.0 || temp_data_float > 99.2 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  LGSO2FT3 out of range (-100 to 99.2)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x57:
							{
								temp_data_float = (float)( pPid[0].Data[0] - 128 ) * (float)(100.0 / 128.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F457  STSO2FT2 = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
								if ( temp_data_float < -100.0 || temp_data_float > 99.2 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  STSO2FT2 out of range (-100 to 99.2)\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( gSid1VariablePidSize == 2 )
								{
									temp_data_float = (float)( pPid[0].Data[1] - 128 ) * (float)(100.0 / 128.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F457  STSO2FT4 = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
									if ( temp_data_float < -100.0 || temp_data_float > 99.2 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  STSO2FT4 out of range (-100 to 99.2)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x58:
							{
								temp_data_float = (float)( pPid[0].Data[0] - 128 ) * (float)(100.0 / 128.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F458  LGSO2FT2 = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
								if ( temp_data_float < -100.0 || temp_data_float > 99.2 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  LGSO2FT2 out of range (-100 to 99.2)\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( gSid1VariablePidSize == 2 )
								{
									temp_data_float = (float)( pPid[0].Data[1] - 128 ) * (float)(100.0 / 128.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F458  LGSO2FT4 = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
									if ( temp_data_float < -100.0 || temp_data_float > 99.2 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  LGSO2FT4 out of range (-100 to 99.2)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x59:
							{
								temp_data_long = ( (pPid[0].Data[0] * 256) +
								                    pPid[0].Data[1] ) * 10;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F459  FRP      = %d kPa\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE && gOBDHybridFlag != TRUE &&
								     temp_data_long == 0 )  // test 5.10 & 10.13
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FRP should not be 0\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x5A:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F45A  APP_R    = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
								if ( temp_data_float > 40 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  APP_R should be less than 40 %%\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x5B:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F45B  BAT_PWR  = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE &&
								     temp_data_float == 0 )  // test 5.10 & 10.13
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  BAT_PWR should not be 0\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x5C:
							{
								temp_data_long = pPid[0].Data[0] - 40;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F45C  EOT      = %d C\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );
								if ( temp_data_long  < -20 || temp_data_long > 150 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EOT out of range (-20 to 150)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x5D:
							{
								temp_data_float = (float)( ( (pPid[0].Data[0] * 256) +
								                             pPid[0].Data[1] ) - 26880 ) / 128;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F45D  FUEL_TIMING = %.2f\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
								if ( temp_data_float < -210 || temp_data_float >= 302 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FUEL_TIMING out of range (-210 to 301)\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x5E:
							{
								temp_data_float = ( (pPid[0].Data[0] * 256) +
								                    pPid[0].Data[1] ) * (float)0.05;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F45E  FUEL_RATE = %.2f L/h\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
									{
										if ( temp_data_float == 0.0 && gOBDHybridFlag != TRUE )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  FUEL_RATE should not be 0 when engine running\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else // test 5.6
									{
										if ( temp_data_float != 0.0 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  FUEL_RATE should be 0 when engine is not running\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;

							case 0x5F:
							{
								temp_data_long = pPid[0].Data[0];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F45F  EMIS_SUP = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );
								if ( temp_data_long  < 0x0E || temp_data_long > 0x10 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EMIS_SUP must be $0E to $10\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x61:
							{
								temp_data_long = pPid[0].Data[0] - 125;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F461  TQ_DD    = %d %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );
							}
							break;

							case 0x62:
							{
								temp_data_long = pPid[0].Data[0] - 125;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F462  TQ_ACT   = %d %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
								{
									if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
									{
										if ( temp_data_long <= 0 )
										{
											if ( !(temp_data_long == 0 && gOBDHybridFlag == TRUE) )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  TQ_ACT out of range\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
									else if ( temp_data_long < 0 )  // test 5.6
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TQ_ACT should not be less than 0 when engine is not running\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x63:
							{
								temp_data_long = (pPid[0].Data[0] * 256) +
								                 pPid[0].Data[1];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F463  TQ_REF   = %d Nm\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE &&
								     temp_data_long <= 0 )  // test 5.10 & 10.13
								{
									if ( !(temp_data_long == 0 && gOBDHybridFlag == TRUE) )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TQ_REF out of range\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x64:
							{
								temp_data_long = pPid[0].Data[0] - 125;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F464  TQ_MAX1  = %d %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
								{
									if ( temp_data_long <= 0 )
									{
										if ( !(temp_data_long == 0 && gOBDHybridFlag == TRUE) )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  TQ_MAX1 out of range\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}

								temp_data_long = pPid[0].Data[1] - 125;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F464  TQ_MAX2  = %d %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
								{
									if ( temp_data_long <= 0 )
									{
										if ( !(temp_data_long == 0 && gOBDHybridFlag == TRUE) )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  TQ_MAX2 out of range\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}

								temp_data_long = pPid[0].Data[2] - 125;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F464  TQ_MAX3  = %d %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
								{
									if ( temp_data_long <= 0 )
									{
										if ( !(temp_data_long == 0 && gOBDHybridFlag == TRUE) )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  TQ_MAX3 out of range\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}

								temp_data_long = pPid[0].Data[3] - 125;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F464  TQ_MAX4  = %d %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
								{
									if ( temp_data_long <= 0 )
									{
										if ( !(temp_data_long == 0 && gOBDHybridFlag == TRUE) )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  TQ_MAX4 out of range\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}

								temp_data_long = pPid[0].Data[4] - 125;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F464  TQ_MAX5  = %d %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
								{
									if ( temp_data_long <= 0 )
									{
										if ( !(temp_data_long == 0 && gOBDHybridFlag == TRUE) )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  TQ_MAX5 out of range\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;

							case 0x65:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F465  Auxiliary I/O support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x1F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Auxiliary I/O  No devices supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( (pPid[0].Data[0] & 0xE0) != 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Auxiliary I/O  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

//								if ( gOBDDieselFlag == TRUE &&
//								     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) &&
//								     (pPid[0].Data[0] & 0x08) == 0 )
//								{
//									Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//									      "ECU %X  Auxiliary I/O A Bit 3 (GPL_STAT_SUPPORTED) must be set for diesel vehicles\n",
//									      GetEcuId ( EcuIndex ) );
//								}


								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F465  Auxiliary I/O Status = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// Auxiliary I/O Status bits
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    PTO_STAT  = %s\n",
								      (pPid[0].Data[0] & 0x01) ? "ON" : "OFF" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    N/D_STAT  = %s\n",
								      (pPid[0].Data[0] & 0x02) ? "DRIVE" : "NEUT" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    N/G_STAT  = %s\n",
								      (pPid[0].Data[0] & 0x04) ? "GEAR" : "NEUT" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    GPL_STAT  = %s\n",
								      (pPid[0].Data[0] & 0x08) ? "ON" : "OFF" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    GEAR_RCMD = %s\n",
								      GEAR[(pPid[0].Data[0])>>4] );
							}
							break;

							case 0x66:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F466  MAF Sensor support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x03) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  MAF Sensor support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFC )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  MAF Sensor support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = (float)( (pPid[0].Data[1] * 256) +
									                           pPid[0].Data[2] ) * (float)0.03125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F466  MAFA     = %.3f g/s\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
									{
										if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
										{
											if ( temp_data_float <= 0 )
											{
												if ( !(temp_data_float == 0.0 && gOBDHybridFlag == TRUE) )
												{
													Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
													      "ECU %X  MAFA out of range\n",
													      GetEcuId ( EcuIndex ) );
												}
											}
										}
										else  // test 5.6
										{
											if ( temp_data_float > 5 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  MAFA should be less than or equal to 5 when engine is not running\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = (float)( (pPid[0].Data[3] * 256) +
									                           pPid[0].Data[4] ) * (float)0.03125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F466  MAFB     = %.3f g/s\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
									{
										if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
										{
											if ( temp_data_float <= 0 )
											{
												if ( !((temp_data_float == 0.0) && (gOBDHybridFlag == TRUE)) )
												{
													Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
													      "ECU %X  MAFB out of range\n",
													      GetEcuId ( EcuIndex ) );
												}
											}
										}
										else  // test 5.6
										{
											if ( temp_data_float > 5 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  MAFB should be less than or equal to 5 when engine is not running\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}
							}
							break;

							case 0x67:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F467  ECT Sensor support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x03) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  ECT Sensor support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFC )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  ECT Sensor support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F467  ECT1     = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  ECT1 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = (pPid[0].Data[2] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F467  ECT2     = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  ECT2 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x68:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F468  IAT Sensor support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x3F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  IAT Sensor support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xC0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  IAT Sensor support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = ( pPid[0].Data[1] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F468  IAT11    = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  IAT11 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = ( pPid[0].Data[2] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F468  IAT12    = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  IAT12 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = (pPid[0].Data[3] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F468  IAT13    = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  IAT13 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = ( pPid[0].Data[4] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F468  IAT21    = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  IAT21 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_long = ( pPid[0].Data[5] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F468  IAT22    = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  IAT22 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x20) != 0 )
								{
									temp_data_long =  (pPid[0].Data[6] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F468  IAT23    = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  IAT23 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x69:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F469  EGR data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x3F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGR data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xC0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGR data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[1] * (float)(100.0 / 255.0);
									temp_EGR_CMD = temp_data_float;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F469  EGR_A_CMD = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE )  // test 5.6
									{
										if ( temp_data_float > 10 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  EGR_A_CMD should be less than or equal to 10 %% when engine is not running\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else  // test 5.10 & 10.13
									{
										if ( gOBDDieselFlag == TRUE )  // compression ignition
										{
											if ( temp_data_float > 100 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_A_CMD should be less than or equal to 100 %% when engine is running\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
										else  // spark ignition
										{
											if ( gOBDHybridFlag == FALSE && temp_data_float > 10 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_A_CMD should be less than or equal to 10 %% when engine is running\n",
												      GetEcuId ( EcuIndex ) );
											}
											else if ( gOBDHybridFlag == TRUE && temp_data_float > 50 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_A_CMD should be less than or equal to 50 %% when engine is running\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[2] * (float)(100.0 / 255.0);
									temp_EGR_ACT = temp_data_float;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F469  EGR_A_ACT = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = (float)(pPid[0].Data[3] - 128 ) * (float)(100.0 / 128.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F469  EGR_A_ERR = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									// if EGR_CMD and EGR_ACT supported, check for proper EGR_ERR
									if ( (pPid[0].Data[0] & 0x01) != 0 && (pPid[0].Data[0] & 0x02) != 0 )
									{
										if ( temp_EGR_CMD == 0 )
										{
											if ( temp_EGR_ACT == 0 && temp_data_float != 0 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_A_ERR should be 0%% when EGR_A_CMD == 0%% and EGR_A_ACT == 0%%\n",
												      GetEcuId ( EcuIndex ) );
											}
											else if ( temp_EGR_ACT > 0 && pPid[0].Data[3] != 0xFF )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_A_ERR should be 99.2%% when EGR_A_CMD == 0%% and EGR_A_ACT > 0%%\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[4] * (float)(100.0 / 255.0);
									temp_EGR_CMD = temp_data_float;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F469  EGR_B_CMD = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE)  // test 5.6
									{
										if ( temp_data_float > 10 )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  EGR_B_CMD should be less than or equal to 10 %% when engine is not runnig\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
									else  // test 5.10 & 10.13
									{
										if ( gOBDDieselFlag == TRUE )  // compression ignition
										{
											if ( temp_data_float > 100 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_B_CMD should be less or equal to 100 %% when engine is running\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
										else  // spark ignition
										{
											if ( gOBDHybridFlag == FALSE && temp_data_float > 10 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_B_CMD should be less than or equal to 10 %% when engine is running\n",
												      GetEcuId ( EcuIndex ) );
											}
											else if ( gOBDHybridFlag == TRUE && temp_data_float > 50 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_B_CMD should be less than or equal to 50 %% when engine is running\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[5] * (float)(100.0 / 255.0);
									temp_EGR_ACT = temp_data_float;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F469  EGR_B_ACT = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x20) != 0 )
								{
									temp_data_float = (float)( pPid[0].Data[6] - 128) * (float)(100.0 / 128.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F469  EGR_B_ERR = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									// if EGR_CMD and EGR_ACT supported, check for proper EGR_ERR
									if ( (pPid[0].Data[0] & 0x08) != 0 && (pPid[0].Data[0] & 0x10) != 0 )
									{
										if ( temp_EGR_CMD == 0 )
										{
											if ( temp_EGR_ACT == 0 && temp_data_float != 0 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_B_ERR should be 0%% when EGR_B_CMD == 0%% and EGR_B_ACT == 0%%\n",
												      GetEcuId ( EcuIndex ) );
											}
											else if ( temp_EGR_ACT > 0 && pPid[0].Data[6] != 0xFF )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGR_B_ERR should be 99.2%% when EGR_B_CMD == 0%% and EGR_B_ACT > 0%%\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}
							}
							break;

							case 0x6A:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F46A  IAF data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  IAF data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  IAF data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[1] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46A  IAF_A_CMD = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[2] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46A  IAF_A_REL = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[3] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46A  IAF_B_CMD = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[4] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46A  IAF_B_REL = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x6B:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F46B  EGRT sensor support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0xFF) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGRT sensor support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that only one range bit is set for each sensor
								temp_data_long = pPid[0].Data[0] >> 4;
								if ( (pPid[0].Data[0] & temp_data_long ) != 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGRT sensor support  Both Normal and Wide Range supported for sensor(s)\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x11) != 0 )
								{
									if ( (pPid[0].Data[0] & 0x01) != 0 )
									{
										temp_data_long = (pPid[0].Data[1]) - 40;
									}
									else
									{
										temp_data_long = ((pPid[0].Data[1]) * 4) - 40;
									}

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46B  EGRT11   = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );

									if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
									{
										if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
										{
											if ( ( ((pPid[0].Data[0] & 0x01) != 0) && (temp_data_long < -20 || temp_data_long > 215) ) ||
											     ( ((pPid[0].Data[0] & 0x10) != 0) && (temp_data_long < -20 || temp_data_long > 500) ) )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGRT11 out of range (-20 to 215 or 500)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
										else  // test 5.6
										{
											if ( temp_data_long < -20 || temp_data_long > 200 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGRT11 out of range (-20 to 200)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}

								if ( (pPid[0].Data[0] & 0x22) != 0 )
								{
									if ( (pPid[0].Data[0] & 0x02) != 0 )
									{
										temp_data_long = (pPid[0].Data[2]) - 40;
									}
									else
									{
										temp_data_long = ((pPid[0].Data[2]) * 4) - 40;
									}

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46B  EGRT12   = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );

									if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
									{
										if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
										{
											if ( ( ((pPid[0].Data[0] & 0x02) != 0) && (temp_data_long < -20 || temp_data_long > 215) ) ||
											     ( ((pPid[0].Data[0] & 0x20) != 0) && (temp_data_long < -20 || temp_data_long > 500) ) )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGRT12 out of range (-20 to 215 or 500)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
										else  // test 5.6
										{
											if ( temp_data_long < -20 || temp_data_long > 200 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGRT12 out of range (-20 to 200)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}

								if ( (pPid[0].Data[0] & 0x44) != 0 )
								{
									if ( (pPid[0].Data[0] & 0x04) != 0 )
									{
										temp_data_long = (pPid[0].Data[3]) - 40;
									}
									else
									{
										temp_data_long = ((pPid[0].Data[3]) * 4) - 40;
									}

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46B  EGRT21   = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );

									if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
									{
										if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
										{
											if ( ( ((pPid[0].Data[0] & 0x04) != 0) && (temp_data_long < -20 || temp_data_long > 215) ) ||
											     ( ((pPid[0].Data[0] & 0x40) != 0) && (temp_data_long < -20 || temp_data_long > 500) ) )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGRT21 out of range (-20 to 215 or 500)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
										else  // test 5.6
										{
											if ( temp_data_long < -20 || temp_data_long > 200 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGRT21 out of range (-20 to 200)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}

								if ( (pPid[0].Data[0] & 0x88) != 0 )
								{
									if ( (pPid[0].Data[0] & 0x08) != 0 )
									{
										temp_data_long = (pPid[0].Data[4]) - 40;
									}
									else
									{
										temp_data_long = ((pPid[0].Data[4]) * 4) - 40;
									}

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46B  EGRT22   = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );

									if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
									{
										if ( gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
										{
											if ( ( ((pPid[0].Data[0] & 0x08) != 0) && (temp_data_long < -20 || temp_data_long > 215) ) ||
											     ( ((pPid[0].Data[0] & 0x80) != 0) && (temp_data_long < -20 || temp_data_long > 500) ) )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGRT22 out of range (-20 to 215 or 500)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
										else  // test 5.6
										{
											if ( temp_data_long < -20 || temp_data_long > 200 )
											{
												Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
												      "ECU %X  EGRT22 out of range (-20 to 200)\n",
												      GetEcuId ( EcuIndex ) );
											}
										}
									}
								}
							}
							break;

							case 0x6C:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F46C  TAC data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TAC data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TAC data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[1] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46C  TAC_A_CMD = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[2] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46C  TP_A_REL  = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDDieselFlag == FALSE && gOBDHybridFlag == FALSE &&
									     temp_data_float > 50 )  // NOT test 10.2
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TP_A_REL should be less than or equal to 50 %% \n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[3] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46C  TAC_B_CMD = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[4] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46C  TP_B_REL  = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDDieselFlag == FALSE && gOBDHybridFlag == FALSE &&
									     temp_data_float > 50 )  // NOT Test 10.2
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TP_B_REL should be less than or equal to 50 %%\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x6D:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F46D  FRP data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x3F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FRP data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xC0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  FRP data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = ( (pPid[0].Data[1] * 256) +
									                   pPid[0].Data[2] ) * 10;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46D  FRP_A_CMD = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = ( (pPid[0].Data[3] * 256) +
									                   pPid[0].Data[4] ) * 10;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46D  FRP_A     = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = ( pPid[0].Data[5] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46D  FRT_A     = %5d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FRT_A out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = ( (pPid[0].Data[6] * 256) +
									                   pPid[0].Data[7] ) * 10;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46D  FRP_B_CMD = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_long = ( (pPid[0].Data[8] * 256) +
									                   pPid[0].Data[9] ) * 10;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46D  FRP_B     = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x20) != 0 )
								{
									temp_data_long = ( pPid[0].Data[10] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46D  FRT_B     = %5d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FRT_B out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x6E:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F46E  ICP system data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  ICP data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  ICP data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = ( (pPid[0].Data[1] * 256) +
									                   pPid[0].Data[2] ) * 10;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46E  ICP_A_CMD = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = ( (pPid[0].Data[3] * 256) +
									                   pPid[0].Data[4] ) * 10;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46E  ICP_A     = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = ( (pPid[0].Data[5] * 256) +
									                   pPid[0].Data[6] ) * 10;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46E  ICP_B_CMD = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = ( (pPid[0].Data[7] * 256) +
									                   pPid[0].Data[8] ) * 10;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46E  ICP_B     = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
							}
							break;

							case 0x6F:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F46F  TCA sensor support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TCA sensor support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TCA sensor support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								// TCA_CINP supported
								if ( (pPid[0].Data[0] & 0x05) != 0 )
								{
									temp_data_long = pPid[0].Data[1];
									//if 8kPa/bit
									if ( (pPid[0].Data[0] & 0x04) != 0 )
									{
										temp_data_long *= 8;
									}
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46F  TCA_CINP  = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								// TCB_CINP supported
								if ( (pPid[0].Data[0] & 0x0A) != 0 )
								{
									temp_data_long = pPid[0].Data[2];
									//if 8kPa/bit
									if ( (pPid[0].Data[0] & 0x08) != 0 )
									{
										temp_data_long *= 8;
									}
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F46F  TCB_CINP  = %5d kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
							}
							break;

							case 0x70:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F470  BP data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x3F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  BP data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xC0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  BP data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[1] * 256) +
									                    pPid[0].Data[2] ) * (float)0.03125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F470  BP_A_CMD  = %4.3f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[3] * 256) +
									                    pPid[0].Data[4] ) * (float)0.03125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F470  BP_A_ACT  = %4.3f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[5] * 256) +
									                    pPid[0].Data[6] ) * (float)0.03125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F470  BP_B_CMD  = %4.3f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[7] * 256) +
									                    pPid[0].Data[8] ) * (float)0.03125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F470  BP_B_ACT  = %4.3f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								// bit 2 (0x04) BP_A Status and bit 5 (0x20) BP_B Status
								if ( (pPid[0].Data[0] & 0x24) != 0 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F470  BP control status = $%02X\n",
									      GetEcuId ( EcuIndex ),
									      pPid[0].Data[9] );
									// check that reserved bits are not set
									if ( pPid[0].Data[9] & 0xF0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  BP control status  Reserved bits set\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x71:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F471  VGT data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x3F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  VGT data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xC0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  VGT data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[1] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F471  VGT_A_CMD = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[2] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F471  VGT_A_ACT = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[3] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F471  VGT_B_CMD = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[4] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F471  VGT_B_ACT = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								// bit 2 (0x04) VGT_A Status and bit 5 (0x20) VGT_B Status
								if ( (pPid[0].Data[0] & 0x24) != 0 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F471  VGT control status = $%02X\n",
									      GetEcuId ( EcuIndex ),
									      pPid[0].Data[5] );
									// check that reserved bits are not set
									if ( pPid[0].Data[5] & 0xF0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F471  VGT control status  Reserved bits set\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x72:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F472  WG data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  WG data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  WG data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[1] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F472  WG_A_CMD  = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[2] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F472  WG_A_ACT  = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[3] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F472  WG_B_CMD  = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = (float)pPid[0].Data[4] * (float)(100.0 / 255.0);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F472  WG_B_ACT  = %2.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x73:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F473  EP data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x03) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EP data support  No sensors supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFC )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EP data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[1] * 256) +
									                    pPid[0].Data[2] ) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F473  EP1      = %4.2f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[3] * 256) +
									                    pPid[0].Data[4] ) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F473  EP2      = %4.2f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x74:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F474  TC_RPM data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x03) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TC_RPM data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFC )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TC_RPM data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] * 256) +
									                 pPid[0].Data[2];
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F474  TCA_RPM  = %d RPM\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = (pPid[0].Data[3] * 256) +
									                 pPid[0].Data[4];
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F474  TCB_RPM  = %d RPM\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
							}
							break;

							case 0x75:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F475  TC A Temp data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TC A Temp data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TC A Temp data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = ( pPid[0].Data[1] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F475  TCA_CINT  = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TCA_CINT out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = ( pPid[0].Data[2] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F475  TCA_COUTT = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TCA_COUTT out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[3] * 256) +
									                      pPid[0].Data[4] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F475  TCA_TINT  = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
									if ( temp_data_float < -20 || temp_data_float > 1000 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TCA_TINT out of range (-20 to 1000)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[5] * 256) +
									                      pPid[0].Data[6] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F475  TCA_TOUTT = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
									if ( temp_data_float < -20 || temp_data_float > 1000 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TCA_TOUTT out of range (-20 to 1000)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x76:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F476  TC B Temp data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TC B Temp data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  TC B Temp data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = ( pPid[0].Data[1] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F476  TCB_CINT  = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TCB_CINT out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = ( pPid[0].Data[2] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F476  TCB_COUTT = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TCB_COUTT out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[3] * 256) +
									                      pPid[0].Data[4] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F476  TCB_TINT  = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
									if ( temp_data_float < -20 || temp_data_float > 1000 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TCB_TINT out of range (-20 to 1000)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[5] * 256) +
									                      pPid[0].Data[6] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F476  TCB_TOUTT = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
									if ( temp_data_float < -20 || temp_data_float > 1000 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  TCB_TOUTT out of range (-20 to 1000)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x77:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F477  CACT data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  CACT data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  CACT data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = ( pPid[0].Data[1] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F477  CACT11   = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  CACT11 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = ( pPid[0].Data[2] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F477  CACT12   = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  CACT12 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = ( pPid[0].Data[3] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F477  CACT21   = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  CACT21 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = ( pPid[0].Data[4] ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F477  CACT22   = %3d C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
									if ( temp_data_long < -20 || temp_data_long > 120 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  CACT22 out of range (-20 to 120)\n",
										      GetEcuId ( EcuIndex ) );
									}
								}
							}
							break;

							case 0x78:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F478  EGT Bank 1 data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGT Bank 1 data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGT Bank 1 data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[1] * 256) +
									                      pPid[0].Data[2] ) * (float).1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F478  EGT11    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[3] * 256) +
									                      pPid[0].Data[4] ) * (float).1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F478  EGT12    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[5] * 256) +
									                      pPid[0].Data[6] ) * (float).1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F478  EGT13    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[7] * 256) +
									                      pPid[0].Data[8] ) * (float).1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F478  EGT14    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x79:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F479  EGT Bank 2 data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGT Bank 2 data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGT Bank 2 data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[1] * 256) +
									                      pPid[0].Data[2] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F479  EGT21    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[3] * 256) +
									                      pPid[0].Data[4] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F479  EGT22    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[5] * 256) +
									                      pPid[0].Data[6] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F479  EGT23    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[7] * 256) +
									                      pPid[0].Data[8] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F479  EGT24    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x7A:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F47A  DPF Bank 1 data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x07) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  DPF Bank 1 data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF8 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  DPF Bank 1 data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = (float)((signed short)( (pPid[0].Data[1] * 256) +
									                                          pPid[0].Data[2] )) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47A  DPF1_DP   = %4.1f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[3] ) * 256) +
									                      pPid[0].Data[4] ) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47A  DPF1_INP  = %4.1f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[5] ) * 256) +
									                      pPid[0].Data[6] ) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47A  DPF1_OUTP = %4.1f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x7B:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F47B  DPF Bank 2 data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x07) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  DPF Bank 2 data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF8 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  DPF Bank 2 data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = (float)((signed short)( (pPid[0].Data[1] * 256) +
									                                          pPid[0].Data[2] )) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47B  DPF2_DP   = %4.1f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[3] ) * 256) +
									                      pPid[0].Data[4] ) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47B  DPF2_INP  = %4.1f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[5] ) * 256) +
									                      pPid[0].Data[6] ) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47B  DPF2_OUTP = %4.1f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x7C:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F47C  DPF Temp data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  DPF Temp data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  DPF Temp data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[1] * 256) +
									                      pPid[0].Data[2] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47C  DPF1_INT  = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[3] * 256) +
									                      pPid[0].Data[4] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47C  DPF1_OUTT = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[5] * 256) +
									                      pPid[0].Data[6] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47C  DPF2_INT  = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_float = ( ( (pPid[0].Data[7] * 256) +
									                      pPid[0].Data[8] ) * (float)0.1 ) - 40;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F47C  DPF2_OUTT = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x7D:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F47D  NOx NTE control area status = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// NOx NTE control area status bits
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    NNTE_IN  = %s\n",
								      (pPid[0].Data[0] & 0x01) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    NNTE_OUT = %s\n",
								      (pPid[0].Data[0] & 0x02) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    NNTE_CAA = %s\n",
								      (pPid[0].Data[0] & 0x04) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    NNTE_DEF = %s\n",
								      (pPid[0].Data[0] & 0x08) ? "YES" : "NO" );
								
								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  NOx NTE control area status  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x7E:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F47E  PM NTE control area status = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// PM NTE control area status bits
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    PNTE_IN  = %s\n",
								      (pPid[0].Data[0] & 0x01) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    PNTE_OUT = %s\n",
								      (pPid[0].Data[0] & 0x02) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    PNTE_CAA = %s\n",
								      (pPid[0].Data[0] & 0x04) ? "YES" : "NO" );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    PNTE_DEF = %s\n",
								      (pPid[0].Data[0] & 0x08) ? "YES" : "NO" );
								
								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PM NTE control area status  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x7F:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F47F  Run Time support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x07) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF8 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] * 16777216) +
									                 (pPid[0].Data[2] * 65536) +
									                 (pPid[0].Data[3] * 256) +
									                 pPid[0].Data[4];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F47F  RUN_TIME  = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F47F  RUN_TIME  = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = (pPid[0].Data[5] * 16777216) +
									                 (pPid[0].Data[6] * 65536) +
									                 (pPid[0].Data[7] * 256) +
									                 pPid[0].Data[8];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F47F  IDLE_TIME = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F47F  IDLE_TIME = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = (pPid[0].Data[9] * 16777216) +
									                 (pPid[0].Data[10] * 65536) +
									                 (pPid[0].Data[11] * 256) +
									                 pPid[0].Data[12];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F47F  PTO_TIME  = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F47F  PTO_TIME  = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}
							}
							break;

							case 0x81:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F481  Run Time for AECD #1-#5 support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x1F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time for AECD #1-#5 support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xE0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time for AECD #1-#5 support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] * 16777216) +
									                 (pPid[0].Data[2] * 65536) +
									                 (pPid[0].Data[3] * 256) +
									                 pPid[0].Data[4];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD1_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD1_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[5] * 16777216) +
									                 (pPid[0].Data[6] * 65536) +
									                 (pPid[0].Data[7] * 256) +
									                 pPid[0].Data[8];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD1_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD1_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = (pPid[0].Data[9] * 16777216) +
									                 (pPid[0].Data[10] * 65536) +
									                 (pPid[0].Data[11] * 256) +
									                 pPid[0].Data[12];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD2_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD2_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[13] * 16777216) +
									                 (pPid[0].Data[14] * 65536) +
									                 (pPid[0].Data[15] * 256) +
									                 pPid[0].Data[16];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD2_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD2_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = (pPid[0].Data[17] * 16777216) +
									                 (pPid[0].Data[18] * 65536) +
									                 (pPid[0].Data[19] * 256) +
									                 pPid[0].Data[20];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD3_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD3_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[21] * 16777216) +
									                 (pPid[0].Data[22] * 65536) +
									                 (pPid[0].Data[23] * 256) +
									                 pPid[0].Data[24];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD3_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD3_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = (pPid[0].Data[25] * 16777216) +
									                 (pPid[0].Data[26] * 65536) +
									                 (pPid[0].Data[27] * 256) +
									                 pPid[0].Data[28];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD4_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD4_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[29] * 16777216) +
									                 (pPid[0].Data[30] * 65536) +
									                 (pPid[0].Data[31] * 256) +
									                 pPid[0].Data[32];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD4_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD4_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_long = (pPid[0].Data[33] * 16777216) +
									                 (pPid[0].Data[34] * 65536) +
									                 (pPid[0].Data[35] * 256) +
									                 pPid[0].Data[36];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD5_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD5_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[37] * 16777216) +
									                 (pPid[0].Data[38] * 65536) +
									                 (pPid[0].Data[39] * 256) +
									                 pPid[0].Data[40];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD5_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F481  AECD5_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}
							}
							break;

							case 0x82:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F482  Run Time for AECD #6-#10 support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x1F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time for AECD #6-#10 support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xE0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time for AECD #6-#10 support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] * 16777216) +
									                 (pPid[0].Data[2] * 65536) +
									                 (pPid[0].Data[3] * 256) +
									                 pPid[0].Data[4];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD6_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD6_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[5] * 16777216) +
									                 (pPid[0].Data[6] * 65536) +
									                 (pPid[0].Data[7] * 256) +
									                 pPid[0].Data[8];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD6_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD6_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = (pPid[0].Data[9] * 16777216) +
									                 (pPid[0].Data[10] * 65536) +
									                 (pPid[0].Data[11] * 256) +
									                 pPid[0].Data[12];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD7_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD7_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[13] * 16777216) +
									                 (pPid[0].Data[14] * 65536) +
									                 (pPid[0].Data[15] * 256) +
									                 pPid[0].Data[16];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD7_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD7_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = (pPid[0].Data[17] * 16777216) +
									                 (pPid[0].Data[18] * 65536) +
									                 (pPid[0].Data[19] * 256) +
									                 pPid[0].Data[20];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD8_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD8_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[21] * 16777216) +
									                 (pPid[0].Data[22] * 65536) +
									                 (pPid[0].Data[23] * 256) +
									                 pPid[0].Data[24];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD8_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD8_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = (pPid[0].Data[25] * 16777216) +
									                 (pPid[0].Data[26] * 65536) +
									                 (pPid[0].Data[27] * 256) +
									                 pPid[0].Data[28];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD1_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD9_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[29] * 16777216) +
									                 (pPid[0].Data[30] * 65536) +
									                 (pPid[0].Data[31] * 256) +
									                 pPid[0].Data[32];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD9_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD9_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_long = (pPid[0].Data[33] * 16777216) +
									                 (pPid[0].Data[34] * 65536) +
									                 (pPid[0].Data[35] * 256) +
									                 pPid[0].Data[36];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD10_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD10_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[37] * 16777216) +
									                 (pPid[0].Data[38] * 65536) +
									                 (pPid[0].Data[39] * 256) +
									                 pPid[0].Data[40];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD10_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F482  AECD10_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}
							}
							break;

							case 0x83:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F483  NOx Sensor Data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  NOx Sensor Data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] * 256) +
									                 pPid[0].Data[2];
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F483  NOX11    = %d ppm\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );

									if ( (pPid[0].Data[0] & 0x10) != 0 && temp_data_long != 0xFFFF )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  NOX11 should report $FFFF when Byte A Bit 4 is set\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = (pPid[0].Data[3] * 256) +
									                 pPid[0].Data[4];
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F483  NOX12    = %d ppm\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );

									if ( (pPid[0].Data[0] & 0x20) != 0 && temp_data_long != 0xFFFF )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  NOX12 should report $FFFF when Byte A Bit 5 is set\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] * 256) +
									                 pPid[0].Data[2];
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F483  NOX21    = %d ppm\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );

									if ( (pPid[0].Data[0] & 0x40) != 0 && temp_data_long != 0xFFFF )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  NOX21 should report $FFFF when Byte A Bit 6 is set\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = (pPid[0].Data[3] * 256) +
									                 pPid[0].Data[4];
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F483  NOX22    = %d ppm\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );

									if ( (pPid[0].Data[0] & 0x80) != 0 && temp_data_long != 0xFFFF )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  NOX22 should report $FFFF when Byte A Bit 7 is set\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long );
									}
								}
							}
							break;

							case 0x84:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F484  MST      = %d C\n",
								      GetEcuId ( EcuIndex ),
								      ( pPid[0].Data[0] ) -40 );
							}
							break;

							case 0x85:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F485  NOx Reagent data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  NOx Reagent data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  NOx Reagent data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[1] * 256) +
									                    pPid[0].Data[2] ) * (float)0.005;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F485  REAG_RATE = %.3f L/h\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[3] * 256) +
									                    pPid[0].Data[4] ) * (float)0.005;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F485  REAG_DEMD = %.3f L/h\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = ( pPid[0].Data[5] * 100) / 255;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F485  REAG_LVL  = %d %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = (pPid[0].Data[6] * 16777216) +
									                 (pPid[0].Data[7] * 65536) +
									                 (pPid[0].Data[8] * 256) +
									                 pPid[0].Data[9];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F485  NWI_TIME  = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F485  NWI_TIME  = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}
							}
							break;

							case 0x86:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F486  PM data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x03) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PM data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFC )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PM data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[1] * 256) +
									                    pPid[0].Data[2] ) * (float)0.0125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F486  PM11     = %.3f mg/m3\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[3] * 256) +
									                    pPid[0].Data[4] ) * (float)0.0125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F486  PM21     = %.3f mg/m3\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x87:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F487  MAP data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x03) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  MAP data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFC )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  MAP data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[1] * 256) +
									                    pPid[0].Data[2] ) * (float)0.03125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F487  MAP_A    = %.3f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
									{
										if ( temp_data_float == 0.0 &&
										     gOBDHybridFlag != TRUE )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  MAP_A must be non-zero\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_float = ( (pPid[0].Data[3] * 256) +
									                    pPid[0].Data[4] ) * (float)0.03125;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F487  MAP_B    = %.3f kPa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == TRUE )  // test 5.10 & 10.13
									{
										if ( temp_data_float == 0.0 &&
										     gOBDHybridFlag != TRUE )
										{
											Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
											      "ECU %X  MAP_B must be non-zero\n",
											      GetEcuId ( EcuIndex ) );
										}
									}
								}
							}
							break;

							case 0x88:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F488  SCR_INDUCE_SYSTEM = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F488  SCR_INDUCE_SYSTEM_HIST12 = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[1] );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F488  SCR_INDUCE_SYSTEM_HIST34 = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[2] );

								temp_data_long = ( (pPid[0].Data[3] * 256) +
								                   pPid[0].Data[4] );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F488  SCR_INDUCE_DIST_1N = %d km\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								temp_data_long = ( (pPid[0].Data[5] * 256) +
								                   pPid[0].Data[6] );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F488  SCR_INDUCE_DIST_1D = %d km\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								temp_data_long = ( (pPid[0].Data[7] * 256) +
								                   pPid[0].Data[8] );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F488  SCR_INDUCE_DIST_2N = %d km\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								temp_data_long = ( (pPid[0].Data[9] * 256) +
								                   pPid[0].Data[10] );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F488  SCR_INDUCE_DIST_3N = %d km\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								temp_data_long = ( (pPid[0].Data[11] * 256) +
								                   pPid[0].Data[12] );
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F488  SCR_INDUCE_DIST_4N = %d km\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );
							}
							break;

							case 0x89:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F489  Run Time for AECD #11-#15 support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x1F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time for AECD #11-#15 support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xE0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time for AECD #11-#15 support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] * 16777216) +
									                 (pPid[0].Data[2] * 65536) +
									                 (pPid[0].Data[3] * 256) +
									                 pPid[0].Data[4];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD11_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD11_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[5] * 16777216) +
									                 (pPid[0].Data[6] * 65536) +
									                 (pPid[0].Data[7] * 256) +
									                 pPid[0].Data[8];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD11_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD11_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = (pPid[0].Data[9] * 16777216) +
									                 (pPid[0].Data[10] * 65536) +
									                 (pPid[0].Data[11] * 256) +
									                 pPid[0].Data[12];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD12_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD12_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[13] * 16777216) +
									                 (pPid[0].Data[14] * 65536) +
									                 (pPid[0].Data[15] * 256) +
									                 pPid[0].Data[16];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD12_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD12_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = (pPid[0].Data[17] * 16777216) +
									                 (pPid[0].Data[18] * 65536) +
									                 (pPid[0].Data[19] * 256) +
									                 pPid[0].Data[20];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD13_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD13_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[21] * 16777216) +
									                 (pPid[0].Data[22] * 65536) +
									                 (pPid[0].Data[23] * 256) +
									                 pPid[0].Data[24];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD13_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD13_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = (pPid[0].Data[25] * 16777216) +
									                 (pPid[0].Data[26] * 65536) +
									                 (pPid[0].Data[27] * 256) +
									                 pPid[0].Data[28];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD41_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD14_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[29] * 16777216) +
									                 (pPid[0].Data[30] * 65536) +
									                 (pPid[0].Data[31] * 256) +
									                 pPid[0].Data[32];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD14_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD14_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_long = (pPid[0].Data[33] * 16777216) +
									                 (pPid[0].Data[34] * 65536) +
									                 (pPid[0].Data[35] * 256) +
									                 pPid[0].Data[36];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD15_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD15_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[37] * 16777216) +
									                 (pPid[0].Data[38] * 65536) +
									                 (pPid[0].Data[39] * 256) +
									                 pPid[0].Data[40];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD15_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F489  AECD15_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}
							}
							break;

							case 0x8A:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F48A  Run Time for AECD #16-#20 support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x1F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time for AECD #16-#20 support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xE0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Run Time for AECD #16-#20 support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( (pPid[0].Data[0] & 0x01) != 0 )
								{
									temp_data_long = (pPid[0].Data[1] * 16777216) +
									                 (pPid[0].Data[2] * 65536) +
									                 (pPid[0].Data[3] * 256) +
									                 pPid[0].Data[4];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD16_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD16_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[5] * 16777216) +
									                 (pPid[0].Data[6] * 65536) +
									                 (pPid[0].Data[7] * 256) +
									                 pPid[0].Data[8];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD16_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD16_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x02) != 0 )
								{
									temp_data_long = (pPid[0].Data[9] * 16777216) +
									                 (pPid[0].Data[10] * 65536) +
									                 (pPid[0].Data[11] * 256) +
									                 pPid[0].Data[12];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD17_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD17_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[13] * 16777216) +
									                 (pPid[0].Data[14] * 65536) +
									                 (pPid[0].Data[15] * 256) +
									                 pPid[0].Data[16];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD17_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD17_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x04) != 0 )
								{
									temp_data_long = (pPid[0].Data[17] * 16777216) +
									                 (pPid[0].Data[18] * 65536) +
									                 (pPid[0].Data[19] * 256) +
									                 pPid[0].Data[20];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD18_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD18_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[21] * 16777216) +
									                 (pPid[0].Data[22] * 65536) +
									                 (pPid[0].Data[23] * 256) +
									                 pPid[0].Data[24];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD18_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD18_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x08) != 0 )
								{
									temp_data_long = (pPid[0].Data[25] * 16777216) +
									                 (pPid[0].Data[26] * 65536) +
									                 (pPid[0].Data[27] * 256) +
									                 pPid[0].Data[28];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD19_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD19_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[29] * 16777216) +
									                 (pPid[0].Data[30] * 65536) +
									                 (pPid[0].Data[31] * 256) +
									                 pPid[0].Data[32];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD19_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD19_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}

								if ( (pPid[0].Data[0] & 0x10) != 0 )
								{
									temp_data_long = (pPid[0].Data[33] * 16777216) +
									                 (pPid[0].Data[34] * 65536) +
									                 (pPid[0].Data[35] * 256) +
									                 pPid[0].Data[36];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD20_TIME1 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD20_TIME1 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}

									temp_data_long = (pPid[0].Data[37] * 16777216) +
									                 (pPid[0].Data[38] * 65536) +
									                 (pPid[0].Data[39] * 256) +
									                 pPid[0].Data[40];
									if ( temp_data_long == -1 )
									{
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD20_TIME2 = n.a.\n" );
									}
									else
									{
										GetHoursMinsSecs ( temp_data_long, &hours, &mins, &secs );
										Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PID $F48A  AECD20_TIME2 = %d sec (%d hrs, %d mins, %d secs)\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_long,
										      hours, mins, secs );
									}
								}
							}
							break;

							case 0x8B:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F48B  Diesel Aftertreatment support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that reserved bit is not set
								if ( pPid[0].Data[0] & 0x80 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Diesel Aftertreatment Status support reserved bit 7 set\n",
									      GetEcuId ( EcuIndex ) );
								}


								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F48B  Diesel Aftertreatment Status = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[1] );

								// Diesel Aftertreatment Status bits
								if ( pPid[0].Data[0] & 0x01 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    DPF_REGEN_STATUS = %s\n",
									      (pPid[0].Data[1] & 0x01) ? "YES" : "NO" );
								}
								if ( pPid[0].Data[0] & 0x02 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    DPF_REGEN_TYPE   = %s\n",
									      (pPid[0].Data[1] & 0x02) ? "ACTIVE" : "PASSIVE" );
								}
								if ( pPid[0].Data[0] & 0x04 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    NOX_ADS_REGEN    = %s\n",
									      (pPid[0].Data[1] & 0x04) ? "YES" : "NO" );
								}
								if ( pPid[0].Data[0] & 0x08 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    NOX_ADS_DESULF   = %s\n",
									      (pPid[0].Data[1] & 0x08) ? "YES" : "NO" );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[1] & 0xF0 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48B  Diesel Aftertreatment Status (BYTE B) Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}


								if ( pPid[0].Data[0] & 0x10 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48B  DPF_REGEN_PCT  = %.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      (float)(pPid[0].Data[2]) * (float)(100.0 / 255.0) );
								}
								if ( pPid[0].Data[0] & 0x20 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48B  DPF_REGEN_AVGT = %d minutes\n",
									      GetEcuId ( EcuIndex ),
									      (pPid[0].Data[3] * 256) +
									      pPid[0].Data[4] );
								}
								if ( pPid[0].Data[0] & 0x40 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48B  DPF_REGEN_AVGD = %d km\n",
									      GetEcuId ( EcuIndex ),
									      (pPid[0].Data[5] * 256) +
									      pPid[0].Data[6] );
								}
							}
							break;

							case 0x8C:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F48C  O2 Sensor (Wide Range) support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );


								// display associated values
								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_float = (float)((pPid[0].Data[1] * 256) + pPid[0].Data[2]) * (float)(.001526);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48C  O2S11_PCT = %3.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((pPid[0].Data[3] * 256) + pPid[0].Data[4]) * (float)(.001526);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48C  O2S12_PCT = %3.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = (float)((pPid[0].Data[5] * 256) + pPid[0].Data[6]) * (float)(.001526);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48C  O2S21_PCT = %3.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = (float)((pPid[0].Data[7] * 256) + pPid[0].Data[8]) * (float)(.001526);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48C  O2S22_PCT = %3.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
								if ( pPid[0].Data[0] & 0x10 )
								{
									temp_data_float = (float)((pPid[0].Data[9] * 256) + pPid[0].Data[10]) * (float)(.000122);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48C  LAMBDA11  = %.3f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
								if ( pPid[0].Data[0] & 0x20 )
								{
									temp_data_float = (float)((pPid[0].Data[11] * 256) + pPid[0].Data[12]) * (float)(.000122);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48C  LAMBDA12  = %.3f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
								if ( pPid[0].Data[0] & 0x40 )
								{
									temp_data_float = (float)((pPid[0].Data[13] * 256) + pPid[0].Data[14]) * (float)(.000122);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48C  LAMBDA21  = %.3f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
								if ( pPid[0].Data[0] & 0x80 )
								{
									temp_data_float = (float)((pPid[0].Data[15] * 256) + pPid[0].Data[16]) * (float)(.000122);
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48C  LAMBDA22  = %.3f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x8D:
							{
								temp_data_float = (float)pPid[0].Data[0] * (float)(100.0 / 255.0);
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F48D  TP_G     = %.1f %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );
							}
							break;

							case 0x8E:
							{
								temp_data_long = (signed long)(pPid[0].Data[0]) - 125;
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F48E  TQ_FR    = %d %%\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );
							}
							break;

							case 0x8F:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F48F  PM Sensor support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PM Sensor Data support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}
								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PM Sensor Data support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								// PM Sensor Operating Status Bank 1 Sensor 1
								if ( pPid[0].Data[0] & 0x01 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48F  PM11 Sensor status = $%02X\n",
									      GetEcuId ( EcuIndex ),
									      pPid[0].Data[1] );

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    PM11_Active = %s\n",
									      (pPid[0].Data[1] & 0x01) ? "YES" : "NO" );

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    PM11_REGEN  = %s\n",
									      (pPid[0].Data[1] & 0x02) ? "YES" : "NO" );

									// check that reserved bit is not set
									if ( pPid[0].Data[1] & 0xFC )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PM11 Sensor Operating Status  Reserved bits set\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								// PM Sensor Normalized Output Bank 1 Sensor 1
								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((signed short)( (pPid[0].Data[2] * 256) +
									                                          pPid[0].Data[3] )) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48F  PM11     = %3.2f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( temp_data_float < -100 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PM11 must be greater than -100%%\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_float );
									}
								}

								// PM Sensor Operating Status Bank 2 Sensor 1
								if ( pPid[0].Data[0] & 0x04 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48F  PM21 Sensor status = $%02X\n",
									      GetEcuId ( EcuIndex ),
									      pPid[0].Data[4] );

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    PM21_Active = %s\n",
									      (pPid[0].Data[4] & 0x01) ? "YES" : "NO" );

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    PM21_REGEN  = %s\n",
									      (pPid[0].Data[4] & 0x02) ? "YES" : "NO" );

									if ( pPid[0].Data[4] & 0xFC )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PM21 Sensor Operating Status  Reserved bits set\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								// PM Sensor Normalized Output Bank 2 Sensor 1
								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = (float)((signed short)( (pPid[0].Data[5] * 256) +
									                                          pPid[0].Data[6] )) * (float)0.01;
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F48F  PM21     = %.2f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( temp_data_float < -100 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  PM21 must be greater than -100%%\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_float );
									}
								}
							}
							break;

							case 0x90:
							{
								// WWH-OBD Vehicle OBD System Information
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F490  WWH-OBD Vehicle OBD System Information = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    MI_DISP_VOBD = $%02X\n",
								      pPid[0].Data[0] & 0x03 );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    MI_MODE_VOBD = $%02X\n",
								      (pPid[0].Data[0] & 0x3C)>>2 );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    VOBD_RDY = %s\n",
								      (pPid[0].Data[0] & 0x40) ? "YES" : "NO" );

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0x80 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X    WWH-OBD Vehicle OBD System Information  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								temp_data_long = (pPid[0].Data[1] * 256) +
								                 pPid[0].Data[2];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F490  VOBD_MI_TIME = %d hrs\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );
							}
							break;

							case 0x91:
							{
								// WWH-OBD Vehicle OBD System Information
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F491  WWH-OBD ECU OBD System Information = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "                    MI_MODE_ECU = $%02X\n",
								      pPid[0].Data[0] & 0x0F );

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  WWH-OBD ECU OBD System Information support  Reserved bits set\n" );
								}

								temp_data_long = (pPid[0].Data[1] * 256) +
								                 pPid[0].Data[2];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F491  OBD_MI_TIME = %d hrs\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );

								temp_data_long = (pPid[0].Data[3] * 256) +
								                 pPid[0].Data[4];
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F491  OBD_B1_TIME = %d hrs\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_long );
							}
							break;

							case 0x92:
							{
								// Fuel System Control Status (Compression Ignition)
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F492  Fuel System Control Status support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0xFF) == 0 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Fuel System Control Status support  No data supported\n" );
								}

								// Fuel System Status
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F492  FUELSYS  = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[1] );

								// Fuel Pressure Control 1 Status
								if ( pPid[0].Data[0] & 0x01 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    FP1_CL  = %c\n",
									      (pPid[0].Data[1] & 0x01) ? '1' : '0' );
								}

								// Fuel Injection Quantity Control 1 Status
								if ( pPid[0].Data[0] & 0x02 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    FIQ1_CL = %c\n",
									      (pPid[0].Data[1] & 0x02) ? '1' : '0' );
								}

								// Fuel Injection Timing Control 1 Status
								if ( pPid[0].Data[0] & 0x04 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    FIT1_CL = %c\n",
									      (pPid[0].Data[1] & 0x04) ? '1' : '0' );
								}

								// Idle Fuel Balance/Contribution Control 1 Status
								if ( pPid[0].Data[0] & 0x08 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    IFB1_CL = %c\n",
									      (pPid[0].Data[1] & 0x08) ? '1' : '0' );
								}

								// Fuel Pressure Control 2 Status
								if ( pPid[0].Data[0] & 0x10 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    FP2_CL  = %c\n",
									      (pPid[0].Data[1] & 0x10) ? '1' : '0' );
								}

								// Fuel Injection Quantity Control 2 Status
								if ( pPid[0].Data[0] & 0x20 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    FIQ2_CL = %c\n",
									      (pPid[0].Data[1] & 0x20) ? '1' : '0' );
								}

								// Fuel Injection Timing Control 2 Status
								if ( pPid[0].Data[0] & 0x40 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    FIT2_CL = %c\n",
									      (pPid[0].Data[1] & 0x40) ? '1' : '0' );
								}

								// Idle Fuel Balance/Contribution Control 2 Status
								if ( pPid[0].Data[0] & 0x80 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    IFB2_CL = %c\n",
									      (pPid[0].Data[1] & 0x80) ? '1' : '0' );
								}
							}
							break;

							case 0x93:
							{
								// WWH-OBD Vehicle OBD Counters support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F493  WWH-OBD Vehicle OBD Counters support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFE )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  WWH-OBD Vehicle OBD Counters support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_long = (pPid[0].Data[1] * 256) +
									                 pPid[0].Data[2];
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F493  MI_TIME_CUM = %d hrs\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
							}
							break;

							case 0x94:
							{
								// NOx Warning And Inducement System support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F494  NOx Warning And Inducement System support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x3F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  NOx Warning And Inducement System support  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xC0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  NOx Warning And Inducement System support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								// NOx Warning And Inducement System support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F494  NOx Warning And Inducement System status  = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								if ( pPid[0].Data[0] & 0x01 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    NOX_WARN_ACT = %s\n",
									      (pPid[0].Data[1] & 0x01) ? "YES" : "NO" );

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    INDUC_L1 = $%02X\n",
									      (pPid[0].Data[1] & 0x06)>>1 );

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    INDUC_L2 = $%02X\n",
									      (pPid[0].Data[1] & 0x18)>>3 );

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "                    INDUC_L3 = $%02X\n",
									      (pPid[0].Data[1] & 0x60)>>5 );

									if ( pPid[0].Data[0] & 0x80 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X    NOx Warning And Inducement Ststem status  Reserved bits set\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								temp_data_long = (pPid[0].Data[2] * 256) +
								                 pPid[0].Data[3];
								if ( pPid[0].Data[0] & 0x02 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F494  REAG_QUAL_TIME = %d hrs\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
								else if ( temp_data_long != 65535 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  REAG_QUAL_TIME must be 65535 if it is not supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								temp_data_long = (pPid[0].Data[4] * 256) +
								                 pPid[0].Data[5];
								if ( pPid[0].Data[0] & 0x04 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F494  REAG_CON_TIME  = %d hrs\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
								else if ( temp_data_long != 65535 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  REAG_CON_TIME must be 65535 if it is not supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								temp_data_long = (pPid[0].Data[6] * 256) +
								                 pPid[0].Data[7];
								if ( pPid[0].Data[0] & 0x08 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F494  REAG_DOSE_TIME = %d hrs\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
								else if ( temp_data_long != 65535 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  REAG_DOSE_TIME must be 65535 if it is not supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								temp_data_long = (pPid[0].Data[8] * 256) +
								                 pPid[0].Data[9];
								if ( pPid[0].Data[0] & 0x10 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F494  EGR_TIME       = %d hrs\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
								else if ( temp_data_long != 65535 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EGR_TIME must be 65535 if it is not supported\n",
									      GetEcuId ( EcuIndex ) );
								}

								temp_data_long = (pPid[0].Data[10] * 256) +
								                 pPid[0].Data[11];
								if ( pPid[0].Data[0] & 0x20 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F494  NOX_DTC_TIME   = %d hrs\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_long );
								}
								else if ( temp_data_long != 65535 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  NOX_DTC_TIME must be 65535 if it is not supported\n",
									      GetEcuId ( EcuIndex ) );
								}
							}
							break;

							case 0x98:
							{
								// Exhaust Gas Temperature Sensor support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F498  Exhaust Gas Temperature Sensor Bank 1 support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Exhaust Gas Temperature Sensor Bank 1 support  No data supported\n" );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Exhaust Gas Temperature Sensor Bank 1 support  Reserved bits set\n" );
								}

								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_float = ((float)((pPid[0].Data[1] * 256) +
									                           pPid[0].Data[2]) * (float).1) - (float)40;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F498  EGT15    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = ((float)((pPid[0].Data[3] * 256) +
									                           pPid[0].Data[4]) * (float).1) - (float)40;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F498  EGT16    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = ((float)((pPid[0].Data[5] * 256) +
									                           pPid[0].Data[6]) * (float).1) - (float)40;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F498  EGT17    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = ((float)((pPid[0].Data[7] * 256) +
									                           pPid[0].Data[8]) * (float).1) - (float)40;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F498  EGT18    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x99:
							{
								// Exhaust Gas Temperature Sensor support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F499  Exhaust Gas Temperature Sensor Bank 2 support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Exhaust Gas Temperature Sensor Bank 2 support  No data supported\n" );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Exhaust Gas Temperature Sensor Bank 2 support  Reserved bits set\n" );
								}

								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_float = ((float)((pPid[0].Data[1] * 256) +
									                           pPid[0].Data[2]) * (float).1) - (float)40;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F499  EGT25    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = ((float)((pPid[0].Data[3] * 256) +
									                           pPid[0].Data[4]) * (float).1) - (float)40;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F499  EGT26    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = ((float)((pPid[0].Data[5] * 256) +
									                           pPid[0].Data[6]) * (float).1) - (float)40;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F499  EGT27    = %.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = ((float)((pPid[0].Data[7] * 256) +
									                           pPid[0].Data[8]) * (float).1) - (float)40;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F499  EGT28    = %3.1f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x9A:
							{
								// Hybrid/EV Vehicle System Data support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F49A  Hybrid/EV Vehicle System Data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Hybrid/EV Vehicle System Data support  Reserved bits set\n" );
								}

								// Hybrid/EV Vehicle state
								if ( pPid[0].Data[0] & 0x01 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49A  HEV_MODE     = %s\n",
									      GetEcuId ( EcuIndex ),
									      (pPid[0].Data[1] & 0x01) ? "CDM" : "CSM" );
								}

								// Enhanced Hybrid/EV Vehicle state
								if ( pPid[0].Data[0] & 0x08 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49A  ENH_HEV_MODE = $%02X\n",
									      GetEcuId ( EcuIndex ),
									      (pPid[0].Data[1] & 0x06)>>1 );
								}

								// Check if reserved bits set
								if ( pPid[0].Data[1] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      " ECU %X  Hybrid/EV Vehicle Charging state  Reserved bits set\n" );
								}

								// Hybrid/EV Battery System Voltage
								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((unsigned short)( (pPid[0].Data[2] * 256) +
									                                            pPid[0].Data[3] )) * (float)0.015625;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49A  HEV_BATT_V   = %.6f V\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( temp_data_float <= 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  HEV_BATT_V must be greater than 0 V\n",
										      GetEcuId ( EcuIndex ),
										      temp_data_float );
									}
								}

								// Hybrid/EV Battery System Current
								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = (float)((signed short)( (pPid[0].Data[4] * 256) +
									                                          pPid[0].Data[5] )) * (float)0.1;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49A  HEV_BATT_A   = %.2f A\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x9B:
							{
								// Diesel Exhaust Fluid Sensor Data support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F49B  Diesel Exhaust Fluid Sensor Data support = $%1X\n",
								      GetEcuId ( EcuIndex ),
								      (pPid[0].Data[0] & 0x0F) );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x07) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Diesel Exhaust Fluid Sensor Data support  No data supported\n" );
								}

								// Diesel Exhaust Fluid Type
								if ( pPid[0].Data[0] & 0x01 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49B  DEF_Type = $%1X\n",
									      GetEcuId ( EcuIndex ),
									      (pPid[0].Data[0] >> 4) );
								}

								// Diesel Exhaust Fluid Concentration
								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[1])) * (float)0.25;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49B  DEF_CON  = %.2f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								// Diesel Exhaust Fluid Tank Temperature
								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[2])) - (float)40.00;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49B  DEF_T    = %f C\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								// Diesel Exhaust Fluid Tank Level
								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[3])) * (float)(100.0/255.0);

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49B  DEF_LVL  = %.2f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x9C:
							{
								// Exhaust Gas Temperature Sensor support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F49C  O2 Sensor Data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0xFF) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  O2 Sensor Data support  No data supported\n" );
								}

								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_float = (float)((pPid[0].Data[1] * 256) +
									                          pPid[0].Data[2]) * (float).001526;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49C  O2S13_PCT = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((pPid[0].Data[3] * 256) +
									                          pPid[0].Data[4]) * (float).001526;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49C  O2S14_PCT = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = (float)((pPid[0].Data[5] * 256) +
									                          pPid[0].Data[6]) * (float).001526;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49C  O2S23_PCT = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = (float)((pPid[0].Data[7] * 256) +
									                          pPid[0].Data[8]) * (float).001526;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49C  O2S24_PCT = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x10 )
								{
									temp_data_float = (float)((pPid[0].Data[9] * 256) +
									                          pPid[0].Data[10]) * (float).000122;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49C  LAMBDA13  = %.6f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x20 )
								{
									temp_data_float = (float)((pPid[0].Data[11] * 256) +
									                          pPid[0].Data[12]) * (float).000122;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49C  LAMBDA14  = %.6f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x40 )
								{
									temp_data_float = (float)((pPid[0].Data[13] * 256) +
									                          pPid[0].Data[14]) * (float).000122;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49C  LAMBDA23  = %.6f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x80 )
								{
									temp_data_float = (float)((pPid[0].Data[15] * 256) +
									                          pPid[0].Data[16]) * (float).000122;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49C  LAMBDA24  = %.6f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0x9D:
							{
								// Engine Fuel Rate
								temp_data_float = (float)((unsigned short)((pPid[0].Data[0] * 256) +
								                                           pPid[0].Data[1])) * (float).02;

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F49D  ENG_FUEL_RATE = %.2f g/s\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  ENG_FUEL_RATE must be 0 g/s when engine is not running\n" );
								}
								else if ( gOBDEngineRunning == TRUE && temp_data_float == 0 && gOBDHybridFlag != TRUE )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  ENG_FUEL_RATE must be greater than 0 g/s when engine is running\n" );
								}

								temp_data_float = (float)((unsigned short)((pPid[0].Data[2] * 256) +
								                                           pPid[0].Data[3])) * (float).02;

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F49D  VEH_FUEL_RATE = %.2f g/s\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  VEH_FUEL_RATE must be 0 g/s when engine is not running\n" );
								}
								else if ( gOBDEngineRunning == TRUE && temp_data_float == 0 && gOBDHybridFlag != TRUE )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  VEH_FUEL_RATE must be greater than 0 g/s when engine is running\n" );
								}
							}
							break;

							case 0x9E:
							{
								// Engine Exhaust Flow Rate
								temp_data_float = (float)((unsigned short)((pPid[0].Data[0] * 256) +
								                                           pPid[0].Data[1])) * (float).2;

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F49E  EXH_RATE = %.2f kg/h\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EXH_RATE must be 0 kg/h when engine is not running\n" );
								}
								else if ( gOBDEngineRunning == TRUE && temp_data_float == 0 && gOBDHybridFlag != TRUE )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  EXH_RATE must be greater than 0 kg/h when engine is running\n" );
								}
							}
							break;

							case 0x9F:
							{
								// Fuel System Percentage Use support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F49F  Fuel System Percentage Use support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x03) == 0 &&  // bits 0,1  Fuel System A,B uses Bank 1
								     (pPid[0].Data[0] & 0x0C) == 0 &&  // bits 2,3  Fuel System A,B uses Bank 2
								     (pPid[0].Data[0] & 0x30) == 0 &&  // bits 4,5  Fuel System A,B uses Bank 3
								     (pPid[0].Data[0] & 0xC0) == 0 )   // bits 6,7  Fuel System A,B uses Bank 4
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Fuel System Percentage Use support  (At least one bank must be supported)\n" );
								}

								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[1]) * (float)(100.0/255.0));

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49F  FUELSYSA_B1 = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FUELSYSA_B1 must be 0 %% when engine is not running\n" );
									}
									else if ( (gOBDEngineRunning == TRUE && temp_data_float != 0) || gOBDHybridFlag == TRUE )
									{
										fPid9FSuccess = TRUE;
									}
								}

								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[2])) * (float)(100.0/255.0);

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49F  FUELSYSB_B1 = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FUELSYSB_B1 must be 0 %% when engine is not running\n" );
									}
									else if ( (gOBDEngineRunning == TRUE && temp_data_float != 0) || gOBDHybridFlag == TRUE )
									{
										fPid9FSuccess = TRUE;
									}
								}

								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[3])) * (float)(100.0/255.0);

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49F  FUELSYSA_B2 = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FUELSYSA_B2 must be 0 %% when engine is not running\n" );
									}
									else if ( (gOBDEngineRunning == TRUE && temp_data_float != 0) || gOBDHybridFlag == TRUE )
									{
										fPid9FSuccess = TRUE;
									}
								}

								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[4])) * (float)(100.0/255.0);

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49F  FUELSYSB_B2 = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FUELSYSB_B2 must be 0 %% when engine is not running\n" );
									}
									else if ( (gOBDEngineRunning == TRUE && temp_data_float != 0) || gOBDHybridFlag == TRUE )
									{
										fPid9FSuccess = TRUE;
									}
								}

								if ( pPid[0].Data[0] & 0x10 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[5])) * (float)(100.0/255.0);

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49F  FUELSYSA_B3 = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FUELSYSA_B3 must be 0 %% when engine is not running\n" );
									}
									else if ( (gOBDEngineRunning == TRUE && temp_data_float != 0) || gOBDHybridFlag == TRUE )
									{
										fPid9FSuccess = TRUE;
									}
								}

								if ( pPid[0].Data[0] & 0x20 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[6])) * (float)(100.0/255.0);

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49F  FUELSYSB_B3 = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FUELSYSB_B3 must be 0 %% when engine is not running\n" );
									}
									else if ( (gOBDEngineRunning == TRUE && temp_data_float != 0) || gOBDHybridFlag == TRUE )
									{
										fPid9FSuccess = TRUE;
									}
								}

								if ( pPid[0].Data[0] & 0x40 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[7])) * (float)(100.0/255.0);

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49F  FUELSYSA_B4 = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FUELSYSA_B4 must be 0 %% when engine is not running\n" );
									}
									else if ( (gOBDEngineRunning == TRUE && temp_data_float != 0) || gOBDHybridFlag == TRUE )
									{
										fPid9FSuccess = TRUE;
									}
								}

								if ( pPid[0].Data[0] & 0x80 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[8])) * (float)(100.0/255.0);

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F49F  FUELSYSB_B4 = %.6f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  FUELSYSB_B4 must be 0 %% when engine is not running\n" );
									}
									else if ( (gOBDEngineRunning == TRUE && temp_data_float != 0) || gOBDHybridFlag == TRUE )
									{
										fPid9FSuccess = TRUE;
									}
								}

								if ( fPid9FSuccess == FALSE && gOBDEngineRunning == TRUE)
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  At least one of the FUELSYS banks must be greater than 0 %% when engine is running\n" );
								}

							}
							break;

							case 0xA1:
							{
								// NOx Sensor Corrected Data support
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F4A1  NOx Sensor Corrected Data support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x0F) == 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  NOx Sensor Corrected Data  No data supported\n",
									      GetEcuId ( EcuIndex ) );
								}


								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_float = (float)((unsigned short)((pPid[0].Data[1] * 256) +
									                                           pPid[0].Data[2]));

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A1  NOXC11   = %5.f ppm\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( (pPid[0].Data[0] & 0x10) == 0x10 && temp_data_float != 0xFFFF )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  NOXC11 Data Availability Bit Set, Sensor Data should report $FFFF\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((unsigned short)((pPid[0].Data[3] * 256) +
									                                           pPid[0].Data[4]));

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A1  NOXC12   = %5.f ppm\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( (pPid[0].Data[0] & 0x20) == 0x10 && temp_data_float != 0xFFFF )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  NOXC12 Data Availability Bit Set, Sensor Data should report $FFFF\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = (float)((unsigned short)((pPid[0].Data[5] * 256) +
									                                           pPid[0].Data[6]));

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A1  NOXC21   = %5.f ppm\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );

									if ( (pPid[0].Data[0] & 0x40) == 0x10 && temp_data_float != 0xFFFF )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  NOXC21 Data Availability Bit Set, Sensor Data should report $FFFF\n",
										      GetEcuId ( EcuIndex ) );
									}
								}

								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = (float)((unsigned short)((pPid[0].Data[7] * 256) +
									                                           pPid[0].Data[8]));

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A1  NOXC22   = %f ppm\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

									if ( (pPid[0].Data[0] & 0x80) == 0x10 && temp_data_float != 0xFFFF )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										      "ECU %X  NOXC22 Data Availability Bit Set, Sensor Data should report $FFFF\n",
										      GetEcuId ( EcuIndex ) );
									}
							}
							break;

							case 0xA2:
							{
								// Cylinder Fuel Rate
								temp_data_float = (float)((unsigned short)((pPid[0].Data[0] * 256) +
								                                           pPid[0].Data[1])) * (float)0.03125;

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F4A2  CYL_RATE = %.6f mg/stroke\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float );

								if ( gOBDEngineRunning == FALSE && temp_data_float != 0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "         CYL_RATE must be 0 mg/stroke when engine is not running\n" );
								}
								else if ( gOBDEngineRunning == TRUE && temp_data_float == 0 && gOBDHybridFlag != TRUE )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "         CYL_RATE must be greater than 0 mg/stroke when engine is running\n" );
								}
							}
							break;

							case 0xA3:
							{
								// Evap System Vapor Pressure
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F4A3  Evaporative System Vapor Pressure support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that required bit is set
								if ( (pPid[0].Data[0] & 0x03) == 0x03 ||
								     (pPid[0].Data[0] & 0x0C) == 0x0C )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Evaporative System Vapor Pressure support  Two different scaling ranges supported for the same sensor\n",
									      GetEcuId ( EcuIndex ) );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xF0 )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Evaporative System Vapor Pressure support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}


								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_float = (float)((signed short)((pPid[0].Data[1] * 256) +
									                                         pPid[0].Data[2])) * (float)0.25;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A3  EVAP_A_VP = %f Pa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((signed short)((pPid[0].Data[3] * 256) +
									                                         pPid[0].Data[4])) * (float)2;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A3  EVAP_A_VP (Wide Range) = %f Pa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x04 )
								{
									temp_data_float = (float)((unsigned short)((pPid[0].Data[5] * 256) +
									                                           pPid[0].Data[6])) * (float)0.25;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A3  EVAP_B_VP = %f Pa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x08 )
								{
									temp_data_float = (float)((unsigned short)((pPid[0].Data[7] * 256) +
									                                            pPid[0].Data[8])) * (float)2;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A3  EVAP_B_VP (Wide Range) = %f Pa\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0xA4:
							{
								// Transmission Actual Gear
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F4A4  Transmission Actual Gear Status support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFC )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Transmission Actual Gear Status support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}


								if ( pPid[0].Data[0] & 0x01 )
								{
									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A4  GEAR_ACT = %s\n",
									      GetEcuId ( EcuIndex ),
									      GEAR[(pPid[0].Data[1])>>4] );
								}

								// check that reserved bits are not set
								if ( pPid[0].Data[1] & 0x0F )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Transmission Actual Gear Status  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}

								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((unsigned short)((pPid[0].Data[2] * 256) +
									                                            pPid[0].Data[3])) * (float)0.001;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A4  GEAR_RAT = %f\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0xA5:
							{
								// Diesel Exhaust Fluid Dosing
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F4A5  Commanded DEF Dosing support = $%02X\n",
								      GetEcuId ( EcuIndex ),
								      pPid[0].Data[0] );

								// check that reserved bits are not set
								if ( pPid[0].Data[0] & 0xFC )
								{
									Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  Commanded DEF Dosing support  Reserved bits set\n",
									      GetEcuId ( EcuIndex ) );
								}


								if ( pPid[0].Data[0] & 0x01 )
								{
									temp_data_float = (float)((unsigned short)(pPid[0].Data[1])) * (float)0.5;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A5  DEF_CMD  = %3.1f %%\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}

								if ( pPid[0].Data[0] & 0x02 )
								{
									temp_data_float = (float)((unsigned short)((pPid[0].Data[2] * 256) +
									                                            pPid[0].Data[3])) * (float)0.0005;

									Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
									      "ECU %X  PID $F4A5  DEF_UCDC = %3.4f L\n",
									      GetEcuId ( EcuIndex ),
									      temp_data_float );
								}
							}
							break;

							case 0xA6:
							{
								temp_data_float = (float)((unsigned short)(pPid[0].Data[0] * 16777216) +
								                                          (pPid[0].Data[1] * 65536) +
								                                          (pPid[0].Data[2] * 256) +
								                                           pPid[0].Data[3]) * (float)0.1;

								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F4A6  ODO      = %.1f km (%.1f miles)\n",
								      GetEcuId ( EcuIndex ),
								      temp_data_float,
								      temp_data_float/(float)1.609344 );
							}
							break;


							default:
							{
								// Non-OBD PID
							}
							break;

						}  // end switch ( pPid[0].PIDLSB )

					}  // end if PIDMSB == F4

					else if ( pPid[0].PIDMSB == 0xF5 )
					{
						// Check various PID values for validity based on vehicle state
						switch ( pPid[0].PIDLSB )
						{
							case 0x01:
							{
								// Capture the response from PID $F501
								// for use with MID verification logic
								memcpy ( &gEcuData[EcuIndex].PIDF501, &pPid[0], sizeof ( PID ) );

								if ( OBDEngineDontCare == FALSE ) // NOT Test 10.2
								{
									if ( VerifyPIDF501 ( pPid, EcuIndex, &ulIMBitsForVehicle ) != PASS )
									{
										Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
										       "PID $F501 failures detected.\n" );
									}
								}
							}
							break;
							case 0x02:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F502  Misfire Counts (Last/Current drive cycle) Cylinders 1-8  NO DECODE CURRENTLY IMPLEMENTED\n",
								      GetEcuId ( EcuIndex ) );
							}
							break;
							case 0x03:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F503  Misfire Counts (Last/Current drive cycle) Cylinders 9-16  NO DECODE CURRENTLY IMPLEMENTED\n",
								      GetEcuId ( EcuIndex ) );
							}
							break;
							case 0x04:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F504  Misfire Counts (10 Trip EWMA) Cylinders 1-8  NO DECODE CURRENTLY IMPLEMENTED\n",
								      GetEcuId ( EcuIndex ) );
							}
							break;
							case 0x05:
							{
								Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F505  Misfire Counts (10 Trip EWMA) Cylinders 9-16  NO DECODE CURRENTLY IMPLEMENTED\n",
								      GetEcuId ( EcuIndex ) );
							}
							break;
						}  // end switch $F5xx

					}  // end else if ( pPid[0].PIDMSB == 0XF5 )

				}  // end else ... datasize != 0

			}  // end if (IsDIDSupported (EcuIndex, IdIndex) == TRUE)

		}  // end for (EcuIndex

	}  // end for (IdIndex



	if ( OBDEngineDontCare == FALSE )  // NOT Test 10.2
	{
		/**************************************************************************
		* "Either Or" required PIDs
		**************************************************************************/
		// All vehicles require Pid $F405 or $F467
		if ( fPidSupported[0x05] == FALSE &&
		     fPidSupported[0x67] == FALSE )
		{
			fReqPidNotSupported = TRUE;
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Neither PID $F405 nor $F467 supported (Support of one required for all vehicles)\n" );
		}

		// Gasoline vehicles require Pid $F40B, $F487, $F410 or $F466
		if ( gOBDDieselFlag == FALSE &&
		     ( fPidSupported[0x0B] == FALSE &&
		       fPidSupported[0x87] == FALSE &&
		       fPidSupported[0x10] == FALSE &&
		       fPidSupported[0x66] == FALSE ) )
		{
			fReqPidNotSupported = TRUE;
			Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $F40B, $F487, $F410 nor $F466 supported (Support of one required for gasoline vehicles)\n" );
		}

		// Gasoline vehicles require Pid $F40F or $F468
		if ( gOBDDieselFlag == FALSE &&
		     ( fPidSupported[0x0F] == FALSE &&
		       fPidSupported[0x68] == FALSE ) )
		{
			fReqPidNotSupported = TRUE;
			Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Neither PID $F40F nor $F468 supported (Support of one required for gasoline vehicles)\n" );
		}

		// Gasoline vehicles require Pid $F413 or $F41D
		if ( gOBDDieselFlag == FALSE &&
		     ( fPidSupported[0x13] == FALSE &&
		       fPidSupported[0x1D] == FALSE ) )
		{
			fReqPidNotSupported = TRUE;
			Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Neither PID $F413 nor $F41D supported (Support of one required for gasoline vehicles)\n" );
		}

		// all vehicles require Pid $F421 or $F44D
		if ( fPidSupported[0x21] == FALSE &&
		     fPidSupported[0x4D] == FALSE )
		{
			fReqPidNotSupported = TRUE;
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Neither PID $F421 nor PID $F44D supported (Support of one required for all vehicles)\n" );
		}

		// MY 2010 diesel vehicles require Pid $F42C or $F469
//		if ( gModelYear >= 2010 &&
//		     gOBDDieselFlag == TRUE &&
//		     ( gUserInput.eComplianceType == US_OBDII ||
//		       gUserInput.eComplianceType == HD_OBD   ||
//		       gUserInput.eComplianceType == CNOBD_WITH_IUMPR ) &&
//		     ( fPidSupported[0x2C] == FALSE &&
//		       fPidSupported[0x69] == FALSE ) )
//		{
//			fReqPidNotSupported = TRUE;
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "Neither PID $F42C nor PID $F469 supported (Support of one required for MY 2010 and beyond diesel vehicles)\n", IdIndex );
//		}

		// All vehicles require Pid $31 or $4E
		if ( fPidSupported[0x31] == FALSE &&
		     fPidSupported[0x4E] == FALSE )
		{
			fReqPidNotSupported = TRUE;
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Neither PID $F431 nor PID $F44E supported (Support of one required for all vehicles)\n" );
		}

		// Gasoline vehicles require Pid $45 or $6C
		if ( gOBDDieselFlag == FALSE &&
		     ( fPidSupported[0x45] == FALSE &&
		       fPidSupported[0x6C] == FALSE ) )
		{
//			if ( gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD )
//			{
//				fReqPidNotSupported = TRUE;
//				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//				      "Neither PID $F445 nor PID $F46C supported (Support of one required for gasoline vehicles)\n", IdIndex );
//			}
//			else
			{
				Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "Neither PID $F445 nor PID $F46C supported (Support of one required for gasoline vehicles)\n", IdIndex );
			}
		}

//		// MY 2013 diesel vehicles require PID $85 or $9B
//		if ( gModelYear >= 2013 &&
//		     gOBDDieselFlag == TRUE &&
//		     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) &&
//		     ( fPidSupported[0x85] == FALSE &&
//		       fPidSupported[0x9B] == FALSE ) )
//		{
//			fReqPidNotSupported = TRUE;
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "Neither PID $F485 nor PID $F49B supported (Support required for MY 2013 and beyond diesel vehicles)\n", IdIndex );
//		}


		/**************************************************************************
		* Individually required PIDs
		**************************************************************************/
		if ( fPidSupported[0x101] == FALSE )
		{
			// Must be supported for all vehicles
			fReqPidNotSupported = TRUE;
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $F501 not supported (Support required for all vehicles)\n" );
		}

		// if $F501 is supported, check the 'required by at least one ECU' bits
		else
		{
			// Must be supported for all vehicles
			if ( (ulIMBitsForVehicle & MIS_SUP) != MIS_SUP )
			{
				fReqPidNotSupported = TRUE;
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "PID $F501 DATA_B Bit 0 (Misfire) not supported by at least one ECU (Support required for all vehicles)\n" );
			}

			if ( (ulIMBitsForVehicle & FUEL_SUP) != FUEL_SUP )
			{
				fReqPidNotSupported = TRUE;
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "PID $F501 DATA_B Bit 1 (Fuel System) not supported by at least one ECU (Support required for all vehicles)\n" );
			}

			if ( (ulIMBitsForVehicle & COOL_SUP) != COOL_SUP )
			{
				fReqPidNotSupported = TRUE;
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "PID $F501 DATA_B Bit 3 (COOL) not supported by at least one ECU (Support required for all vehicles)\n" );
			}

			// Must be supported for all spark ignition vehicles
			if ( gOBDDieselFlag == FALSE )
			{
				if ( (ulIMBitsForVehicle & CAT_SUP) != CAT_SUP )
				{
					fReqPidNotSupported = TRUE;
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "PID $F501 DATA_C Bit 0 (Catalyst) not supported by at least one ECU (Support required for all spark ignition vehicles)\n" );
				}

				if ( (ulIMBitsForVehicle & EVAP_SUP) != EVAP_SUP )
				{
					fReqPidNotSupported = TRUE;
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "PID $F501 DATA_D Bit 0 (EVAP) not supported by at least one ECU (Support required for all spark ignition vehicles)\n" );
				}
			}
			// Must be supported for all compression ignition vehicles
			else
			{
				if ( (ulIMBitsForVehicle & HCCAT_SUP) != HCCAT_SUP )
				{
					fReqPidNotSupported = TRUE;
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "PID $F501 DATA_C Bit 1 (HCCAT) not supported by at least one ECU (Support required for all compression ignition vehicles)\n" );
				}

				if ( (ulIMBitsForVehicle & SCR_SUP) != SCR_SUP )
				{
					fReqPidNotSupported = TRUE;
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "PID $F501 DATA_C Bit 3 (SCR) not supported by at least one ECU (Support required for all compression ignition vehicles)\n" );
				}

				if ( (ulIMBitsForVehicle & PM_SUP) != PM_SUP &&
				     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) )
				{
					fReqPidNotSupported = TRUE;
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "PID $F501 DATA_D Bit 3 (PM) not supported by at least one ECU (Support required for all compression ignition vehicles)\n" );
				}
			}
		}

		for ( IdIndex = 0xF401; IdIndex <= (0xF400+MAX_PIDS); IdIndex++ )
		{
			if ( fPidSupported[IdIndex-0xF400] == FALSE )
			{
				switch ( IdIndex )
				{
					case 0xF404:
					case 0xF40C:
					case 0xF41C:
					{
						// Must be supported for all vehicles
						fReqPidNotSupported = TRUE;
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "PID $%04X not supported (Support required for all vehicles)\n",
						      IdIndex );
					}
					break;

					case 0xF406:
					case 0xF407:
					{
//						// Warn if not supported for gasoline vehicles
//						if ( gOBDDieselFlag == FALSE &&
//						     gUserInput.eComplianceType == US_OBDII != OBDBr_NO_IUMPR )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for gasoline vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF40D:
					{
						// Must be supported for all light duty vehicles
						if ( gUserInput.eVehicleType == LD )
						{
							fReqPidNotSupported = TRUE;
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "PID $%04X not supported (Support required for all light duty vehicles)\n",
							      IdIndex );
						}
						// Warn if not supported for gasoline vehicles
						else if ( gUserInput.eVehicleType == MD ||  gUserInput.eVehicleType == HD )
						{
							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "PID $%04X not supported (Support required for all medium and heavy duty vehicles)\n",
							      IdIndex );
						}
					}
					break;

					case 0xF403:
					case 0xF411:
					{
						// Must be supported for gasoline vehicles
						if ( gOBDDieselFlag == FALSE )
						{
							fReqPidNotSupported = TRUE;
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "PID $%04X not supported (Support required for gasoline vehicles)\n",
							      IdIndex );
						}
					}
					break;

					case 0xF40E:
					{
						// Must be supported for gasoline vehicles
						if ( gOBDDieselFlag == FALSE )
						{
							fReqPidNotSupported = TRUE;
							if ( gOBDHybridFlag == TRUE )
							{
								Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "PID $%04X not supported (Support required for gasoline vehicles)\n",
								      IdIndex );
							}
							else
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "PID $%04X not supported (Support required for gasoline vehicles)\n",
								      IdIndex );
							}
						}
					}
					break;

					case 0xF42F:
					{
						// Warn if not supported for all vehicles
						fReqPidNotSupported = TRUE;
						Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "PID $%04X not supported (Support required for all vehicles)\n",
						      IdIndex );
					}
					break;

//					case 0xF441:
					case 0xF443:
					{
//						// Must be supported for US_OBDII
//						if ( gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for all vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gUserInput.eComplianceType != HD_OBD )
						{
							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "PID $%04X not supported\n",
							      IdIndex );
						}
					}
					break;

					case 0xF41F:
					case 0xF430:
					case 0xF433:
					case 0xF442:
					{
						// Must be supported for all vehicles
						fReqPidNotSupported = TRUE;
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "PID $%04X not supported (Support required for all vehicles)\n",
						      IdIndex );
					}
					break;

					case 0xF42E:
					{
						// Warn if not supported for gasoline vehicles
						if ( gOBDDieselFlag == FALSE )
						{
							fReqPidNotSupported = TRUE;
							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "PID $%04X not supported (Support required for gasoline vehicles)\n",
							      IdIndex );
						}
					}
					break;

					case 0xF444:
					{
						// Must be supported for gasoline vehicles
						if ( gOBDDieselFlag == FALSE )
						{
//							if (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD)
//							{
//								fReqPidNotSupported = TRUE;
//								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//								      "PID $%04X not supported (Support required for gasoline vehicles)\n",
//								      IdIndex );
//							}
//							else
							{
								Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "PID $%04X not supported\n",
								      IdIndex );
							}
						}
					}
					break;

					case 0xF446:
					{
						// Must be supported for US vehicles
//						if ( gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for all vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF449:
					{
						// Must be supported for diesel vehicles
						if ( gOBDDieselFlag == TRUE )
						{
							fReqPidNotSupported = TRUE;
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "PID $%04X not supported (Support required for diesel vehicles)\n",
							      IdIndex );
						}
					}
					break;

					case 0xF451:
					{
						// Must be supported for MY 2015 vehicles
//						if ( gModelYear >= 2015  &&
//						     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2015 and beyond vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF45C:
					case 0xF45D:
					case 0xF45E:
					case 0xF465:
					{
						// Warn if not supported for MY 2010 diesel vehicles
//						if ( gModelYear >= 2010 &&
//						     gOBDDieselFlag == TRUE &&
//						     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) )
//						{
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2010 and beyond diesel vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF45B:
					{
						// Must be supported for MY 2013 hybrid vehicles
//						if ( gModelYear >= 2013 &&
//						     (gUserInput.ePwrTrnType == HEV || gUserInput.ePwrTrnType == PHEV) &&
//						     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2013 and beyond hybrid vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF461:
					{
						// Must be supported for MY 2010 diesel vehicles
//						if ( gModelYear >= 2010 &&
//						     gOBDDieselFlag == TRUE &&
//						     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2010 and beyond diesel vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF462:
					{
						// Must be supported for MY 2010 diesel vehicles
//						if ( gModelYear >= 2010 &&
//						     gOBDDieselFlag == TRUE &&
//						     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2010 and beyond diesel vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2019 && gModelYear < 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2021 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF463:
					{
						// Must be supported for MY 2010 heavy duty diesel vehicles
//						if ( gModelYear >= 2010 &&
//						     gOBDDieselFlag == TRUE &&
//						     gUserInput.eComplianceType == HD_OBD )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2010 and beyond heavy duty diesel vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2019 &&
//						          gOBDDieselFlag == TRUE &&
//						          gUserInput.eComplianceType == US_OBDII && gUserInput.eVehicleType == MD )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond medium duty diesel vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2019 && gModelYear < 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2021 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF47D:
					case 0xF47E:
					case 0xF481:
					{
						// Must be supported for MY 2010 heavy duty diesel vehicles
//						if ( gModelYear >= 2010 &&
//						     gOBDDieselFlag == TRUE &&
//						     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) &&
//						     (gUserInput.eVehicleType == HD || gUserInput.eVehicleType == MD) )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2010 and beyond heavy duty diesel vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF47F:
					{
						// Must be supported for MY 2010 heavy duty vehicles
//						if ( gModelYear >= 2010 &&
//						     ( gUserInput.eComplianceType == HD_OBD ||
//						       (gOBDDieselFlag == TRUE && gUserInput.eVehicleType == MD) ) )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2010 and beyond heavy duty vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF48B:
					{
						// Must be supported for MY 2013 diesel vehicles
//						if ( gModelYear >= 2013 &&
//						     gOBDDieselFlag == TRUE &&
//						     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2013 and beyond diesel vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF488:
					{
						// Must be supported for MY 2013 heavy duty diesel vehicles
//						if ( gModelYear >= 2013 &&
//						     gOBDDieselFlag == TRUE &&
//						     (gUserInput.eComplianceType == US_OBDII || gUserInput.eComplianceType == HD_OBD) &&
//						     (gUserInput.eVehicleType == HD || gUserInput.eVehicleType == MD) )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2013 and beyond heavy duty diesel vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF48E:
					{
						// Must be supported for MY 2010 heavy duty diesel vehicles
//						if ( gModelYear >= 2010 &&
//						     gOBDDieselFlag == TRUE &&
//						     gUserInput.eComplianceType == HD_OBD )
//						{
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2010 and beyond heavy duty diesel vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2019 &&
//						          gOBDDieselFlag == TRUE &&
//						          gUserInput.eComplianceType == US_OBDII && gUserInput.eVehicleType == MD )
//						{
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond heavy duty diesel vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2019 && gModelYear < 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2021 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF49A:
					{
						// Must be supported for MY 2019 PHEV vehicles
//						if ( gModelYear >= 2019 &&
//						     gOBDPlugInFlag == TRUE &&
//						     gUserInput.eComplianceType == US_OBDII )
//						{
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond plug-in hybrid electric vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF49D:
					case 0xF49E:
					{
						// Must be supported for MY 2016 heavy duty diesel vehicles
//						if ( gModelYear >= 2016 &&
//						     gOBDDieselFlag == TRUE &&
//						     gUserInput.eComplianceType == HD_OBD )
//						{
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2016 and beyond heavy duty diesel vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2019 &&
//						          gOBDDieselFlag == TRUE &&
//						          gUserInput.eComplianceType == US_OBDII && gUserInput.eVehicleType == MD )
//						{
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond medium duty diesel vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2019 && gModelYear < 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2021 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF4A2:
					{
						// Must be supported for MY 2016 heavy duty diesel vehicles
//						if ( gModelYear >= 2016 &&
//						     gOBDDieselFlag == TRUE &&
//						     gUserInput.eComplianceType == HD_OBD )
//						{
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2016 and beyond heavy duty diesel vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2019 &&
//						          gOBDDieselFlag == TRUE &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond OBD-II diesel vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF4A5:
					{
//						if ( gModelYear >= 2019 &&
//						     gOBDDieselFlag == TRUE &&
//						     gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond OBD-II diesel vehicles)\n",
//						}
					}
					break;

					case 0xF4A6:
					{
//						if ( gModelYear >= 2019 && gModelYear < 2021 &&
//						     gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( WARNING, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2019 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
//						else if ( gModelYear >= 2021 &&
//						          gUserInput.eComplianceType == US_OBDII )
//						{
//							fReqPidNotSupported = TRUE;
//							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//							      "PID $%04X not supported (Support required for MY 2021 and beyond OBD-II vehicles)\n",
//							      IdIndex );
//						}
					}
					break;

					case 0xF501:
					case 0xF502:
					case 0xF503:
					case 0xF504:
					case 0xF505:
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "PID $%04X not supported (Support required by all vehicles)\n",
						      IdIndex );
					}
					break;
					
				}  // end switch ( IdIndex )

			}  // end if ( fPidSupported[IdIndex] == FALSE )
			
			// if fPidSupported[IdIndex] == TRUE
			else
			{
				switch ( IdIndex )
				{
					case 0xF401:
					case 0xF402:
					case 0xF441:
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "PID $%04X should not be supported by a UDS vehicle.\n",
						      IdIndex );
					}
					break;
				}  // end switch ( IdIndex )
			}  // end else (fPidSupported[IdIndex] == TRUE)
		}  // end for ( IdIndex


		if ( fReqPidNotSupported == FALSE )
		{
			Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "All required PIDs supported!\n" );
		}
		else
		{
			Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Not all required PIDs are supported\n" );
		}


		// Request group support
//		if ( VerifyGroupDiagnosticSupport () == FAIL )
		if ( RequestGroupDIDSupportData ( PIDREQUEST ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Problems verifying group diagnostic support.\n" );
		}

		// Link active test to verify communication remained active for ALL protocols
		if ( VerifyLinkActive () != PASS )
		{
			return FAIL;
		}
	} // end if ( OBDEngineDontCare == FALSE)

	if ( ulInit_FailureCount != GetFailureCount () )
	{
		// There could have been early/late responses that weren't treated as FAIL
		// or PIDs out of range
		return FAIL;
	}

	return PASS;
}


/*
********************************************************************************
**
**  FUNCTION    DiplayM01P01
**
**  Purpose     Isolated function to display bit name and value of PID $F401
**
********************************************************************************
*/
void DisplayPID01 ( unsigned long EcuIndex,        // index of associated ECU
                    BOOL bCC )                     // Called from ClearCodes?
{
	PID  *pPID;

	pPID = &gEcuData[EcuIndex].PID01;

	if ( !bCC )
	{
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  PID $F401  IM Readiness Data  Byte A = $%02X  B = $%02X  C = $%02X  D = $%02X\n",
		      GetEcuId ( EcuIndex ),
		      pPID->Data[0],
		      pPID->Data[1],
		      pPID->Data[2],
		      pPID->Data[3] );

		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "         %-9s = %-7d    %-9s = %s\n",
		      "DTC_CNT",  (pPID->Data[0] & 0x7F),
		      "MIL",      (pPID->Data[0] & 0x80) ? "ON" : "OFF" );
	}
	else
	{
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  PID $F401  IM Readiness Data (from the last PID $F401 request)  Byte B = $%02X  C = $%02X  D = $%02X\n",
		      GetEcuId ( EcuIndex ),
		      pPID->Data[1],
		      pPID->Data[2],
		      pPID->Data[3] );
	}

	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      "MIS_SUP",  (pPID->Data[1] & 0x01) ? "1 (YES)" : "0 (NO)",
	      "MIS_RDY",  (pPID->Data[1] & 0x10) ? "1 (NO)" : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      "FUEL_SUP", (pPID->Data[1] & 0x02) ? "1 (YES)" : "0 (NO)",
	      "FUEL_RDY", (pPID->Data[1] & 0x20) ? "1 (NO)" : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      "CCM_SUP",  (pPID->Data[1] & 0x04) ? "1 (YES)" : "0 (NO)",
	      "CCM_RDY",  (pPID->Data[1] & 0x40) ? "1 (NO)" : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      gOBDDieselFlag ? "HCCAT_SUP" : "CAT_SUP",  (pPID->Data[2] & 0x01) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "HCCAT_RDY" : "CAT_RDY",  (pPID->Data[3] & 0x01) ? "1 (NO)" : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      gOBDDieselFlag ? "NCAT_SUP"  : "HCAT_SUP", (pPID->Data[2] & 0x02) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "NCAT_RDY"  : "HCAT_RDY", (pPID->Data[3] & 0x02) ? "1 (NO)" : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      gOBDDieselFlag ? "reserved"  : "EVAP_SUP", (pPID->Data[2] & 0x04) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "reserved"  : "EVAP_RDY", (pPID->Data[3] & 0x04) ? "1 (NO)" : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      gOBDDieselFlag ? "BP_SUP"    : "AIR_SUP",  (pPID->Data[2] & 0x08) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "BP_RDY"    : "AIR_RDY",  (pPID->Data[3] & 0x08) ? "1 (NO)" : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      gOBDDieselFlag ? "EGS_SUP"   : "O2S_SUP",  (pPID->Data[2] & 0x20) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "EGS_RDY"   : "O2S_RDY",  (pPID->Data[3] & 0x20) ? "1 (NO)" : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %-7s    %-9s = %s\n",
	      gOBDDieselFlag ? "PM_SUP"    : "HTR_SUP",  (pPID->Data[2] & 0x40) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "PM_RDY"    : "HTR_RDY",  (pPID->Data[3] & 0x40) ? "1 (NO)" : "0 (YES)" );

	if ( !bCC )
	{
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "         %-9s = %-7s    %-9s = %s\n",
		      "EGR_SUP",  (pPID->Data[2] & 0x80) ? "1 (YES)" : "0 (NO)",
		      "EGR_RDY",  (pPID->Data[3] & 0x80) ? "1 (NO)" : "0 (YES)" );
	}
}


/*
********************************************************************************
**  FUNCTION    VerifyPID01
**
**  Purpose
********************************************************************************
*/
STATUS VerifyPID01 ( PID           *pPid,              // pointer to PID Data
                     unsigned long  EcuIndex,          // index of associated ECU
                     unsigned long *pulBitsSupported ) // bit map of bits that must be set on at least one ECU
{
	BOOL bTestFailed = FALSE;

	DisplayPID01 ( EcuIndex, FALSE );

	// Check if the MIL light is ON (PID $F401 Data A Bit 7)
	if ( pPid[0].Data[0] & 0x80 )
	{
		if ( gOBDDTCStored == FALSE )
		{
			// MIL is ON when there should not be a stored DTC
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  MIL status failure.\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Check if any DTC status bits (PID $F401 Data A Bits 0-6)
	if ( pPid[0].Data[0] & 0x7F )
	{
		// if DTC status bit(s) set when there should not be a stored DTC
		if ( gOBDDTCStored == FALSE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  DTC status failure.\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data B, BIT 0 - Misfire support
	if ( (pPid[0].Data[1] & 0x01) == 0x01 )
	{
		*pulBitsSupported = *pulBitsSupported | DATA_B_BIT_0;
	}

	// Evaluate Data B, BIT 1 - Fuel support
	if ( (pPid[0].Data[1] & 0x02) == 0x02 )
	{
		*pulBitsSupported = *pulBitsSupported | DATA_B_BIT_1;
	}

	// Evaluate Data B, BIT 2 - CCM support
	if ( (pPid[0].Data[1] & 0x04) != 0x04 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  CCM monitor must be supported (CCM_SUP = 1 (YES))\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Check Data B BIT 3 state
	// only check if MY 2010 and beyond and ECU does not only supports CCM requirements (PID $F401 Data B bit 2==1)
	if ( gModelYear >= 2010 &&
	     ( (gEcuData[EcuIndex].PID01.Data[1] & 0x03) != 0x00 ||
	       (gEcuData[EcuIndex].PID01.Data[2] & 0xFF) != 0x00 ) )
	{
		if ( (pPid[0].Data[1] & 0x08) == 0x00 &&
		     gOBDDieselFlag == 1  )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Data B bit 3 clear (Spark Ignition) does not match user input (Diesel)\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
		else if ( (pPid[0].Data[1] & 0x08) == 0x08 &&
		          gOBDDieselFlag == 0 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Data B bit 3 set (Compression Ignition) does not match user input (Non Diesel)\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data B, BIT 4 - Misfire
	if ( (pPid[0].Data[1] & 0x01) == 0x01 )
	 {
		// if Misfire is supported
		if ( gOBDDieselFlag == FALSE )
		{
			// if it's a spark engine
			if ( (pPid[0].Data[1] & 0x10) != 0x00 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Misfire Monitor must be 0 (MIS_SUP = NO) for spark ignition vehicles or controllers that do not support misfire.\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
		}
		else
		{
			// Bit 4 may be 0 or 1 for compression ignition w/ or wo/ Engine running
			// except in Test 11 where it must be complete (0).
			if ( gTestPhase == eTestPerformanceCounters )
			{
				if ( (pPid[0].Data[1] & 0x10) != 0x00 )
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  Misfire Monitor must be 0 (MIS_SUP = NO) for compression ignition vehicles.\n",
					      GetEcuId ( EcuIndex ) );
					bTestFailed = TRUE;
				}
			}
		}
	}
	// If Misfire unsupported it must indicate complete
	if ( (pPid[0].Data[1] & 0x01) == 0x00 &&
	     (pPid[0].Data[1] & 0x10) == 0x10 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Misfire Monitor (MIS_SUP = 0 (NO) ) must indicate complete (MIS_RDY = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data B, BIT 5 - Fuel status
	if ( (pPid[0].Data[1] & 0x02 ) == 0x02 )
	{
		// only check for tests 11.3 and 11.7 (not 10.6)
		if ( gTestPhase == eTestPerformanceCounters &&
		     (pPid[0].Data[1] & 0x20) == 0x20 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported Fuel System Monitor (FUEL_SUP = 1 (YES)) must indicate complete (FUEL_RDY = 0 (YES)).\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}
	// If fuel system monitor unsupported it must indicate complete
	if ( (pPid[0].Data[1] & 0x02) == 0x00 &&
	     (pPid[0].Data[1] & 0x20) == 0x20 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Fuel System Monitor (FUEL_SUP = 0 (NO)) must indicate complete (FUEL_RDY = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data B, BIT 6 - CCM status
	if ( (pPid[0].Data[1] & 0x40) == 0x40 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  CCM must indicate complete (CCM_RDY = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Check Data B bit 7 if reserved bit is set
	if ( pPid[0].Data[1] & 0x80 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Reserved I/M readiness Data B bit 7 set\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data C, BIT 0
	if ( pPid[0].Data[2] & 0x01 )
	{
		*pulBitsSupported = *pulBitsSupported | DATA_C_BIT_0;
	}

	// Evaluate Data C, BIT 2
	if ( pPid[0].Data[2] & 0x04 )
	{
		if ( gOBDDieselFlag != TRUE /*&&
		     (gUserInput.eComplianceType == US_OBDII ||
		      gUserInput.eComplianceType == HD_OBD   ||
		      gUserInput.eComplianceType == CNOBD_WITH_IUMPR )*/ )
		{
			*pulBitsSupported = *pulBitsSupported | DATA_C_BIT_2;
		}
		else if ( gOBDDieselFlag == TRUE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data C bit 2 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data C, BIT 3
	if ( pPid[0].Data[2] & 0x08 )
	{
		*pulBitsSupported = *pulBitsSupported | DATA_C_BIT_3;
	}

	// Evaluate Data C, BIT 4
	if ( pPid[0].Data[2] & 0x10 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Reserved I/M readiness Data C bit 4 set\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data C, BIT 5
	if ( pPid[0].Data[2] & 0x20 )
	{
		*pulBitsSupported = *pulBitsSupported | DATA_C_BIT_5;
	}

	// Evaluate Data C, BIT 6
	if ( pPid[0].Data[2] & 0x40 )
	{
		*pulBitsSupported = *pulBitsSupported | DATA_C_BIT_6;
	}

	// Evaluate Data C, BIT 7
	if ( pPid[0].Data[2] & 0x80 )
	{
		*pulBitsSupported = *pulBitsSupported | DATA_C_BIT_7;
	}

	// Evaluate Data D, BIT 2
	if ( (pPid[0].Data[3] & 0x04) == 0x04 &&
	     gOBDDieselFlag == TRUE )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Reserved I/M readiness Data D bit 2 set\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data D, BIT 4
	if ( pPid[0].Data[3] & 0x10 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Reserved I/M readiness Data D bit 4 set\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	if ( gOBDEngineRunning == FALSE && gTestPhase != eTestPerformanceCounters )  // test 5.6 & 10.6
	{
		//--------------------------------------------------------------------------
		// supported monitors should not be complete (except O2 on spark ignition)

		// DATA_D Bit 0
		if ( (pPid[0].Data[2] & 0x01) == 0x01 &&
		     (pPid[0].Data[3] & 0x01) == 0x00 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
			      GetEcuId ( EcuIndex ),
			      gOBDDieselFlag ? "HCCAT" : "CAT",
			      gOBDDieselFlag ? "HCCAT" : "CAT" );
			bTestFailed = TRUE;
		}

		// DATA_D Bit 1
		if ( (pPid[0].Data[2] & 0x02) == 0x02 &&
		     (pPid[0].Data[3] & 0x02) == 0x00 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
			      GetEcuId ( EcuIndex ),
			      gOBDDieselFlag ? "NCAT" : "HCAT",
			      gOBDDieselFlag ? "NCAT" : "HCAT" );
			bTestFailed = TRUE;
		}

		// DATA_D Bit 2
		if ( (pPid[0].Data[2] & 0x04) == 0x04 &&
		     (pPid[0].Data[3] & 0x04) == 0x00 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
			      GetEcuId ( EcuIndex ),
			      gOBDDieselFlag ? "RESERVED2" : "EVAP",
			      gOBDDieselFlag ? "RESERVED2" : "EVAP" );
			bTestFailed = TRUE;
		}

		// DATA_D Bit 3
		if ( (pPid[0].Data[2] & 0x08) == 0x08 &&
		     (pPid[0].Data[3] & 0x08) == 0x00 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
			      GetEcuId ( EcuIndex ),
			      gOBDDieselFlag ? "BP" : "AIR",
			      gOBDDieselFlag ? "BP" : "AIR" );
			bTestFailed = TRUE;
		}

		// DATA_D Bit 4
		if ( (pPid[0].Data[2] & 0x10) == 0x10 &&
		     (pPid[0].Data[3] & 0x10) == 0x00 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitor (RESERVED Byte C Bit 4 = 1) must indicate incomplete (RESERVED Byte D Bit 4 = 1)\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}

		// DATA_D Bit 5
		if ( (pPid[0].Data[2] & 0x20) == 0x20 &&
		     (pPid[0].Data[3] & 0x20) == 0x00 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
			      GetEcuId ( EcuIndex ),
			      gOBDDieselFlag ? "EGS" : "O2S",
			      gOBDDieselFlag ? "EGS" : "O2S" );
			bTestFailed = TRUE;
		}

		// DATA_D Bit 6
		if ( (pPid[0].Data[2] & 0x40) == 0x40 &&
		     (pPid[0].Data[3] & 0x40) == 0x00 &&
		     gOBDDieselFlag )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
			      GetEcuId ( EcuIndex ),
			      gOBDDieselFlag ? "PM" : "HTR",
			      gOBDDieselFlag ? "PM" : "HTR" );
			bTestFailed = TRUE;
		}

		// DATA_D Bit 7
		if ( (pPid[0].Data[2] & 0x80) == 0x80 &&
		     (pPid[0].Data[3] & 0x80) == 0x00 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitor (EGR_SUP = 1 (YES)) must indicate incomplete (EGR_RDY = 1 (NO))\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}


	//--------------------------------------------------------------------------
	// unsupported monitors should be complete

	// DATA_D Bit 0
	if ( (pPid[0].Data[2] & 0x01) == 0x00 &&
	     (pPid[0].Data[3] & 0x01) == 0x01 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported monitor (%s_SUP = N) must indicate complete (%s_RDY = 0 (YES))\n",
		      GetEcuId ( EcuIndex ),
		      gOBDDieselFlag ? "HCCAT" : "CAT",
		      gOBDDieselFlag ? "HCCAT" : "CAT" );
		bTestFailed = TRUE;
	}

	// DATA_D Bit 1
	if ( (pPid[0].Data[2] & 0x02) == 0x00 &&
	     (pPid[0].Data[3] & 0x02) == 0x02 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported monitor (%s_SUP = 0 (NO)) must indicate complete (%s_RDY = 0 (YES))\n",
		      GetEcuId ( EcuIndex ),
		      gOBDDieselFlag ? "NCAT" : "HCAT",
		      gOBDDieselFlag ? "NCAT" : "HCAT" );
		bTestFailed = TRUE;
	}

	// DATA_D Bit 2
	if ( (pPid[0].Data[2] & 0x04) == 0x00 &&
	     (pPid[0].Data[3] & 0x04) == 0x04 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported monitor (%s_SUP = 0 (NO)) must indicate complete (%s_RDY = 0 (YES))\n",
		      GetEcuId ( EcuIndex ),
		      gOBDDieselFlag ? "RESERVED2" : "EVAP",
		      gOBDDieselFlag ? "RESERVED2" : "EVAP" );
		bTestFailed = TRUE;
	}

	// DATA_D Bit 3
	if ( (pPid[0].Data[2] & 0x08) == 0x00 &&
	     (pPid[0].Data[3] & 0x08) == 0x08 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported monitor (%s_SUP = 0 (NO)) must indicate complete (%s_RDY = 0 (YES))\n",
		      GetEcuId ( EcuIndex ),
		      gOBDDieselFlag ? "BP" : "AIR",
		      gOBDDieselFlag ? "BP" : "AIR" );
		bTestFailed = TRUE;
	}

	// DATA_D Bit 4
	if ( (pPid[0].Data[2] & 0x10) == 0x00 &&
	     (pPid[0].Data[3] & 0x10) == 0x10 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported monitor (RESERVED Byte C Bit 4 = 0) must indicate complete (RESERVED Byte D Bit 4 = 0)\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// DATA_D Bit 5
	if ( (pPid[0].Data[2] & 0x20) == 0x00 &&
	     (pPid[0].Data[3] & 0x20) == 0x20 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported monitor (%s_SUP = 0 (NO)) must indicate complete (%s_RDY = 0 (YES))\n",
		      GetEcuId ( EcuIndex ),
		      gOBDDieselFlag ? "EGS" : "O2S",
		      gOBDDieselFlag ? "EGS" : "O2S" );
		bTestFailed = TRUE;
	}

	// DATA_D Bit 6
	if ( (pPid[0].Data[2] & 0x40) == 0x00 &&
	     (pPid[0].Data[3] & 0x40) == 0x40 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported monitor (%s_SUP = 0 (NO)) must indicate complete (%s_RDY = 0 (YES))\n",
		      GetEcuId ( EcuIndex ),
		      gOBDDieselFlag ? "PM" : "HTR",
		      gOBDDieselFlag ? "PM" : "HTR" );
		bTestFailed = TRUE;
	}

	// DATA_D Bit 7
	if ( (pPid[0].Data[2] & 0x80) == 0x00 &&
	     (pPid[0].Data[3] & 0x80) == 0x80 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X : Unsupported monitor (EGR_SUP = 0 (NO)) must indicate complete (EGR_RDY = 0 (YES))\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	if ( gTestPhase == eTestPerformanceCounters )
	{
		if ( pPid[0].Data[3] != 0 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported monitors (SUP = 1 (YES)) must be complete (RDY = 0 (YES)) after I/M drive cycle\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}
	else
	{
		return PASS;
	}
}


/*
********************************************************************************
**  FUNCTION    VerifyPID1C
**
**  Purpose
********************************************************************************
*/
STATUS VerifyPID1C ( PID          *pPid,        // pointer to PID Data
                     unsigned long EcuIndex )   // index of associated ECU
{
	BOOL bTestFailed = FALSE;


	// Make sure value is in the valid range before the lookup
	if ( pPid[0].Data[0] > MAX_OBD_TYPES )
	{
		pPid[0].Data[0] = 0;
	}

	// Vehicle should support OBD-II
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "ECU %X  PID $F41C  OBDSUP   = $%02X (%s)\n",
	      GetEcuId ( EcuIndex ),
	      pPid[0].Data[0],
	      OBD_TYPE[pPid[0].Data[0]] );

	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		if ( gUserInput.eComplianceType == US_OBDII )
		{
			if ( pPid[0].Data[0] != 0x01 &&     // CARB OBD II
			     pPid[0].Data[0] != 0x03 &&     // OBDI and OBD II
			     pPid[0].Data[0] != 0x07 &&     // EOBD and OBD II
			     pPid[0].Data[0] != 0x09 &&     // EOBD, OBD and OBD II
			     pPid[0].Data[0] != 0x0B &&     // JOBD and OBD II
			     pPid[0].Data[0] != 0x0D &&     // JOBD, EOBD, and OBD II
			     pPid[0].Data[0] != 0x0F &&
			     pPid[0].Data[0] != 0x22 )      // OBD, OBD II and HD_OBD
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Not a Light Duty OBD II ECU\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
		}
		else if ( gUserInput.eComplianceType == HD_OBD )
		{
			if ( pPid[0].Data[0] != 0x13 &&     // HD OBD-C
			     pPid[0].Data[0] != 0x14 &&     // HD OBD
			     pPid[0].Data[0] != 0x22 )      // OBD, OBD II and HD_OBD
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Not a Heavy Duty OBD II ECU\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
		}
		else if ( gUserInput.eComplianceType == EOBD_WITH_IUMPR ||
		          gUserInput.eComplianceType == EOBD_NO_IUMPR )
		{
			if ( pPid[0].Data[0] != 0x06 &&     // EOBD
			     pPid[0].Data[0] != 0x07 &&     // EOBD and OBD II
			     pPid[0].Data[0] != 0x08 &&     // EOBD and OBD
			     pPid[0].Data[0] != 0x09 &&     // EOBD, OBD and OBD II
			     pPid[0].Data[0] != 0x0C &&     // JOBD and EOBD
			     pPid[0].Data[0] != 0x0D &&     // JOBD, EOBD, and OBD II
			     pPid[0].Data[0] != 0x0E &&     // EURO IV B1
			     pPid[0].Data[0] != 0x0F )      // EURO V B2
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Not a Light Duty EOBD ECU\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
		}
		else if ( gUserInput.eComplianceType == HD_EOBD )
		{
			if ( pPid[0].Data[0] != 0x17 &&     // HD EOBD-I
			     pPid[0].Data[0] != 0x18 &&     // HD EOBD-I N
			     pPid[0].Data[0] != 0x19 &&     // HD EOBD-II
			     pPid[0].Data[0] != 0x1A &&     // HD EOBD-II N
			     pPid[0].Data[0] != 0x21 )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Not a Heavy Duty EOBD ECU\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
		}
		else if ( gUserInput.eComplianceType == IOBD_NO_IUMPR || gUserInput.eComplianceType == HD_IOBD_NO_IUMPR )
		{
			if ( pPid[0].Data[0] != 0x20 )      // IOBD
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Not a IOBD II ECU\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
		}
		else if ( gUserInput.eComplianceType == OBDBr_NO_IUMPR )
		{
			if ( pPid[0].Data[0] != 0x1C &&    // OBDBr-1
			     pPid[0].Data[0] != 0x1D &&    // OBDBr-2
			     pPid[0].Data[0] != 0x23 &&    // OBDBr-3
			     pPid[0].Data[0] != 0x2A )     // OBDBr-D
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Not a OBDBr ECU\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
		}
		else if ( gUserInput.eComplianceType == CNOBD_WITH_IUMPR )
		{
			if ( pPid[0].Data[0] != 0x29 )     // CN-OBD-6
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Not a China6 ECU\n",
				      GetEcuId ( EcuIndex ) );
				bTestFailed = TRUE;
			}
		}
	}

	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}
	else
	{
		return PASS;
	}
}


/*
********************************************************************************
**  FUNCTION    VerifyPID41
**
**  Purpose     Isolated function to verify PID $F441 data
********************************************************************************
*/
STATUS VerifyPID41 ( PID          *pPid,               // pointer to PID Data
                     unsigned long EcuIndex,            // index of associated ECU
                     unsigned char OBDEngineDontCare )  // copy of OBDEngineDontCare flag from calling routine
{
#define  MAX_STRING_NUMBER  8
#define  MAX_NAME_LENGTH    12

	BOOL bTestFailed = FALSE;
	unsigned int ByteIndex;
	unsigned int BitIndex;
	unsigned int BitMask;
	unsigned int BitMask2;
	char String[MAX_STRING_NUMBER][MAX_NAME_LENGTH];
	unsigned char StringIndex;


	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "ECU %X  PID $F441  Monitor Status  Byte A = $%02X  B = $%02X  C = $%02X  D = $%02X\n",
	      GetEcuId ( EcuIndex ),
	      pPid[0].Data[0],
	      pPid[0].Data[1],
	      pPid[0].Data[2],
	      pPid[0].Data[3] );

	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      "MIS_ENA",   (pPid[0].Data[1] & 0x01) ? "1 (YES)" : "0 (NO)",
	      "MIS_CMPL",  (pPid[0].Data[1] & 0x10) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      "FUEL_ENA",  (pPid[0].Data[1] & 0x02) ? "1 (YES)" : "0 (NO)",
	      "FUEL_CMPL", (pPid[0].Data[1] & 0x20) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      "CCM_ENA",   (pPid[0].Data[1] & 0x04) ? "1 (YES)" : "0 (NO)",
	      "CCM_CMPL",  (pPid[0].Data[1] & 0x40) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      gOBDDieselFlag ? "HCCAT_ENA"  : "CAT_ENA",  (pPid[0].Data[2] & 0x01) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "HCCAT_CMPL" : "CAT_CMPL", (pPid[0].Data[3] & 0x01) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      gOBDDieselFlag ? "NCAT_ENA"  : "HCAT_ENA",  (pPid[0].Data[2] & 0x02) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "NCAT_CMPL" : "HCAT_CMPL", (pPid[0].Data[3] & 0x02) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      gOBDDieselFlag ? "reserved" : "EVAP_ENA",   (pPid[0].Data[2] & 0x04) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "reserved" : "EVAP_CMPL",  (pPid[0].Data[3] & 0x04) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      gOBDDieselFlag ? "BP_ENA"  : "AIR_ENA",     (pPid[0].Data[2] & 0x08) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "BP_CMPL" : "AIR_CMPL",    (pPid[0].Data[3] & 0x08) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      gOBDDieselFlag ? "EGS_ENA"  : "O2S_ENA",    (pPid[0].Data[2] & 0x20) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "EGS_CMPL" : "O2S_CMPL",   (pPid[0].Data[3] & 0x20) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      gOBDDieselFlag ? "PM_ENA"  : "HTR_ENA",     (pPid[0].Data[2] & 0x40) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "PM_CMPL" : "HTR_CMPL",    (pPid[0].Data[3] & 0x40) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "         %-9s = %s    %-10s = %s\n",
	      "EGR_ENA",   (pPid[0].Data[2] & 0x80) ? "1 (YES)" : "0 (NO)",
	      "EGR_CMPL",  (pPid[0].Data[3] & 0x80) ? "1 (NO)"  : "0 (YES)" );




	// check that bits set in PID $41 match support in PID $01
	for ( ByteIndex = 1; ByteIndex <= 3; ByteIndex++ )
	{
		if ( ByteIndex == 1 )
		{
			strcpy_s ( String[0], MAX_NAME_LENGTH, "MIS" );
			strcpy_s ( String[1], MAX_NAME_LENGTH, "FUEL" );
			strcpy_s ( String[2], MAX_NAME_LENGTH, "CCM" );
			strcpy_s ( String[3], MAX_NAME_LENGTH, "no display" );
			strcpy_s ( String[4], MAX_NAME_LENGTH, "MIS" );
			strcpy_s ( String[5], MAX_NAME_LENGTH, "FUEL" );
			strcpy_s ( String[6], MAX_NAME_LENGTH, "CCM" );
			strcpy_s ( String[7], MAX_NAME_LENGTH, "reserved" );

			BitMask2 = 0x01;
		}
		else
		{
			strcpy_s ( String[0], MAX_NAME_LENGTH, gOBDDieselFlag ? "HCCAT"    : "CAT" );
			strcpy_s ( String[1], MAX_NAME_LENGTH, gOBDDieselFlag ? "NCAT"     : "HCAT" );
			strcpy_s ( String[2], MAX_NAME_LENGTH, gOBDDieselFlag ? "reserved" : "EVAP" );
			strcpy_s ( String[3], MAX_NAME_LENGTH, gOBDDieselFlag ? "BP"       : "AIR" );
			strcpy_s ( String[4], MAX_NAME_LENGTH, "reserved" );
			strcpy_s ( String[5], MAX_NAME_LENGTH, gOBDDieselFlag ? "EGS"      : "O2S" );
			strcpy_s ( String[6], MAX_NAME_LENGTH, gOBDDieselFlag ? "PM"       : "HTR" );
			strcpy_s ( String[7], MAX_NAME_LENGTH, "EGR" );

			BitMask2 = 0x01;
		}


		for ( BitIndex = 0, BitMask = 0x01, StringIndex = 0; BitIndex < 8; BitIndex++, BitMask<<=1, StringIndex++ )
		{
			// reserved bit check
			if ( ( ByteIndex == 1 && BitIndex == 7 && (pPid[0].Data[1] & BitMask) == BitMask ) ||  // reserved bit is set
			     ( ByteIndex == 2 && BitIndex == 4 && (pPid[0].Data[2] & BitMask) == BitMask ) ||  // reserved bit is set
			     ( ByteIndex == 3 && BitIndex == 4 && (pPid[0].Data[3] & BitMask) == BitMask ) )   // reserved bit is set
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  PID $F441  Reserved Bit (Byte %c Bit %d) set (1) (must be clear (0))\n",
				      GetEcuId ( EcuIndex ),
				      'A'+ ByteIndex,
				      BitIndex );
			}

			// Compression Ignition (Diesel) bit check
			// Compression bit should match user input
			else if ( ByteIndex == 1 && BitIndex == 3 )
			{
				// only check if MY 2010 and beyond and ECU does not only supports CCM requirements (PID $F401 Data B bit 2==1)
				if ( gModelYear >= 2010 &&
				     ( (gEcuData[EcuIndex].PID01.Data[1] & 0x03) != 0x00 ||
				       (gEcuData[EcuIndex].PID01.Data[2] & 0xFF) != 0x00 ) )
				{
					if ( (pPid[0].Data[1] & 0x08) == 0x00 &&
					     gOBDDieselFlag == 1 )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  PID $F441  Data B bit 3 clear (Spark Ignition) does not match user input (Diesel)\n",
						      GetEcuId ( EcuIndex ) );
						bTestFailed = TRUE;
					}
					else if ( (pPid[0].Data[1] & 0x08) == 0x08 &&
					          gOBDDieselFlag == FALSE )
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  PID $F441  Data B bit 3 set (Compression Ignition) does not match user input (Non Diesel)\n",
						      GetEcuId ( EcuIndex ) );
						bTestFailed = TRUE;
					}
				}
			}

			// Enable/Support bit check
			// enabled monitors should be supported
			else if ( ( ( ByteIndex == 1 && BitIndex >= 0 && BitIndex <=2 ) ||
			            ByteIndex == 2 ) &&
			          (pPid[0].Data[ByteIndex] & BitMask) == BitMask  &&      // enabled       in PID $41
			          (gEcuData[EcuIndex].PID01.Data[ByteIndex] & BitMask) == 0x00 )  // not supported in PID $01
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  If monitor enabled (%s_ENA = 1 (YES)) in PID $F441 it must indicate supported (%s_SUP = 1 (YES)) in PID $F401\n",
				      GetEcuId ( EcuIndex ),
				      String[StringIndex],
				      String[StringIndex] );
				bTestFailed = TRUE;
			}


			else
			{
				// Incomplete/Supported bit check
				// incomplete monitors should be supported
				if ( (pPid[0].Data[ByteIndex] & BitMask) == BitMask &&                                                           // not complete  in PID $41
				     ( ( ByteIndex == 1 && BitIndex >= 4 && BitIndex <=6 && (gEcuData[EcuIndex].PID01.Data[1] & BitMask2) == 0x00 ) ||   // not supported in PID $01
				       ( ByteIndex == 3 &&                                  (gEcuData[EcuIndex].PID01.Data[2] & BitMask)  == 0x00 ) ) )  // not supported in PID $01
				{
					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
					      "ECU %X  If monitor not complete (%s_CMPL = 1 (NO)) in DID $4F41 it must indicate supported (%s_SUP = 1 (YES)) in PID $F401\n",
					      GetEcuId ( EcuIndex ),
					      String[StringIndex],
					      String[StringIndex] );
					bTestFailed = TRUE;
				}
				if ( ByteIndex == 1 && BitIndex >= 4 && BitIndex <=6 )
				{
					BitMask2<<=1;
				}


				// Byte 3 (D) only checks */
				if ( ByteIndex == 3 )
				{
					// Supported/Incomplete bit check
					// supported monitors should be incomplete (when engine off)
					if ( OBDEngineDontCare == FALSE && gOBDEngineRunning == FALSE &&  // test 5.6 ONLY; NOT Test 5.10, 10.2 & 10.13
					     (gEcuData[EcuIndex].PID01.Data[2] & BitMask) == BitMask &&   // supported in PID $01
					     (gEcuData[EcuIndex].PID01.Data[3] & BitMask) == 0x00 )       // complete  in PID $41
					{
						if ( BitIndex != 6 || gOBDDieselFlag == TRUE )  // O2 Heater monitor is allowed to complete for gasoline vehicles
						{
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "ECU %X  If monitor supported (%s_SUP = 1 (YES)) in PID $F401 it must indicate incomplete (%s_CMPL = 1 (NO)) in PID $F441\n",
							      GetEcuId ( EcuIndex ),
							      String[StringIndex],
							      String[StringIndex] );
							bTestFailed = TRUE;
						}
					}


					// Unupported/Complete bit check */
					// unsupported monitors should be complete
					if ( (gEcuData[EcuIndex].PID01.Data[2] & BitMask) == 0x00 &&    // not supported in PID $01
					     (pPid[0].Data[3] & BitMask) == BitMask )           // not complete  in PID $41
					{
						Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
						      "ECU %X  If monitor unsupported (%s_SUP = 0 (NO)) in PID $F401 it must indicate complete (%s_CMPL = 0 (YES)) in PID $F441\n",
						      GetEcuId ( EcuIndex ),
						      String[StringIndex],
						      String[StringIndex] );
						bTestFailed = TRUE;
					}
				}
			}
		}
	}


	if (bTestFailed == TRUE)
	{
		return FAIL;
	}
	else
	{
		return PASS;
	}
}


/*
********************************************************************************
**  FUNCTION    VerifyPIDF501
**
**  Purpose
********************************************************************************
*/
void DisplayPIDF501 ( unsigned long EcuIndex,  // index of associated ECU
                      BOOL bCC )               // Set to TRUE if called from ClearCodes
{
	PID           *pPid;          // pointer to PID Data

	pPid = &gEcuData[EcuIndex].PIDF501;

	if ( !bCC )
	{
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  PID $F501  Byte A = $%02X  B = $%02X  C = $%02X  D = $%02X  E = $%02X  F = $%02X\n",
		      GetEcuId ( EcuIndex ),
		      pPid->Data[0],
		      pPid->Data[1],
		      pPid->Data[2],
		      pPid->Data[3],
		      pPid->Data[4],
		      pPid->Data[5] );
	
		// $F501 Byte A (0) decode
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "                    %-9s = %-7d    %-10s = %s\n",
		      "DTC_CNT",  (pPid->Data[0] & 0x7F),
		      "MIL",      (pPid->Data[0] & 0x80) ? "ON" : "OFF" );
	}
	else
	{
		Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  PID $F501  IM Readiness Data (from the last PID $F501 request)  B = $%02X  C = $%02X  D = $%02X  E = $%02X  F = $%02X\n",
		      GetEcuId ( EcuIndex ),
		      pPid->Data[1],
		      pPid->Data[2],
		      pPid->Data[3],
		      pPid->Data[4],
		      pPid->Data[5] );
	}

	// $F501 Byte B (1) decode
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "MIS_SUP",   (pPid->Data[1] & 0x01) ? "1 (YES)" : "0 (NO)",
	      "MIS_CMPL",  (pPid->Data[1] & 0x10) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "FUEL_SUP",  (pPid->Data[1] & 0x02) ? "1 (YES)" : "0 (NO)",
	      "FUEL_CMPL", (pPid->Data[1] & 0x20) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "CCM_SUP",   (pPid->Data[1] & 0x04) ? "1 (YES)" : "0 (NO)",
	      "CCM_CMPL",  (pPid->Data[1] & 0x40) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "COOL_SUP",  (pPid->Data[1] & 0x08) ? "1 (YES)" : "0 (NO)",
	      "COOL_CMPL", (pPid->Data[1] & 0x80) ? "1 (NO)"  : "0 (YES)" );

	// $F501 Byte C (2) decode
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "reserved" : "CAT_SUP",    (pPid->Data[2] & 0x01) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "reserved" : "CAT_CMPL",   (pPid->Data[2] & 0x10) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "HCCAT_SUP"  : "reserved", (pPid->Data[2] & 0x02) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "HCCAT_CMPL" : "reserved", (pPid->Data[2] & 0x20) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "reserved" : "HCAT_SUP",   (pPid->Data[2] & 0x04) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "reserved" : "HCAT_CMPL",  (pPid->Data[2] & 0x40) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "SCR_SUP"  : "reserved",   (pPid->Data[2] & 0x08) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "SCR_CMPL" : "reserved",   (pPid->Data[2] & 0x80) ? "1 (NO)"  : "0 (YES)" );

	// $F501 Byte D (3) decode
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "reserved" : "EVAP_SUP",   (pPid->Data[3] & 0x01) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "reserved" : "EVAP_CMPL",  (pPid->Data[3] & 0x10) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "reserved" : "AIR_SUP",    (pPid->Data[3] & 0x02) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "reserved" : "AIR_CMPL",   (pPid->Data[3] & 0x20) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "BP_SUP"  : "reserved",    (pPid->Data[3] & 0x04) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "BP_CMPL" : "reserved",    (pPid->Data[3] & 0x40) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "PM_SUP"  : "reserved",    (pPid->Data[3] & 0x08) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "PM_CMPL" : "reserved",    (pPid->Data[3] & 0x80) ? "1 (NO)"  : "0 (YES)" );

	// $F501 Byte E (4) decode
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "EGS_SUP",    (pPid->Data[4] & 0x01) ? "1 (YES)" : "0 (NO)",
	      "EGS_CMPL",   (pPid->Data[4] & 0x10) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "PCVCV_SUP",  (pPid->Data[4] & 0x02) ? "1 (YES)" : "0 (NO)",
	      "PCVCV_CMPL", (pPid->Data[4] & 0x20) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "EGR_SUP",    (pPid->Data[4] & 0x04) ? "1 (YES)" : "0 (NO)",
	      "EGR_CMPL",   (pPid->Data[4] & 0x40) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "VVT_SUP",    (pPid->Data[4] & 0x08) ? "1 (YES)" : "0 (NO)",
	      "VVT_CMPL",   (pPid->Data[4] & 0x80) ? "1 (NO)"  : "0 (YES)" );
	
	// $F501 Byte F (5) decode
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "DOR_SUP",    (pPid->Data[5] & 0x01) ? "1 (YES)" : "0 (NO)",
	      "DOR_CMPL",   (pPid->Data[5] & 0x10) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "CSERS_SUP",  (pPid->Data[5] & 0x02) ? "1 (YES)" : "0 (NO)",
	      "CSERS_CMPL", (pPid->Data[5] & 0x20) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      gOBDDieselFlag ? "NAC_SUP"  : "reserved",  (pPid->Data[5] & 0x04) ? "1 (YES)" : "0 (NO)",
	      gOBDDieselFlag ? "NAC_CMPL" : "reserved",  (pPid->Data[5] & 0x40) ? "1 (NO)"  : "0 (YES)" );
	Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "                    %-9s = %-7s    %-10s = %s\n",
	      "ERS_SUP",    (pPid->Data[5] & 0x08) ? "1 (YES)" : "0 (NO)",
	      "ERS_CMPL",   (pPid->Data[5] & 0x80) ? "1 (NO)"  : "0 (YES)" );
}


/*
********************************************************************************
**  FUNCTION    VerifyPIDF501
**
**  Purpose
********************************************************************************
*/
STATUS VerifyPIDF501 ( PID *pPid,
                       unsigned long EcuIndex,           // index of associated ECU
                       unsigned long *pulBitsSupported ) // bit map of bits that must be set on at least one ECU
{
	BOOL bTestFailed = FALSE;

	DisplayPIDF501 ( EcuIndex, FALSE );

	// Check if the MIL light is ON (PID $F501 Data A Bit 7)
	if ( pPid[0].Data[0] & 0x80 )
	{
		if ( gOBDDTCStored == FALSE )
		{
			// MIL is ON when there should not be a stored DTC
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  MIL status failure.\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Check for DTC Count (PID $F501 Data A Bits 0-6)
	if ( pPid[0].Data[0] & 0x7F )
	{
		// If DTC status bit(s) set when there should not be a stored DTC
		if ( gOBDDTCStored == FALSE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  DTC status failure.\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data B, BIT 0 - Misfire support
	if ( (pPid[0].Data[1] & 0x01) == 0x01)
	{
		*pulBitsSupported = *pulBitsSupported | MIS_SUP;
	}

	// Evaluate Data B, BIT 1 - Fuel support
	if ( (pPid[0].Data[1] & 0x02) == 0x02)
	{
		*pulBitsSupported = *pulBitsSupported | FUEL_SUP;
	}

	// Evaluate Data B, BIT 2 - CCM support
	if ( (pPid[0].Data[1] & 0x04) != 0x04 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  CCM monitor must be supported (CCM_SUP = 1 (YES)) by all ECUs\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data B BIT 3 - Cooling support
	if ( (pPid[0].Data[1] & 0x08) == 0x08 )
	{
		*pulBitsSupported = *pulBitsSupported | COOL_SUP;
	}

	// Evaluate Data B, BIT 4 - Misfire status
	// if Misfire is supported
//	if ( (pPid[0].Data[1] & 0x01) == 0x01)
//	{
//		// if it's a spark engine
//		if ( gOBDDieselFlag == FALSE )
//		{
//			if ( (pPid[0].Data[1] & 0x10) != 0x10 )
//			{
//				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//				      "ECU %X  Misfire Monitor must be 1 (MIS_CMPL = 1 (NO)) for spark ignition vehicles or controllers that do not support misfire (MIS_SUP = 0 (NO)).\n",
//				      GetEcuId ( EcuIndex ) );
//				bTestFailed = TRUE;
//			}
//		}
//		// else it's a compression engine
//		else
//		{
//			// Bit 4 may be 0 or 1 for compression ignition w/ or wo/ Engine running
//			// except in Test 11 where it must be complete (0)./
//			if ( gTestPhase == eTestPerformanceCounters )
//			{
//				if ( (pPid[0].Data[1] & 0x10) != 0x00 )
//				{
//					Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//					      "ECU %X  Misfire Monitor must be 0 (MIS_CMPL = 0 (YES)) for compression ignition vehicles.\n",
//					      GetEcuId ( EcuIndex ) );
//					bTestFailed = TRUE;
//				}
//			}
//		}
//	}
	// If Misfire unsupported it must indicate complete
	if ( (pPid[0].Data[1] & 0x01) == 0x00 &&
	     (pPid[0].Data[1] & 0x10) == 0x10 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Misfire Monitor (MIS_SUP = 0 (NO)) must indicate complete (MIS_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}
	
	// Evaluate Data B, BIT 5 - Fuel status
	if ( (pPid[0].Data[1] & 0x02) == 0x02 )
	{
		// only check for tests 11.3 and 11.7 (not 10.6)
		if ( gTestPhase == eTestPerformanceCounters &&
		     (pPid[0].Data[1] & 0x20) == 0x20 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Supported Fuel System Monitor (FUEL_SUP = 1 (YES)) must indicate complete (FUEL_RDY = 0 (YES)).\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}
	// If fuel system monitor unsupported it must indicate complete
	if ( (pPid[0].Data[1] & 0x02) == 0x00 &&
	     (pPid[0].Data[1] & 0x20) == 0x20 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Fuel System Monitor (FUEL_SUP = 0 (NO)) must indicate complete (FUEL_RDY = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data B, BIT 6 - CCM status
//	if ( (pPid[0].Data[1] & 0x40) == 0x40 )
//	{
//		// NONE
//	}

	// Evaluate Data B, BIT 7 - Cooling status
//	if ( (pPid[0].Data[1] & 0x80) == 0x80 )
//	{
//		// NONE
//	}


	// Evaluate Data C, BIT 0 - CAT support
	if ( pPid[0].Data[2] & 0x01 )
	{
		if ( gOBDDieselFlag != TRUE )
		{
			*pulBitsSupported = *pulBitsSupported | CAT_SUP;
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data C bit 0 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}
	
	// Evaluate Data C, BIT 1 - HCCAT support
	if ( pPid[0].Data[2] & 0x02 )
	{
		if ( gOBDDieselFlag == TRUE )
		{
			*pulBitsSupported = *pulBitsSupported | HCCAT_SUP;
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data C bit 1 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}
	
	// Evaluate Data C, BIT 2 - HCAT support
	if ( pPid[0].Data[2] & 0x04 )
	{
		if ( gOBDDieselFlag == TRUE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data C bit 2 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data C, BIT 3 - SCR support
	if ( pPid[0].Data[2] & 0x08 )
	{
		if ( gOBDDieselFlag == TRUE )
		{
			*pulBitsSupported = *pulBitsSupported | SCR_SUP;
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data C bit 3 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data C, BIT 4 - CAT status
	if ( (pPid[0].Data[2] & 0x01) == 0x00 &&
	     (pPid[0].Data[2] & 0x10) == 0x10 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Catalyst Monitor (CAT_SUP = 0 (NO)) must indicate complete (CAT_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data C, BIT 5 - HCCAT status
	if ( (pPid[0].Data[2] & 0x02) == 0x00 &&
	     (pPid[0].Data[2] & 0x20) == 0x20 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported NMHC Monitor (HCCAT_SUP = 0 (NO)) must indicate complete (HCCAT_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data C, BIT 6 - HCCAT status
	if ( (pPid[0].Data[2] & 0x04) == 0x00 &&
	     (pPid[0].Data[2] & 0x40) == 0x40 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Heated Catalyst Monitor (HCCAT_SUP = 0 (NO)) must indicate complete (HCCAT_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data C, BIT 7 - SCR status
	if ( (pPid[0].Data[2] & 0x08) == 0x00 &&
	     (pPid[0].Data[2] & 0x80) == 0x80 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported SCR Monitor (SCR_SUP = 0 (NO)) must indicate complete (SCR_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}


	// Evaluate Data D, BIT 0 - EVAP support
	if ( pPid[0].Data[3] & 0x01 )
	{
		if ( gOBDDieselFlag != TRUE )
		{
			*pulBitsSupported = *pulBitsSupported | EVAP_SUP;
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data D bit 0 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data D, BIT 1 - AIR-SUP
	if ( pPid[0].Data[3] & 0x02 )
	{
		if ( gOBDDieselFlag == TRUE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data D bit 1 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data D, BIT 2 - BP_SUP
	if ( pPid[0].Data[3] & 0x04 )
	{
		if ( gOBDDieselFlag != TRUE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data D bit 2 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data D, BIT 3 - PM_SUP
	if ( pPid[0].Data[3] & 0x08 )
	{
		if ( gOBDDieselFlag == TRUE )
		{
			*pulBitsSupported = *pulBitsSupported | PM_SUP;
		}
		else
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data D bit 3 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data D, BIT 4 - EVAP status
	if ( (pPid[0].Data[3] & 0x01) == 0x00 &&
	     (pPid[0].Data[3] & 0x10) == 0x10 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Evaporative Monitor (EVAP_SUP = 0 (NO)) must indicate complete (EVAP_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data D, BIT 5 - AIR status
	if ( (pPid[0].Data[3] & 0x02) == 0x00 &&
	     (pPid[0].Data[3] & 0x20) == 0x20 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Air Monitor (AIR_SUP = 0 (NO)) must indicate complete (AIR_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data D, BIT 6 - BP status
	if ( (pPid[0].Data[3] & 0x04) == 0x00 &&
	     (pPid[0].Data[3] & 0x40) == 0x40 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Boost Monitor (BP_SUP = 0 (NO)) must indicate complete (BP_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data D, BIT 7 - PM status
	if ( (pPid[0].Data[3] & 0x08) == 0x00 &&
	     (pPid[0].Data[3] & 0x80) == 0x80 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported DPF Monitor (PM_SUP = 0 (NO)) must indicate complete (PM_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}


	// Evaluate Data E, BIT 0 - EGS support
	if ( (pPid[0].Data[4] & 0x01) != 0x01 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  EGS monitor must be supported (EGS_SUP = 1 (YES)) by all ECUs\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data E, BIT 1 - PCVCV support
	if ( (pPid[0].Data[4] & 0x02) != 0x02 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  PCVCV monitor must be supported (PCVCV_SUP = 1 (YES)) by all ECUs\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data E, BIT 2 - EGR_SUP
	if ( (pPid[0].Data[4] & 0x04) != 0x04 )
	{
		if ( gOBDDieselFlag == TRUE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  EGR monitor must be supported (EGR_SUP = 1 (YES)) by all compression ignition engine ECUs\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data E, BIT 3 - VVT_SUP
//	if ( pPid[0].Data[4] & 0x08 )
//	{
//		// NONE
//	}

	// Evaluate Data E, BIT 4 - EGS status
//	if ( (pPid[0].Data[4] & 0x01) == 0x00 &&
//	     (pPid[0].Data[4] & 0x10) == 0x10 )
//	{
//		// NONE
//	}

	// Evaluate Data E, BIT 5 - PCVCV status
//	if ( (pPid[0].Data[4] & 0x02) == 0x00 &&
//	     (pPid[0].Data[4] & 0x20) == 0x20 )
//	{
//		// NONE
//	}

	// Evaluate Data E, BIT 6 - EGR status
	if ( (pPid[0].Data[3] & 0x04) == 0x00 &&
	     (pPid[0].Data[3] & 0x40) == 0x40 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported EGR Monitor (EGR_SUP = 0 (NO)) must indicate complete (EGR_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data E, BIT 7 - EGR status
	if ( (pPid[0].Data[3] & 0x08) == 0x00 &&
	     (pPid[0].Data[3] & 0x80) == 0x80 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported EGR Monitor (EGR_SUP = 0 (NO)) must indicate complete (EGR_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}


	// Evaluate Data F, BIT 0 - DOR_SUP
//	if ( pPid[0].Data[5] & 0x01 )
//	{
//		// NONE
//	}

	// Evaluate Data F, BIT 1 - CSERS_SUP
	if ( (pPid[0].Data[5] & 0x02) != 0x02)
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  CSERS monitor must be supported (CSERS_SUP = 1 (YES)) by all ECUs\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}
	
	// Evaluate Data F, BIT 2 - NAC_SUP
	if ( (pPid[0].Data[5] & 0x04) == 0x04 )
	{
		if ( gOBDDieselFlag != TRUE )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "ECU %X  Reserved I/M readiness Data F bit 2 set\n",
			      GetEcuId ( EcuIndex ) );
			bTestFailed = TRUE;
		}
	}

	// Evaluate Data F, BIT 3 - ERS_SUP
//	if ( pPid[0].Data[5] & 0x08 )
//	{
//	// NONE
//	}

	// Evaluate Data F, BIT 4 - DOR status
	if ( (pPid[0].Data[5] & 0x01) == 0x00 &&
	     (pPid[0].Data[5] & 0x10) == 0x10 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Direct Ozone Radiator Monitor (DOR_SUP = 0 (NO)) must indicate complete (DOR_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data F, BIT 5 - CSERS status
//	if ( (pPid[0].Data[5] & 0x02) == 0x00 &&
//	     (pPid[0].Data[5] & 0x20) == 0x20 )
//	{
//		// NONE
//	}

	// Evaluate Data F, BIT 6 - NAC status
	if ( (pPid[0].Data[5] & 0x04) == 0x00 &&
	     (pPid[0].Data[5] & 0x40) == 0x40 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported NAC Monitor (NAC_SUP = 0 (NO)) must indicate complete (NAC_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}

	// Evaluate Data F, BIT 7 - ERS status
	if ( (pPid[0].Data[5] & 0x08) == 0x00 &&
	     (pPid[0].Data[5] & 0x80) == 0x80 )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "ECU %X  Unsupported Other Emission-Related Strategy Monitor (ERS_SUP = 0 (NO)) must indicate complete (ERS_CMPL = 0 (YES)).\n",
		      GetEcuId ( EcuIndex ) );
		bTestFailed = TRUE;
	}


//	if ( gOBDEngineRunning == FALSE &&
//	     gTestPhase != eTestPerformanceCounters )  // test 5.6 & 10.6
//	{
//		/------------------------------------------------------------------------
//		// supported monitors should not be complete (except O2 on spark ignition)
//	
//		// DATA_D Bit 0
//		if ( (pPid[0].Data[2] & 0x01) == 0x01 &&
//		     (pPid[0].Data[3] & 0x01) == 0x00 )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
//			      GetEcuId ( EcuIndex ),
//			      gOBDDieselFlag ? "HCCAT" : "CAT",
//			      gOBDDieselFlag ? "HCCAT" : "CAT" );
//			bTestFailed = TRUE;
//		}
//	
//		// DATA_D Bit 1
//		if ( (pPid[0].Data[2] & 0x02) == 0x02 &&
//		     (pPid[0].Data[3] & 0x02) == 0x00 )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
//			      GetEcuId ( EcuIndex ),
//			      gOBDDieselFlag ? "NCAT" : "HCAT",
//			      gOBDDieselFlag ? "NCAT" : "HCAT" );
//			bTestFailed = TRUE;
//		}
//	
//		// DATA_D Bit 2
//		if ( (pPid[0].Data[2] & 0x04) == 0x04 &&
//		     (pPid[0].Data[3] & 0x04) == 0x00 )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
//			      GetEcuId ( EcuIndex ),
//			      gOBDDieselFlag ? "RESERVED2" : "EVAP",
//			      gOBDDieselFlag ? "RESERVED2" : "EVAP" );
//			bTestFailed = TRUE;
//		}
//	
//		// DATA_D Bit 3
//		if ( (pPid[0].Data[2] & 0x08) == 0x08 &&
//		     (pPid[0].Data[3] & 0x08) == 0x00 )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
//			      GetEcuId ( EcuIndex ),
//			      gOBDDieselFlag ? "BP" : "AIR",
//			      gOBDDieselFlag ? "BP" : "AIR" );
//			bTestFailed = TRUE;
//		}
//	
//		// DATA_D Bit 4
//		if ( (pPid[0].Data[2] & 0x10) == 0x10 &&
//		     (pPid[0].Data[3] & 0x10) == 0x00 )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supported monitor (RESERVED Byte C Bit 4 = 1) must indicate incomplete (RESERVED Byte D Bit 4 = 1)\n",
//			      GetEcuId ( EcuIndex ) );
//			bTestFailed = TRUE;
//		}
//	
//		// DATA_D Bit 5
//		if ( (pPid[0].Data[2] & 0x20) == 0x20 &&
//		     (pPid[0].Data[3] & 0x20) == 0x00 )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
//			      GetEcuId ( EcuIndex ),
//			      gOBDDieselFlag ? "EGS" : "O2S",
//			      gOBDDieselFlag ? "EGS" : "O2S" );
//			bTestFailed = TRUE;
//		}
//	
//		// DATA_D Bit 6
//		if ( (pPid[0].Data[2] & 0x40) == 0x40 &&
//		     (pPid[0].Data[3] & 0x40) == 0x00 &&
//		     gOBDDieselFlag )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supported monitor (%s_SUP = 1 (YES)) must indicate incomplete (%s_RDY = 1 (NO))\n",
//			      GetEcuId ( EcuIndex ),
//			      gOBDDieselFlag ? "PM" : "HTR",
//			      gOBDDieselFlag ? "PM" : "HTR" );
//			bTestFailed = TRUE;
//		}
//	
//		// DATA_D Bit 7
//		if ( (pPid[0].Data[2] & 0x80) == 0x80 &&
//		     (pPid[0].Data[3] & 0x80) == 0x00 )
//		{
//			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
//			      "ECU %X  Supported monitor (EGR_SUP = 1 (YES)) must indicate incomplete (EGR_RDY = 1 (NO))\n",
//			      GetEcuId ( EcuIndex ) );
//			bTestFailed = TRUE;
//		}
//	}  // end if gOBDEngineRunning == FALSE && gTestPhase != eTestPerformanceCounters )  // test 5.6 & 10.6


	if ( bTestFailed == TRUE )
	{
		return FAIL;
	}
	else
	{
		return PASS;
	}
}


/*
********************************************************************************
**  FUNCTION    VerifyIM_Ready
**
**  Purpose     Isolated function to request PID $F501 and check IM Ready
********************************************************************************
*/
STATUS VerifyIM_Ready ( void )
{
	SID_REQ       SidReq;
	unsigned long EcuIndex;

	unsigned long temp_data_long;
	unsigned short PidDataCnt = 0;

	BOOL          bTestFailed = FALSE;

	PID          *pPid;

	if ( IsDIDSupported ( -1, 0xF501 ) == FALSE )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "PID $F501 not supported\n" );
		bTestFailed = TRUE;
	}

	SidReq.SID    = 0x22;
	SidReq.NumIds = 1;
	SidReq.DID[0] = 0xF501;
	if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
	{
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "PID $F501 request\n" );
		bTestFailed = TRUE;
	}
	else
	{
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( gOBDResponse[EcuIndex].PidSize != 0 )
			{
				PidDataCnt++;
			
				pPid = (PID *)&gOBDResponse[EcuIndex].Pid[0];

				// Verify that all PID data is valid
				if ( VerifyPIDF501 ( pPid, EcuIndex, &temp_data_long ) != PASS )
				{
					bTestFailed = TRUE;
				}
			}
		}

		if ( PidDataCnt == 0 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "No PID $F501 data\n" );
			bTestFailed = TRUE;
		}
	}

	// if this is test 11.3, check for appropriate value of PID $F431 and $F44E
	if ( gTestPhase == eTestPerformanceCounters && gTestSubsection == 3 )
	{
		// If PID $31 is supported, request it
		if ( IsDIDSupported ( -1, 0xF431 ) == TRUE )
		{
			SidReq.SID    = 0x22;
			SidReq.NumIds = 1;
			SidReq.DID[0] = 0xF431;
			if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "PID $F431 request\n" );
				bTestFailed = TRUE;
			}
			else
			{
				// Check the data to see if it is valid
				for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
				{
					if ( IsDIDSupported ( EcuIndex, 0xF431 ) == TRUE )
					{
						pPid = (PID *)&gOBDResponse[EcuIndex].Pid[0];
						temp_data_long = ((pPid->Data[0] * 256) +
						                  pPid->Data[1]);
						if ( gOBDResponse[EcuIndex].PidSize != 0 )
						{
							if ( temp_data_long == 0 )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F431 is 0\n",
								      GetEcuId ( EcuIndex ) );
								bTestFailed = TRUE;
							}
						}
						else
						{
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "ECU %X  PID $F431 size is 0\n",
							      GetEcuId ( EcuIndex ) );
							bTestFailed = TRUE;
						}
					}
				}
			}
		}

		// If PID $4E is supported, request it
		if ( IsDIDSupported ( -1, 0xF44E ) == TRUE )
		{
			SidReq.SID    = 0x22;
			SidReq.NumIds = 1;
			SidReq.DID[0] = 0xF44E;
			if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
			{
				Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "PID $F44E request\n" );
				bTestFailed = TRUE;
			}
			else
			{
				// Check the data to see if it is valid
				for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++)
				{
					if ( IsDIDSupported ( EcuIndex, 0xF44E ) == TRUE )
					{
						pPid = (PID *)&gOBDResponse[EcuIndex].Pid[0];
						temp_data_long = ((pPid->Data[0] * 256) +
						                  pPid->Data[1]);
						if ( gOBDResponse[EcuIndex].PidSize != 0 )
						{
							if ( temp_data_long == 0 )
							{
								Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
								      "ECU %X  PID $F44E is 0\n",
								      GetEcuId ( EcuIndex ) );
								bTestFailed = TRUE;
							}
						}
						else
						{
							Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
							      "ECU %X  PID $F44E size is 0\n",
							      GetEcuId ( EcuIndex ) );
							bTestFailed = TRUE;
						}
					}
				}
			}
		}
	}

	if (bTestFailed == TRUE)
	{
		return FAIL;
	}
	else
	{
		return PASS;
	}
}


/*
*****************************************************************************
**
**  Function:  DetermineVariablePidSize
**
**  Purpose:   Determine number of data bytes in PIDs $06 - $09, $55 - $58
**
*****************************************************************************
*/
STATUS DetermineVariablePidSize ( void )
{
	SID_REQ SidReq;

	PID    *pPid1;
	unsigned char pid[OBD_MAX_ECUS];
	unsigned long EcuIdIndex[OBD_MAX_ECUS];

	unsigned long EcuIndex, numResp;

	// only need to check once
	if ( gSid1VariablePidSize != 0 )
	{
		return PASS;
	}

	// -1 ==> cannot determine the PID size
	gSid1VariablePidSize = -1;

	// only check if needed
	if ( IsDIDSupported ( -1, 0xF406 ) == FALSE &&
	     IsDIDSupported ( -1, 0xF407 ) == FALSE &&
	     IsDIDSupported ( -1, 0xF408 ) == FALSE &&
	     IsDIDSupported ( -1, 0xF409 ) == FALSE &&
	     IsDIDSupported ( -1, 0xF455 ) == FALSE &&
	     IsDIDSupported ( -1, 0xF456 ) == FALSE &&
	     IsDIDSupported ( -1, 0xF457 ) == FALSE &&
	     IsDIDSupported ( -1, 0xF458 ) == FALSE )
	{
		return PASS;
	}

	// cannot support both PID $13 and $1D
	if ( IsDIDSupported ( -1, 0xF413 ) == TRUE &&
	     IsDIDSupported ( -1, 0xF41D ) == TRUE )
	{
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( IsDIDSupported ( EcuIndex, 0xF413 ) == TRUE )
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Supports PID $F413\n",
				      GetEcuId ( EcuIndex ));
			}

			if ( IsDIDSupported ( EcuIndex, 0xF41D ) == TRUE )
			{
				Log ( INFORMATION, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
				      "ECU %X  Supports PID $F41D\n",
				      GetEcuId ( EcuIndex ));
			}
		}
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      "Both PID $F413 and $F41D are supported\n" );
		return FAIL;
	}

	// check PID $13 first
	if ( IsDIDSupported ( -1, 0xF413 ) == TRUE )
	{
		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = 0xF413;

		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $F413 request\n" );
			return FAIL;
		}

		numResp = 0;
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( IsDIDSupported ( EcuIndex, 0xF413 ) == TRUE )
			{
				if ( gOBDResponse[EcuIndex].PidSize > 0 )
				{
					pPid1 = (PID *)(gOBDResponse[EcuIndex].Pid);
					if ( pPid1->PIDLSB == 0x13 )
					{
						EcuIdIndex[numResp] = EcuIndex;
						pid[numResp++] = pPid1->Data[0];
					}
				}
			}
		}

		if ( numResp == 0 || numResp > 2 )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $F413 supported by %d ECUs (1 to 2 expected)\n",
			      numResp);
			return FAIL;
		}

		gSid1VariablePidSize = 1;
		return PASS;
	}

	// check PID $1D second
	if ( IsDIDSupported ( -1, 0xF41D ) == TRUE )
	{
		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = 0xF41D;

		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $F41D request\n" );
			return FAIL;
		}

		numResp = 0;
		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( IsDIDSupported ( EcuIndex, 0xF41D ) == TRUE )
			{
				if ( gOBDResponse[EcuIndex].PidSize > 0 )
				{
					pPid1 = (PID *)(gOBDResponse[EcuIndex].Pid);
					if ( pPid1->PIDLSB == 0x1D )
					{
						EcuIdIndex[numResp] = EcuIndex;
						pid[numResp++] = pPid1->Data[0];
					}
				}
			}
		}

		if ( numResp == 1 )
		{
			gSid1VariablePidSize = (pid[0] == 0xff) ? 2 : 1;
			return PASS;
		}

		if ( numResp == 2 )
		{
			if ( (pid[0] | pid[1]) != 0xff)
			{
				gSid1VariablePidSize = 1;
				return PASS;
			}

			if ( (pid[0] == 0xf0 && pid[1] == 0x0f) ||
			     (pid[0] == 0x0f && pid[1] == 0xf0) )
			{
				gSid1VariablePidSize = 2;
				return PASS;
			}

			if ( (pid[0] == 0xcc && pid[1] == 0x33) ||
			     (pid[0] == 0x33 && pid[1] == 0xcc) )
			{
				gSid1VariablePidSize = 2;
				return PASS;
			}

			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "Unable to determine number of data bytes in PIDs $F406 - $F409\n" );
			return FAIL;
		}

		// numResp == 0 or numResp > 2
		Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
		      " PID $F41D supported by %d ECUs ( 1 to 2 expected)\n",
		      numResp);
		return FAIL;
	}

	Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
	      "Neither PID $F413 nor $F41D are supported\n" );
	return FAIL;
}


/*
*****************************************************************************
**
**  Function:  GetPID4FArray
**
**  Purpose:   copy PID $4F values into an array.
**
*****************************************************************************
*/
STATUS GetPID4FArray ( void )
{
	SID_REQ SidReq;
	PID *pPid1;
	unsigned int EcuIndex;

	for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
	{
		Pid4F[EcuIndex][0] = 0x00;
		Pid4F[EcuIndex][1] = 0x00;
		Pid4F[EcuIndex][2] = 0x00;
		Pid4F[EcuIndex][3] = 0x00;
	}

	if ( IsDIDSupported ( -1, 0xF44F ) == TRUE )
	{
		SidReq.SID = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = 0xF44F;

		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $F44F request\n" );
			return FAIL;
		}

		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( IsDIDSupported ( EcuIndex, 0xF44F ) == TRUE )
			{
				if ( gOBDResponse[EcuIndex].PidSize > 0 )
				{
					pPid1 = (PID *)(gOBDResponse[EcuIndex].Pid);
					if ( pPid1->PIDLSB == 0x4F )
					{
						Pid4F[EcuIndex][0] = pPid1->Data[0];
						Pid4F[EcuIndex][1] = pPid1->Data[1];
						Pid4F[EcuIndex][2] = pPid1->Data[2];
						Pid4F[EcuIndex][3] = pPid1->Data[3];
					}
				}
			}
		}
	}
	return PASS;
}


/*
*****************************************************************************
**
**  Function:  GetPID50Array
**
**  Purpose:   copy PID $50 values into an array.
**
*****************************************************************************
*/
STATUS GetPID50Array ( void )
{
	SID_REQ SidReq;
	PID *pPid1;
	unsigned int EcuIndex;

	if ( IsDIDSupported ( -1, 0xF450 ) == TRUE )
	{
		SidReq.SID    = 0x22;
		SidReq.NumIds = 1;
		SidReq.DID[0] = 0xF450;

		if ( SidRequest ( &SidReq, SID_REQ_NORMAL ) == FAIL )
		{
			Log ( FAILURE, SCREENOUTPUTON, LOGOUTPUTON, NO_PROMPT,
			      "PID $F450 request\n" );
			return FAIL;
		}

		for ( EcuIndex = 0; EcuIndex < gOBDNumEcus; EcuIndex++ )
		{
			if ( IsDIDSupported ( EcuIndex, 0xF450 ) == TRUE )
			{
				if ( gOBDResponse[EcuIndex].PidSize > 0 )
				{
					pPid1 = (PID *)(gOBDResponse[EcuIndex].Pid);
					if ( pPid1->PIDLSB == 0x50 )
					{
						Pid50[EcuIndex][0] = pPid1->Data[0];
						Pid50[EcuIndex][1] = pPid1->Data[1];
						Pid50[EcuIndex][2] = pPid1->Data[2];
						Pid50[EcuIndex][3] = pPid1->Data[3];
					}
				}
			}
		}
	}
	return PASS;
}


/*
*****************************************************************************
**
**  Function:  GetHoursMinsSecs
**
**  Purpose:   Convert the time (in seconds) into hours, minutes and seconds .
**
*****************************************************************************
*/
void GetHoursMinsSecs ( unsigned long time, unsigned long *hours, unsigned long *mins, unsigned long *secs )
{

	*hours = time/3600;

	*mins = (time - (*hours * 3600)) / 60;

	*secs = time - (*hours * 3600) - (*mins * 60);
}
